// This example shows how to do kriging. It is based on the Matern
// kernel on a regular grid. The Matern kernel is parameterized by
// scaling factors, one for each dimension. The factors are
// predefined; but in practice they come from a maximum likelihood
// estimation. In the process of kriging, the linear solves use a
// Toeplitz solver with circulant preconditioning.

#include <mpi.h>
#include <math.h>
#include <iostream>
#include <ScalaGAUSS.hpp>

void GenRandPoint(int NDim, long int *Dim, double *Range, double *Coord);
void GenGridPoint(int NDim, long int *Dim, double *Range, double *Coord);

int main(int argc, char *argv[]) {

  MPI_Init(&argc, &argv);
  MPI_Comm Comm = MPI_COMM_WORLD;

  //-------------------------------------------------------------------------
  // Set up parameters

  // Grid
  int NDim = 3;
  long Dim[] = {10, 10, 10};
  GRID_TYPE GridType = FIXED_DOMAIN;
  double Range[] = {100.0, 100.0, 100.0};

  // Partitioning
  int NProc[] = {2, 1, 1};

  // Kernel Param
  double Nu = 1.0;
  int LenTheta = NDim;
  double Theta[] = {10.0, 13.0, 7.0};
  bool bNeedPartials[] = {true, true, true};

  // Linear solver
  double CG_RTol = 1e-10;
  int CG_MaxIt = 200;
  bool bElimAllReduce = true;

  try {

  //-------------------------------------------------------------------------
  // Setup everything

  // Grid
  RegularGrid Grid(Comm);
  Grid.SetGrid(NDim, Dim, GridType, Range, NProc, bElimAllReduce);

  // GPSample
  RegularGridData GPSample(Comm);
  GPSample.SetGrid(Grid);
  GPSample.SetSymBernoulli(); // bogus data

  // Kernel
  Matern Kernel;
  Kernel.Setup(NDim, Nu, LenTheta, Theta);

  // Toeplitz solver param
  ToepSolverParam mToepSolverParam(CG_RTol, CG_MaxIt);

  // CovarMat
  RegularGridCovarMat<Matern> CovarMat(Comm);
  CovarMat.SetGrid(Grid);
  CovarMat.SetToepSolverParam(mToepSolverParam);
  CovarMat.UseKernel(Kernel, bNeedPartials);

  // CovarVec
  RegularGridCovarVec<Matern> CovarVec(Comm);
  CovarVec.SetGrid(Grid);

  // Kriging
  Kriging<RegularGridData, RegularGridCovarMat<Matern> > K;

  //-------------------------------------------------------------------------
  // Krige
  double Coord[3], Prediction, Variance;

  int Rank;
  MPI_Comm_rank(Comm, &Rank);

  // Krige a general location
  GenRandPoint(NDim, Dim, Range, Coord);
  CovarVec.SetVal(Kernel, Coord);
  K.SimpleKriging(GPSample, CovarMat, (RegularGridData &)CovarVec,
                  &Prediction, &Variance);
  if (Rank == 0) {
    printf("Krige a general location. Prediction = %g, Variance = %g\n",
           Prediction, Variance);
  }

  // Krige a grid point
  GenGridPoint(NDim, Dim, Range, Coord);
  CovarVec.SetVal(Kernel, Coord);
  K.SimpleKriging(GPSample, CovarMat, (RegularGridData &)CovarVec,
                  &Prediction, &Variance);
  if (Rank == 0) {
    printf("Krige a grid point (prediction should be +/-1, variance shoule be 0). Prediction = %g, Variance = %g\n",
           Prediction, Variance);
  }

  }
  catch (std::string Msg) {
    ExceptionHandler(Msg, Comm);
  }

  MPI_Finalize();
  return 0;

}

//-------------------------------------------------------------------------
void GenRandPoint(int NDim, long int *Dim, double *Range, double *Coord) {
  int i;
  srandom(time(NULL));
  for (i = 0; i < NDim; i++) {
    Coord[i] = (double)random() / RAND_MAX * Range[i];
  }
}

//-------------------------------------------------------------------------
void GenGridPoint(int NDim, long int *Dim, double *Range, double *Coord) {
  int i;
  srandom(time(NULL));
  for (i = 0; i < NDim; i++) {
    Coord[i] = ( Range[i]/(Dim[i]-1) ) * ( random()%Dim[i] );
  }
}

// Author: J. Chen
// $Date: 2013/03/28 15:50:43$
