// This example shows how to compute MLE. It uses the PETSC SNES
// library to solve the nonlinear score equations. The scores are
// approximated by using sample average approximation (SAA). This
// example uses the Matern kernel on a regular grid, possibly with
// filtering. The Matern kernel is parameterized by scaling factors,
// one for each dimension. In the evaluation of the score functions,
// the linear solves use a Toeplitz solver with circulant
// preconditioning.

// Use the PETSC option -help to show this help
static char help[] = "Maximum likelihood estimation for large-scale Gaussian\n\
process on 3D regular grid with Matern kernel. Arguments:\n\
 -Dim           : Size of grid. Default: 20,20,20\n\
 -Range         : Physical range of grid. Default: 100.0,100.0,100.0\n\
 -NProc         : Process grid. Default: <np>,1,1\n\
 -Nu            : Matern smoothness. Default: 1.0\n\
 -GuessTheta    : Initial guess of parameters. Default: 9.0,14.0,8.0\n\
 -NumRandVec    : Number of random vectors in SAA. Default: 50\n\
 -NumLapFilter  : Number of filtering applied to process. Default is optimal\n\
 -CG_RTol       : Relative residual tolerance of linear solve. Default: 1e-10\n\
 -CG_MaxIt      : Max iteration count of linear solve. Default: 200\n\
 -ElimAllReduce : Whether eliminate Allreduce in linear solve. Default: 1\n\
 -DataFile      : (Mandatory) Data file.\n\n";

#include <mpi.h>
#include <math.h>
#include <iostream>
#include "ExampleMLE.hpp"

int main(int argc, char **argv) {

  PetscErrorCode ierr;
  ierr = PetscInitialize(&argc, &argv, (char *)0, help); CHKERRQ(ierr);
  MPI_Comm Comm = MPI_COMM_WORLD;

  try {

  //-------------------------------------------------------------------------
  // Setup context for nonlinear solver

  // MLE param
  MLEParam mMLEParam(Comm);
  mMLEParam.GetFromOptions();
  mMLEParam.PrintParams();

  // Grid
  RegularGrid Grid(Comm);
  Grid.SetGrid(mMLEParam.NDim, mMLEParam.Dim, mMLEParam.GridType,
               mMLEParam.Range, mMLEParam.NProc, mMLEParam.bElimAllReduce);

  // Data
  double *MyData = NULL;
  long MyDataLen = Grid.GetPartitioning(NULL, NULL);
  New_1D_Array<double, long>(&MyData, MyDataLen);
  ReadData(mMLEParam.DataFile, Grid, MyData, Comm);

  // Start clock here
  // I do not want to include the IO time for reading data
  double t1, t2;
  MPI_Barrier(Comm);
  t1 = MPI_Wtime();

  // GPSample
  RegularGridData GPSample(Comm);
  GPSample.SetGrid(Grid);
  GPSample.SetVal(MyData, MyDataLen);
  Delete_1D_Array<double>(&MyData);
  GPSample.ApplyLapFilter(mMLEParam.NumLapFilter);

  // Kernel
  Matern Kernel;
  Kernel.Setup(mMLEParam.NDim, mMLEParam.Nu, mMLEParam.LenTheta,
               mMLEParam.Theta0);

  // Toeplitz solver param
  ToepSolverParam mToepSolverParam(mMLEParam.CG_RTol, mMLEParam.CG_MaxIt);

  // CovarMat
  RegularGridCovarMat<Matern> CovarMat(Comm);
  CovarMat.SetGrid(Grid);
  CovarMat.SetToepSolverParam(mToepSolverParam);
  CovarMat.UseKernel(Kernel, mMLEParam.bNeedPartials);
  CovarMat.ApplyLapFilter(mMLEParam.NumLapFilter);

  // ScoreFunc
  ApproxScoreFunc<RegularGridData, RegularGridCovarMat<Matern> > S;
  S.SetGPSample(&GPSample);
  S.SetCovarMat(&CovarMat);
  S.SetNumRandVec(mMLEParam.NumRandVec);

  // F and J
  double *F = NULL;
  New_1D_Array<double, int>(&F, mMLEParam.LenTheta);
  double **J = NULL;
  New_2D_Array<double, int, int>(&J, mMLEParam.LenTheta, mMLEParam.LenTheta);

  // Wrap everything in Ctx (which is passed in to the snes solver)
  ScoreCtx Ctx(&Grid, &Kernel, &CovarMat, &S, mMLEParam.LenTheta,
               mMLEParam.NumLapFilter, mMLEParam.bNeedPartials, F, J);

  //-------------------------------------------------------------------------
  // Setup nonlinear solver and solve

  // Create nonlinear solver context
  SNES snes;
  ierr = SNESCreate(PETSC_COMM_SELF, &snes); CHKERRQ(ierr);

  // Create vector data structures
  Vec theta, f;
  ierr = VecCreate(PETSC_COMM_SELF, &theta); CHKERRQ(ierr);
  ierr = VecSetSizes(theta, PETSC_DECIDE, mMLEParam.LenTheta); CHKERRQ(ierr);
  ierr = VecSetFromOptions(theta); CHKERRQ(ierr);
  ierr = VecDuplicate(theta, &f); CHKERRQ(ierr);

  // Set function evaluation routine
  ierr = SNESSetFunction(snes, f, FormFunction, (void*)(&Ctx)); CHKERRQ(ierr);

  // Create matrix data structure
  Mat jac;
  ierr = MatCreate(PETSC_COMM_SELF, &jac); CHKERRQ(ierr);
  ierr = MatSetSizes(jac, PETSC_DECIDE, PETSC_DECIDE, mMLEParam.LenTheta,
                     mMLEParam.LenTheta); CHKERRQ(ierr);
  ierr = MatSetFromOptions(jac); CHKERRQ(ierr);
  ierr = MatSeqAIJSetPreallocation(jac, mMLEParam.LenTheta,
                                   PETSC_NULL); CHKERRQ(ierr);

  // Set Jacobian evaluation routine
  ierr = SNESSetJacobian(snes, jac, jac, FormJacobian,
                         (void*)(&Ctx)); CHKERRQ(ierr);

  // Customize nonlinear solver; set runtime options
  ierr = SNESSetFromOptions(snes); CHKERRQ(ierr);

  // Set monitoring routine
  ierr = SNESMonitorSet(snes, Monitor, (void*)(&Ctx), 0);CHKERRQ(ierr);

  // Set initial guess
  ierr = FormInitialGuess(theta, mMLEParam.Theta0,
                          mMLEParam.LenTheta); CHKERRQ(ierr);

  // Solve nonlinear system
  ierr = SNESSolve(snes, PETSC_NULL, theta); CHKERRQ(ierr);

  //-------------------------------------------------------------------------
  // Solution summary

  // Estimates
  ierr = SNESGetSolution(snes, &theta); CHKERRQ(ierr);

  // Variance of the estimates
  double **I = NULL;
  New_2D_Array<double, int, int>(&I, mMLEParam.LenTheta, mMLEParam.LenTheta);
  double **invI = NULL;
  New_2D_Array<double, int, int>(&invI, mMLEParam.LenTheta, mMLEParam.LenTheta);
  double **G = NULL;
  New_2D_Array<double, int, int>(&G, mMLEParam.LenTheta, mMLEParam.LenTheta);
  double **invG = NULL;
  New_2D_Array<double, int, int>(&invG, mMLEParam.LenTheta, mMLEParam.LenTheta);
  S.GetVariance(I, invI, G, invG);

  PrintSolSummary(mMLEParam.LenTheta, theta, I, invI, G, invG, Comm);

  //-------------------------------------------------------------------------
  // Clean up

  ierr = VecDestroy(&theta); CHKERRQ(ierr);
  ierr = VecDestroy(&f); CHKERRQ(ierr);
  ierr = MatDestroy(&jac); CHKERRQ(ierr);
  ierr = SNESDestroy(&snes); CHKERRQ(ierr);

  Delete_1D_Array<double>(&F);
  Delete_2D_Array<double, int>(&J, mMLEParam.LenTheta);
  Delete_2D_Array<double, int>(&I, mMLEParam.LenTheta);
  Delete_2D_Array<double, int>(&invI, mMLEParam.LenTheta);
  Delete_2D_Array<double, int>(&G, mMLEParam.LenTheta);
  Delete_2D_Array<double, int>(&invG, mMLEParam.LenTheta);

  // End clock here
  t2 = MPI_Wtime();
  PrintTimeStat(t1, t2, "Time:", Comm);

  }
  catch (std::string Msg) {
    ExceptionHandler(Msg, Comm);
  }

  ierr = PetscFinalize(); CHKERRQ(ierr);
  return 0;

}

//-------------------------------------------------------------------------
PetscErrorCode FormInitialGuess(Vec theta0, double *Theta0, int LenTheta) {
  double *tmp = NULL;
  VecGetArray(theta0, &tmp);
  Set_Array<double, int>(tmp, Theta0, LenTheta);
  VecRestoreArray(theta0, &tmp);
  return 0;
}

//-------------------------------------------------------------------------
PetscErrorCode FormFunction(SNES snes, Vec theta, Vec f, void *ctx) {

  // Context
  ScoreCtx *Ctx = (ScoreCtx *)ctx;

  // Convert theta to Theta, f to F
  double *Theta = NULL, *F = NULL;
  VecGetArrayRead(theta, (const PetscScalar**)&Theta);
  VecGetArray(f, &F);

  // Calculation
  if (Ctx->Kernel->IsSameTheta(Theta)) {
    Ctx->S->GetScores(F, false);
  }
  else {
    Ctx->Kernel->SetTheta(Theta, Ctx->LenTheta);
    Ctx->CovarMat->SetGrid(*(Ctx->Grid));
    Ctx->CovarMat->UseKernel(*(Ctx->Kernel), Ctx->bNeedPartials);
    Ctx->CovarMat->ApplyLapFilter(Ctx->NumLapFilter);
    Ctx->S->SetCovarMat(Ctx->CovarMat);
    Ctx->S->GetScores(F, false);
  }

  // Restore theta and f
  VecRestoreArrayRead(theta, (const PetscScalar**)&Theta);
  VecRestoreArray(f, &F);

  return 0;
}

//-------------------------------------------------------------------------
PetscErrorCode FormJacobian(SNES snes, Vec theta, Mat *jac,
                            Mat *B, MatStructure *flag, void *ctx) {

  // Context
  ScoreCtx *Ctx = (ScoreCtx *)ctx;

  // Convert theta to Theta and allocate Jacobian J
  double *Theta = NULL, **J = NULL;
  VecGetArrayRead(theta, (const PetscScalar**)&Theta);
  New_2D_Array<double, int>(&J, Ctx->LenTheta, Ctx->LenTheta);

  // Calculation
  if (Ctx->Kernel->IsSameTheta(Theta)) {
    Ctx->S->GetJacobian(J);
  }
  else {
    Ctx->Kernel->SetTheta(Theta, Ctx->LenTheta);
    Ctx->CovarMat->SetGrid(*(Ctx->Grid));
    Ctx->CovarMat->UseKernel(*(Ctx->Kernel), Ctx->bNeedPartials);
    Ctx->CovarMat->ApplyLapFilter(Ctx->NumLapFilter);
    Ctx->S->SetCovarMat(Ctx->CovarMat);
    Ctx->S->GetJacobian(J);
  }

  // Restore theta and set jac
  VecRestoreArrayRead(theta, (const PetscScalar**)&Theta);
  int i;
  int *j = NULL;
  New_1D_Array<int, int>(&j, Ctx->LenTheta);
  for (i = 0; i < Ctx->LenTheta; i++) {
    j[i] = i;
  }
  for (i = 0; i < Ctx->LenTheta; i++) {
    MatSetValues(*jac, 1, &i, Ctx->LenTheta, j, J[i], INSERT_VALUES);
  }
  MatAssemblyBegin(*jac, MAT_FINAL_ASSEMBLY);
  MatAssemblyEnd(*jac, MAT_FINAL_ASSEMBLY);
  if (*B != *jac) {
    MatAssemblyBegin(*B, MAT_FINAL_ASSEMBLY);
    MatAssemblyEnd(*B, MAT_FINAL_ASSEMBLY);
  }

  // Cleanup
  Delete_2D_Array<double, int>(&J, Ctx->LenTheta);
  Delete_1D_Array<int>(&j);

  return 0;
}

//-------------------------------------------------------------------------
PetscErrorCode Monitor(SNES snes, PetscInt its, PetscReal fnorm, void *ctx) {
  ScoreCtx *Ctx = (ScoreCtx *)ctx;
  PetscErrorCode ierr;
  Vec x;
  const PetscScalar *xx;
  PetscInt i;
  ierr = SNESGetSolution(snes, &x); CHKERRQ(ierr);
  ierr = VecGetArrayRead(x, &xx); CHKERRQ(ierr);
  ierr = PetscPrintf(PETSC_COMM_WORLD, "%D   Theta: ", its); CHKERRQ(ierr);
  for (i = 0; i < Ctx->LenTheta; i++) {
    ierr = PetscPrintf(PETSC_COMM_WORLD, "%G ", xx[i]); CHKERRQ(ierr);
  }
  ierr = PetscPrintf(PETSC_COMM_WORLD, "  SNES Function norm %G\n",
                     fnorm); CHKERRQ(ierr);
  ierr = VecRestoreArrayRead(x, &xx); CHKERRQ(ierr);
  return 0;
}

//-------------------------------------------------------------------------
void ReadData(char *filename, RegularGrid &grid, double *data, MPI_Comm &comm) {
  // Open file
  MPI_File fh;
  if (MPI_File_open(comm, filename, MPI_MODE_RDONLY, MPI_INFO_NULL, &fh)) {
    throw std::string("Error: Fail to open ") + std::string(filename)
      + std::string("!\n");
  }

  // Grid and partitioning
  int ndim = grid.GetNDim();
  long *dim = NULL, *dim_start = NULL, *dim_len = NULL, *lower = NULL;
  long *dim_end = NULL, *idx = NULL;
  New_1D_Array<long, int>(&dim, ndim);
  New_1D_Array<long, int>(&dim_start, ndim);
  New_1D_Array<long, int>(&dim_len, ndim);
  New_1D_Array<long, int>(&lower, ndim);
  New_1D_Array<long, int>(&dim_end, ndim);
  New_1D_Array<long, int>(&idx, ndim);
  grid.GetDim(dim);
  grid.GetPartitioning(dim_start, dim_len);

  // Create a multi-index
  int i;
  long total_len = 1;
  for (i = 0; i < ndim; i++) {
    dim_end[i] = dim_start[i] + dim_len[i];
    lower[i] = 0;
    idx[i] = dim_start[i];
    total_len *= dim_len[i];
  }
  MultiIdx<long> midx(ndim, dim_start, dim_end, lower, dim, idx);

  // Certain constants
  int cp_len = (int)(dim_len[ndim-1]);  // copy a fiber each time
  long cp_cnt = (total_len/cp_len);     // copy count
  size_t sd = sizeof(double);           // size of a double
  double *ptr_data;                     // pointer to data
  long offset, offset_new, offset_delta;

  // Move file pointer to the starting location
  offset = midx.GetOffset();
  MPI_File_seek(fh, offset*sd, MPI_SEEK_SET);

  // Read in fiber by fiber
  offset_delta = 0;
  ptr_data = data;
  long j;
  for (j = 0; j < cp_cnt; j++) {

    // update file pointer
    MPI_File_seek(fh, offset_delta*sd, MPI_SEEK_CUR);

    // read
    MPI_Status status;
    MPI_File_read(fh, ptr_data, cp_len, MPI_DOUBLE, &status);

    // check the success of read
    int count;
    MPI_Get_count(&status, MPI_DOUBLE, &count);
    if (count != cp_len) {
      throw std::string("Error: Fail to load data from ")
          + std::string(filename) + std::string("!\n");
    }

    // find location of the next fiber
    ptr_data += cp_len;
    for (i = 0; i < cp_len; i++) {
      midx.Next(idx);
    }
    offset_new = midx.GetOffset();
    offset_delta = offset_new - offset - cp_len;
    offset = offset_new;
  }

  // Close file
  MPI_File_close(&fh);

  // Clean up
  Delete_1D_Array<long>(&dim);
  Delete_1D_Array<long>(&dim_start);
  Delete_1D_Array<long>(&dim_len);
  Delete_1D_Array<long>(&lower);
  Delete_1D_Array<long>(&dim_end);
  Delete_1D_Array<long>(&idx);
}

//-------------------------------------------------------------------------
void PrintTimeStat(double t1, double t2, const char *header, MPI_Comm &comm) {
  double t_min, t_max, t_mean;
  double t = t2 - t1;
  MPI_Reduce(&t, &t_min, 1, MPI_DOUBLE, MPI_MIN, 0, comm);
  MPI_Reduce(&t, &t_max, 1, MPI_DOUBLE, MPI_MAX, 0, comm);
  MPI_Reduce(&t, &t_mean, 1, MPI_DOUBLE, MPI_SUM, 0, comm);

  int rank, nproc;
  MPI_Comm_rank(comm, &rank);
  MPI_Comm_size(comm, &nproc);
  t_mean /= nproc;

  if (rank == 0) {
    printf("%s mean %gs, max %gs, min %gs\n", header, t_mean, t_max, t_min);
  }
}

//-------------------------------------------------------------------------
void PrintSolSummary(int len_theta, Vec estim,
                     double **fisher, double **var_mle,
                     double **sandwich, double **var_mle_saa,
                     MPI_Comm &comm) {
  int rank;
  MPI_Comm_rank(comm, &rank);
  if (rank != 0) {
    return;
  }

  int i, j;
  const PetscScalar *est;
  VecGetArrayRead(estim, &est);
  printf("Estimates: ");
  for (i = 0; i < len_theta; i++) {
    printf("%g ", est[i]);
  }
  printf("\n");
  VecRestoreArrayRead(estim, &est);

  printf("Fisher matrix: ");
  for (i = 0; i < len_theta; i++) {
    for (j = 0; j < len_theta; j++) {
      printf("%g ", fisher[i][j]);
    }
    printf("; ");
  }
  printf("\n");

  printf("Std of MLE: ");
  for (i = 0; i < len_theta; i++) {
    printf("%g ", sqrt(var_mle[i][i]));
  }
  printf("\n");

  printf("Sandwich matrix: ");
  for (i = 0; i < len_theta; i++) {
    for (j = 0; j < len_theta; j++) {
      printf("%g ", sandwich[i][j]);
    }
    printf("; ");
  }
  printf("\n");

  printf("Std of MLE with SAA: ");
  for (i = 0; i < len_theta; i++) {
    printf("%g ", sqrt(var_mle_saa[i][i]));
  }
  printf("\n");
}

//-------------------------------------------------------------------------
MLEParam::MLEParam(MPI_Comm &_Comm) {
  Comm = _Comm;
  NDim = NDIM;
  Dim[0] = 20;
  Dim[1] = 20;
  Dim[2] = 20;
  GridType = FIXED_DOMAIN;
  Range[0] = 100.0;
  Range[1] = 100.0;
  Range[2] = 100.0;
  CalNProc();
  Nu = 1.0;
  LenTheta = NDim;
  Theta0[0] = 9.0;
  Theta0[1] = 14.0;
  Theta0[2] = 8.0;
  NumRandVec = 50;
  CalNumLapFilter();
  CG_RTol = 1e-10;
  CG_MaxIt = 200;
  bElimAllReduce = true;
  bNeedPartials[0] = true;
  bNeedPartials[1] = true;
  bNeedPartials[2] = true;
  DataFile[0] = '\0';
}

//-------------------------------------------------------------------------
void MLEParam::CalNProc(void) {
  int NP, i;
  MPI_Comm_size(Comm, &NP);
  NProc[0] = NP;
  for (i = 1; i < NDim; i++) {
    NProc[i] = 1;
  }
}

//-------------------------------------------------------------------------
void MLEParam::CalNumLapFilter(void) {
  NumLapFilter = (int)(round((2.0*Nu+NDim)/4.0));
}

//-------------------------------------------------------------------------
PetscErrorCode MLEParam::GetFromOptions(void) {
  int MaxLen;
  PetscErrorCode ierr;

  MaxLen = NDim;
  int dim[NDim];
  ierr = PetscOptionsGetIntArray(PETSC_NULL, "-Dim", dim, &MaxLen,
                                 PETSC_NULL); CHKERRQ(ierr);
  if (MaxLen == NDim) {
    for (int i = 0; i < NDim; i++) {
      Dim[i] = dim[i];
    }
  }

  MaxLen = NDim;
  ierr = PetscOptionsGetRealArray(PETSC_NULL, "-Range", Range, &MaxLen,
                                  PETSC_NULL); CHKERRQ(ierr);

  MaxLen = NDim;
  ierr = PetscOptionsGetIntArray(PETSC_NULL, "-NProc", NProc, &MaxLen,
                                 PETSC_NULL); CHKERRQ(ierr);

  ierr = PetscOptionsGetReal(PETSC_NULL, "-Nu", &Nu,
                             PETSC_NULL); CHKERRQ(ierr);
  CalNumLapFilter();

  MaxLen = NDim;
  ierr = PetscOptionsGetRealArray(PETSC_NULL, "-GuessTheta", Theta0, &MaxLen,
                                  PETSC_NULL); CHKERRQ(ierr);

  ierr = PetscOptionsGetInt(PETSC_NULL, "-NumRandVec", &NumRandVec,
                            PETSC_NULL); CHKERRQ(ierr);

  ierr = PetscOptionsGetInt(PETSC_NULL, "-NumLapFilter", &NumLapFilter,
                            PETSC_NULL); CHKERRQ(ierr);

  ierr = PetscOptionsGetReal(PETSC_NULL, "-CG_RTol", &CG_RTol,
                             PETSC_NULL); CHKERRQ(ierr);

  ierr = PetscOptionsGetInt(PETSC_NULL, "-CG_MaxIt", &CG_MaxIt,
                            PETSC_NULL); CHKERRQ(ierr);

  int iElimAllReduce = 1;
  ierr = PetscOptionsGetInt(PETSC_NULL, "-ElimAllReduce", &iElimAllReduce,
                            PETSC_NULL); CHKERRQ(ierr);
  if (iElimAllReduce != 1) {
    if (iElimAllReduce == 0) {
      bElimAllReduce = false;
    }
    else {
      bElimAllReduce = true;
    }
  }

  MaxLen = MAX_FLEN;
  ierr = PetscOptionsGetString(PETSC_NULL, "-DataFile", DataFile, MaxLen,
                               PETSC_NULL); CHKERRQ(ierr);
  if (strlen(DataFile) == 0) {
    throw std::string("Error: No data file specified!\n");
  }

  return 0;
}

//-------------------------------------------------------------------------
void MLEParam::PrintParams(void) {
  int Rank;
  MPI_Comm_rank(Comm, &Rank);
  if (Rank != 0) {
    return;
  }

  int i;
  std::cout << "Dim: ";
  for (i = 0; i < NDim; i++) {
    std::cout << Dim[i] << " ";
  }
  std::cout << std::endl;

  std::cout << "Range: ";
  for (i = 0; i < NDim; i++) {
    std::cout << Range[i] << " ";
  }
  std::cout << std::endl;

  std::cout << "NProc: ";
  for (i = 0; i < NDim; i++) {
    std::cout << NProc[i] << " ";
  }
  std::cout << std::endl;

  std::cout << "Nu: " << Nu << std::endl;

  std::cout << "Theta0: ";
  for (i = 0; i < NDim; i++) {
    std::cout << Theta0[i] << " ";
  }
  std::cout << std::endl;

  std::cout << "NumRandVec: " << NumRandVec << std::endl;

  std::cout << "NumLapFilter: " << NumLapFilter << std::endl;

  std::cout << "CG_RTol: " << CG_RTol << std::endl;

  std::cout << "CG_MaxIt: " << CG_MaxIt << std::endl;

  std::cout << "bElimAllReduce: ";
  if (bElimAllReduce) {
    std::cout << "true" << std::endl;
  }
  else {
    std::cout << "false" << std::endl;
  }

  std::cout << "DataFile: " << std::string(DataFile) << std::endl;
}

//-------------------------------------------------------------------------
ScoreCtx::ScoreCtx(RegularGrid *_Grid, Matern *_Kernel,
                   RegularGridCovarMat<Matern> *_CovarMat,
                   ApproxScoreFunc<RegularGridData,
                                   RegularGridCovarMat<Matern> > *_S,
                   int _LenTheta, int _NumLapFilter, bool *_bNeedPartials,
                   double *_F, double **_J) {
  Grid = _Grid;
  Kernel = _Kernel;
  CovarMat = _CovarMat;
  S = _S;
  LenTheta = _LenTheta;
  NumLapFilter = _NumLapFilter;
  bNeedPartials = _bNeedPartials;
  F = _F;
  J = _J;
}

// Author: J. Chen
// $Date: 2013/03/28 15:50:43$
