// This example shows how to compute the score functions and the
// Jacobian. The score functions are approximated by using sample
// average approximation (SAA). This example uses the Matern kernel on
// a regular grid, possibly with filtering. The Matern kernel is
// parameterized by scaling factors, one for each dimension. In the
// evaluation of the score functions, the linear solves use a Toeplitz
// solver with circulant preconditioning.

#include <mpi.h>
#include <math.h>
#include <iostream>
#include <ScalaGAUSS.hpp>

void PrintFJ(double *F, double **J, int LenTheta, MPI_Comm Comm);

int main(int argc, char *argv[]) {

  MPI_Init(&argc, &argv);
  MPI_Comm Comm = MPI_COMM_WORLD;

  //-------------------------------------------------------------------------
  // Set up parameters

  // Grid
  int NDim = 3;
  long Dim[] = {10, 10, 10};
  GRID_TYPE GridType = FIXED_DOMAIN;
  double Range[] = {100.0, 100.0, 100.0};

  // Partitioning
  int NProc[] = {2, 1, 1};

  // Kernel Param
  double Nu = 1.0;
  int LenTheta = NDim;
  double Theta[] = {10.0, 13.0, 7.0};

  // Linear solver
  int NumLapFilter = (int)(round((2.0*Nu+NDim)/4.0));
  double CG_RTol = 1e-10;
  int CG_MaxIt = 200;
  bool bElimAllReduce = true;

  // SAA
  int NumRandVec = 50;

  // Nonlinear solver
  bool bNeedPartials[] = {true, true, true};

  try {

  //-------------------------------------------------------------------------
  // Load/Generate GP sample

  // Grid
  RegularGrid Grid(Comm);
  Grid.SetGrid(NDim, Dim, GridType, Range, NProc, bElimAllReduce);

  // Get partitioning
  long DimStart[3], DimLen[3];
  long MyDataLen = Grid.GetPartitioning(DimStart, DimLen);

  // Generate some bogus data
  // This only shows how to use MyDataLen
  // The actual GP sample should be loaded from somewhere else
  double *MyData = NULL;
  New_1D_Array<double, long>(&MyData, MyDataLen);
  Set_Array<double, long>(MyData, 1.0, MyDataLen);
  Delete_1D_Array<double>(&MyData);

  //-------------------------------------------------------------------------
  // Setup score function

  // GPSample
  RegularGridData GPSample(Comm);
  GPSample.SetGrid(Grid);
  GPSample.SetSymBernoulli(); // bogus data
  GPSample.ApplyLapFilter(NumLapFilter);

  // Kernel
  Matern Kernel;
  Kernel.Setup(NDim, Nu, LenTheta, Theta);

  // Toeplitz solver param
  ToepSolverParam mToepSolverParam(CG_RTol, CG_MaxIt);

  // CovarMat
  RegularGridCovarMat<Matern> CovarMat(Comm);
  CovarMat.SetGrid(Grid);
  CovarMat.SetToepSolverParam(mToepSolverParam);
  CovarMat.UseKernel(Kernel, bNeedPartials);
  CovarMat.ApplyLapFilter(NumLapFilter);

  // ScoreFunc
  ApproxScoreFunc<RegularGridData, RegularGridCovarMat<Matern> > S;
  S.SetGPSample(&GPSample);
  S.SetCovarMat(&CovarMat);
  S.SetNumRandVec(NumRandVec);

  //-------------------------------------------------------------------------
  // Compute scores

  // F and J
  double *F = NULL;
  New_1D_Array<double, int>(&F, LenTheta);
  double **J = NULL;
  New_2D_Array<double, int, int>(&J, LenTheta, LenTheta);
  S.GetScores(F, false);
  PrintFJ(F, J, LenTheta, Comm);

  // When Theta changes, redo the following
  Theta[0] += 1.0;
  Kernel.SetTheta(Theta, LenTheta);
  CovarMat.SetGrid(Grid);
  CovarMat.UseKernel(Kernel, bNeedPartials);
  CovarMat.ApplyLapFilter(NumLapFilter);
  S.SetCovarMat(&CovarMat);
  S.GetScores(F, true);
  S.GetJacobian(J);
  PrintFJ(F, J, LenTheta, Comm);

  // Clean up
  Delete_1D_Array<double>(&F);
  Delete_2D_Array<double, int>(&J, LenTheta);

  }
  catch (std::string Msg) {
    ExceptionHandler(Msg, Comm);
  }

  MPI_Finalize();
  return 0;

}

//-------------------------------------------------------------------------
void PrintFJ(double *F, double **J, int LenTheta, MPI_Comm Comm) {
  int Rank;
  MPI_Comm_rank(Comm, &Rank);
  if (Rank != 0) {
    return;
  }

  int i, j;
  std::cout << "F: ";
  for (i = 0; i < LenTheta; i++) {
    std::cout << F[i] << ", ";
  }
  std::cout << "\nJ:\n";
  for (i = 0; i < LenTheta; i++) {
    for (j = 0; j < LenTheta; j++) {
      std::cout << J[i][j] << ", ";
    }
    std::cout << std::endl;
  }
  std::cout << std::endl;
}

// Author: J. Chen
// $Date: 2013/03/28 15:50:43$
