// This example shows the use of multilevel Toeplitz linear
// solver. The matrix is generated by using the Matern kernel on a
// regular grid.

#include <mpi.h>
#include <math.h>
#include <time.h>
#include <iostream>
#include <ScalaGAUSS.hpp>

void PrintRes(double Res, MPI_Comm Comm);
void PrintTime(double T1, double T2, MPI_Comm Comm);

int main(int argc, char *argv[]) {

  MPI_Init(&argc, &argv);
  MPI_Comm Comm = MPI_COMM_WORLD;

  //-------------------------------------------------------------------------
  // Set up parameters

  // Grid
  int NDim = 3;
  long Dim[] = {41, 36, 27};
  GRID_TYPE GridType = FIXED_DOMAIN;
  double Range[] = {100.0, 100.0, 100.0};

  // Partitioning
  int NProc[] = {2, 2, 1};

  // Kernel Param
  double Nu = 1.0;
  int LenTheta = NDim;
  double Theta[] = {1.3, 1.5, 1.2};

  // Linear solver
  int NumLapFilter = (int)(round((2.0*Nu+NDim)/4.0));
  double CG_RTol = 1e-9;
  int CG_MaxIt = 200;
  bool bElimAllReduce = true;
  bool bNeedPartials[] = {true, false, false};

  try {

  //-------------------------------------------------------------------------
  // Linear solve

  // Grid
  RegularGrid Grid(Comm);
  Grid.SetGrid(NDim, Dim, GridType, Range, NProc, bElimAllReduce);

  // RHS
  RegularGridData RHS(Comm);
  RHS.SetGrid(Grid);
  RHS.SetData(1.0);
  RHS.ApplyLapFilter(NumLapFilter);
  RHS.SetData(1.0);

  // Kernel
  Matern Kernel;
  Kernel.Setup(NDim, Nu, LenTheta, Theta);

  // Solver param
  ToepSolverParam mToepSolverParam(CG_RTol, CG_MaxIt);

  // Matrix
  RegularGridCovarMat<Matern> KernelMatrix(Comm);
  KernelMatrix.SetGrid(Grid);
  KernelMatrix.SetToepSolverParam(mToepSolverParam);
  KernelMatrix.UseKernel(Kernel, bNeedPartials);
  KernelMatrix.ApplyLapFilter(NumLapFilter);

  // Initial guess
  RegularGridData X(RHS);
  X.SetData(0.0);

  // Linear solve
  double T1, T2;
  T1 = MPI_Wtime();
  KernelMatrix.KSolve(RHS, X);
  T2 = MPI_Wtime();
  PrintTime(T1, T2, Comm);

  // Check result
  RegularGridData Y(RHS);
  KernelMatrix.KMult(X, Y);
  double Res = RHS.RelDiff(Y);
  PrintRes(Res, Comm);

  }
  catch (std::string Msg) {
    ExceptionHandler(Msg, Comm);
  }

  MPI_Finalize();
  return 0;

}

//-------------------------------------------------------------------------
void PrintRes(double Res, MPI_Comm Comm) {
  int Rank;
  MPI_Comm_rank(Comm, &Rank);
  if (Rank == 0) {
    std::cout << "Relative residual norm: " << Res << std::endl << std::endl;
  }
}

//-------------------------------------------------------------------------
void PrintTime(double T1, double T2, MPI_Comm Comm) {
  double T_min, T_max, T_mean;
  double T = T2 - T1;
  MPI_Reduce(&T, &T_min, 1, MPI_DOUBLE, MPI_MIN, 0, Comm);
  MPI_Reduce(&T, &T_max, 1, MPI_DOUBLE, MPI_MAX, 0, Comm);
  MPI_Reduce(&T, &T_mean, 1, MPI_DOUBLE, MPI_SUM, 0, Comm);

  int Rank, NP;
  MPI_Comm_rank(Comm, &Rank);
  MPI_Comm_size(Comm, &NP);
  T_mean /= NP;

  if (Rank == 0) {
    std::cout << "Time: mean " << T_mean << "s, max " << T_max << "s, min "
              << T_min << "s" << std::endl;
  }
}

// Author: J. Chen
// $Date: 2013/03/28 15:50:43$
