#include <math.h>
#include <time.h>
#include "Data.hpp"
#include "Utility.hpp"

//--------------------------------------------------------------------------
Data::Data() {
  Init();
}

//--------------------------------------------------------------------------
void Data::Init(void) {
  Len = 0;
  Val = NULL;
}

//--------------------------------------------------------------------------
Data::Data(MPI_Comm &comm) {
  Init();
  Comm = comm;
}

//--------------------------------------------------------------------------
Data::Data(const Data &D) {
  Val = NULL;
  DeepCopy(D);
}

//--------------------------------------------------------------------------
void Data::DeepCopy(const Data &D) {
  Len = D.Len;
  New_1D_Array<double, long>(&Val, Len);
  Set_Array<double, long>(Val, D.Val, Len);
  Comm = D.Comm;
}

//--------------------------------------------------------------------------
Data& Data::operator= (const Data &D) {
  if (this == &D) {
    return *this;
  }
  DeAllocMem();
  DeepCopy(D);
  return *this;
}

//--------------------------------------------------------------------------
void Data::DeAllocMem(void) {
  Delete_1D_Array<double>(&Val);
}

//--------------------------------------------------------------------------
Data::~Data() {
  DeAllocMem();
}

//--------------------------------------------------------------------------
void Data::SetLen(long len) {
  if (Len == len) {
    return;
  }
  Delete_1D_Array<double>(&Val);
  Len = len;
  New_1D_Array<double, long>(&Val, Len);
}

//--------------------------------------------------------------------------
void Data::SetVal(double val) {
  Set_Array<double, long>(Val, val, Len);
}

//--------------------------------------------------------------------------
void Data::SetVal(double val, long len) {
  Delete_1D_Array<double>(&Val);
  Len = len;
  New_1D_Array<double, long>(&Val, Len);
  SetVal(val);
}

//--------------------------------------------------------------------------
void Data::SetVal(double *val) {
  Set_Array<double, long>(Val, val, Len);
}

//--------------------------------------------------------------------------
void Data::SetVal(double *val, long len) {
  Delete_1D_Array<double>(&Val);
  Len = len;
  New_1D_Array<double, long>(&Val, Len);
  SetVal(val);
}

//--------------------------------------------------------------------------
void Data::SetSymBernoulli(void) {
  int rank;
  MPI_Comm_rank(Comm, &rank);
  unsigned int seed = time(NULL) + (unsigned int)rank;
  SetSymBernoulli(seed);
}

//--------------------------------------------------------------------------                                                                   
void Data::SetSymBernoulli(unsigned int seed) {
  srandom(seed);
  long i;
  for (i = 0; i < Len; i++) {
    if (random() <= RAND_MAX/2) {
      Val[i] = -1.0;
    }
    else {
      Val[i] = +1.0;
    }
  }
}

//--------------------------------------------------------------------------
void Data::SetSymBernoulli(long len) {
  Delete_1D_Array<double>(&Val);
  Len = len;
  New_1D_Array<double, long>(&Val, Len);
  SetSymBernoulli();
}

//--------------------------------------------------------------------------                                                                   
void Data::SetSymBernoulli(long len, unsigned int seed) {
  Delete_1D_Array<double>(&Val);
  Len = len;
  New_1D_Array<double, long>(&Val, Len);
  SetSymBernoulli(seed);
}

//--------------------------------------------------------------------------
long Data::GetLen(void) {
  return Len;
}

//--------------------------------------------------------------------------
double* Data::GetValPtr(void) {
  return Val;
}

//--------------------------------------------------------------------------
double Data::InProd(Data &B) {
  if (Len != B.GetLen()) {
    throw std::string("Data::InProd: Vectors have different lengths!\n");
  }

  double inprod_local = 0.0, inprod;
  double *Bptr = B.GetValPtr();
  long i;
  for (i = 0; i < Len; i++) {
    inprod_local += Val[i] * Bptr[i];
  }
  MPI_Allreduce(&inprod_local, &inprod, 1, MPI_DOUBLE, MPI_SUM, Comm);

  return inprod;
}

//--------------------------------------------------------------------------
double Data::Norm2(void) {
  return sqrt(InProd(*this));
}

//--------------------------------------------------------------------------
double Data::RelDiff(Data &B) {
  if (Len != B.GetLen()) {
    throw std::string("Data::RelDiff: Vectors have different lengths!\n");
  }

  double diff_and_norm_local[2] = {0.0, 0.0}, diff_and_norm[2];
  double *Bptr = B.GetValPtr();
  long i;
  for (i = 0; i < Len; i++) {
    diff_and_norm_local[0] += square(Val[i]-Bptr[i]);
    diff_and_norm_local[1] += square(Val[i]);
  }
  MPI_Allreduce(&diff_and_norm_local, &diff_and_norm, 2,
                MPI_DOUBLE, MPI_SUM, Comm);

  return sqrt(diff_and_norm[0]/diff_and_norm[1]);
}

// Author: J. Chen
// $Date: 2013/03/28 15:50:43$
