#include <math.h>
#include "Kernel.hpp"
#include "Utility.hpp"

//--------------------------------------------------------------------------
Matern::Matern() {
  D = 0;
  Nu = 0.0;
  LenTheta = 0;
  Theta = NULL;
  k0 = 0.0;
  twonu = 0.0;
  sqrt2nu = 0.0;
}

//--------------------------------------------------------------------------
Matern::Matern(const Matern &M) {
  Theta = NULL;
  DeepCopy(M);
}

//--------------------------------------------------------------------------
void Matern::DeepCopy(const Matern &M) {
  D = M.D;
  Nu = M.Nu;
  LenTheta = M.LenTheta;
  New_1D_Array<double, int>(&Theta, LenTheta);
  Set_Array<double, int>(Theta, M.Theta, LenTheta);
  k0 = M.k0;
  twonu = M.twonu;
  sqrt2nu = M.sqrt2nu;
}

//--------------------------------------------------------------------------
Matern& Matern::operator= (const Matern &M) {
  if (this == &M) {
    return *this;
  }
  DeAllocMem();
  DeepCopy(M);
  return *this;
}

//--------------------------------------------------------------------------
void Matern::DeAllocMem(void) {
  Delete_1D_Array<double>(&Theta);
}

//--------------------------------------------------------------------------
Matern::~Matern() {
  DeAllocMem();
}

//--------------------------------------------------------------------------
#define SETUP_BODY {                                \
    Delete_1D_Array<double>(&Theta);                \
    D = d;                                          \
    Nu = nu;                                        \
    LenTheta = len_theta;                           \
    New_1D_Array<double, int>(&Theta, LenTheta);    \
    Set_Array<double, int>(Theta, theta, LenTheta); \
    ComputeConstants();                             \
  }

//--------------------------------------------------------------------------
void Matern::Setup(int d, double nu, int len_theta, double theta) {
  SETUP_BODY;
}

//--------------------------------------------------------------------------
void Matern::Setup(int d, double nu, int len_theta, double *theta) {
  SETUP_BODY;
}

//--------------------------------------------------------------------------
#define SET_THETA_BODY {                                                \
    if (LenTheta != len_theta) {                                        \
      throw std::string("Matern::SetTheta: Length of theta has changed!\n"); \
    }                                                                   \
    Set_Array<double, int>(Theta, theta, LenTheta);                     \
    ComputeConstants();                                                 \
  }

//--------------------------------------------------------------------------
void Matern::SetTheta(double theta, int len_theta) {
  SET_THETA_BODY;
}

//--------------------------------------------------------------------------
void Matern::SetTheta(double *theta, int len_theta) {
  SET_THETA_BODY;
}

//--------------------------------------------------------------------------
int Matern::GetLenTheta(void) {
  return LenTheta;
}

//--------------------------------------------------------------------------
bool Matern::Compare(Matern &M) {
  if (D != M.D) { return false; }
  if (Nu != M.Nu) { return false; }
  if (LenTheta != M.LenTheta) { return false; }
  int i;
  for (i = 0; i < LenTheta; i++) {
    if (Theta[i] != M.Theta[i]) { return false; }
  }
  return true;
}

//--------------------------------------------------------------------------
bool Matern::IsSameTheta(double *theta) {
  int i;
  for (i = 0; i < LenTheta; i++) {
    if (Theta[i] != theta[i]) {
      return false;
    }
  }
  return true;
}

//--------------------------------------------------------------------------
double Matern::GetValues(double *x, double *p, double **pp, bool x_is_0) {
  int i, j;

  // If x is zero, everything is simplified
  if (x_is_0) {
    if (p) {
      Set_Array<double, int>(p, 0.0, LenTheta);
      if (pp) {
        for (i = 0; i < LenTheta; i++) {
          Set_Array<double, int>(pp[i], 0.0, i+1);
        }
      }
    }
    return 1.0;
  }

  // If x is not zero, compute r
  double *sq_scaled_xi = NULL;
  New_1D_Array<double, int>(&sq_scaled_xi, D);
  double r = 0.0;
  for (i = 0; i < D; i++) {
    sq_scaled_xi[i] = square( x[i] / Theta[i] );
    r += sq_scaled_xi[i];
  }
  r = sqrt2nu * sqrt(r);

  // Just in case x is zero but the user does not set x_is_0
  if (r == 0.0) {
    if (p) {
      Set_Array<double, int>(p, 0.0, LenTheta);
      if (pp) {
        for (i = 0; i < LenTheta; i++) {
          Set_Array<double, int>(pp[i], 0.0, i+1);
        }
      }
    }
    Delete_1D_Array<double>(&sq_scaled_xi);
    return 1.0;
  }

  // Compute r^nu * besselk(nu,r), r^(nu-1) * besselk(nu-1,r)
  // and r^(nu-2) * besselk(|nu-1|-1,r)
  double bnur, bnur1, bnur2;
  double rpownu, rpownu1, rpownu2;
  double rpownu_bnur, rpownu_bnur1, rpownu_bnur2 = 0;
  rpownu = pow(r, Nu);
  if (!p) {
    besselk2(Nu, r, &bnur);
  }
  else {
    rpownu1 = rpownu / r;
    if (!pp) {
      besselk2(Nu, r, &bnur, &bnur1);
    }
    else {
      rpownu2 = rpownu1 / r;
      besselk2(Nu, r, &bnur, &bnur1, &bnur2);
      rpownu_bnur2 = rpownu2 * bnur2;
    }
    rpownu_bnur1 = rpownu1 * bnur1;
  }
  rpownu_bnur = rpownu * bnur;

  // Assemble
  double k = rpownu_bnur / k0;
  double *sq_scaled_xi_theta = NULL;
  if (p) {
    New_1D_Array<double, int>(&sq_scaled_xi_theta, D);
    for (i = 0; i < D; i++) {
      sq_scaled_xi_theta[i] = sq_scaled_xi[i] / Theta[i];
    }

    double twonu_rpownu_bnur1_k0 = twonu * rpownu_bnur1 / k0;
    for (i = 0; i < LenTheta; i++) {
      p[i] = twonu_rpownu_bnur1_k0 * sq_scaled_xi_theta[i];
    }

    if (pp) {
      double sq_twonu = square(twonu);
      double sq_twonu_rpownu_bnur2_k0 = sq_twonu * rpownu_bnur2 / k0;
      for (i = 0; i < LenTheta; i++) {
        for (j = 0; j < i; j++) {
          pp[i][j] = sq_twonu_rpownu_bnur2_k0
            * (sq_scaled_xi_theta[i] * sq_scaled_xi_theta[j]);
        }
        for (j = i; j <= i; j++) {
          pp[i][j] = -3.0 * p[j] / Theta[j]
            + sq_twonu_rpownu_bnur2_k0 * square( sq_scaled_xi_theta[j] );
        }
      }
    }
  }

  // Clean up
  Delete_1D_Array<double>(&sq_scaled_xi);
  Delete_1D_Array<double>(&sq_scaled_xi_theta);

  return k;
}

//--------------------------------------------------------------------------
void Matern::ComputeConstants(void) {
  k0 = pow(2.0, Nu-1) * tgamma(Nu);
  twonu = 2.0 * Nu;
  sqrt2nu = sqrt(twonu);
}

// Author: J. Chen
// $Date: 2013/03/28 15:50:43$
