#ifndef _KERNEL_
#define _KERNEL_
// Defines several covariance kernels.

//--------------------------------------------------------------------------
// Matern kernel:
//
//           r^nu * besselk(nu,r)
//   k(x) = ----------------------
//           2^(nu-1) * gamma(nu)
// 
// where r is the elliptical radius
// 
//                        [     [    x_i^2    ]]
//   r = sqrt[2*nu] * sqrt[ sum [ ----------- ]]
//                        [  i  [  theta_i^2  ]]
//
// and nu > 0. First partial (w.r.t. theta_i):
//
//    2*nu * r^(nu-1) * besselk(nu-1,r)       x_i^2
//   ----------------------------------- * -----------
//           2^(nu-1) * gamma(nu)           theta_i^3
//
// Second partial (w.r.t. theta_i and theta_j):
//
// i != j:
//         (2*nu)^2 * r^(nu-2) * besselk(|nu-1|-1,r)         x_i^2 * x_j^2
//        ------------------------------------------- * -----------------------
//                    2^(nu-1) * gamma(nu)               theta_i^3 * theta_j^3
//
// i == j:
//         -3 * (2*nu) * r^(nu-1) * besselk(nu-1,r)       x_j^2
//        ------------------------------------------ * -----------
//                    2^(nu-1) * gamma(nu)              theta_j^4
//
//            (2*nu)^2 * r^(nu-2) * besselk(|nu-1|-1,r)      x_j^4
//         + ------------------------------------------ * -----------
//                       2^(nu-1) * gamma(nu)              theta_j^6
//--------------------------------------------------------------------------
class Matern {

public:

  // Constructor, destructor
  Matern();
  Matern(const Matern &M);
  Matern& operator= (const Matern &M);
  ~Matern();

  // Set
  void Setup(int d, double nu, int len_theta, double  theta);
  void Setup(int d, double nu, int len_theta, double *theta);
  void SetTheta(double  theta, int len_theta);
  void SetTheta(double *theta, int len_theta);

  // Get
  int GetLenTheta(void);

  // Compare
  bool Compare(Matern &M);
  bool IsSameTheta(double *theta);

  // Major computational component
  double                 // Output: Matern function value
  GetValues(double *x,   // Input:  Vector x
            double *p,   // Output: First order partials, if not NULL
            double **pp, // Output: Second order partials, if p and pp not NULL
                         //         Since symmetric, only output lower tri half
                         //         pp can be allocated using New_2D_Array_Sym()
            bool x_is_0 = false // Input: If x is zero, everything is simplified
            );

private:

  // Needed by constructor and destructor
  void DeepCopy(const Matern &M);
  void DeAllocMem(void);

  // Needed by Setup()
  void ComputeConstants(void);

  int D;                        // Spatial dimension
  double Nu;                    // Smoothness
  int LenTheta;                 // Length of Theta
  double *Theta;                // Theta

  double k0;                    // = pow(2.0, Nu-1) * tgamma(Nu);
  double twonu;                 // = 2.0 * Nu;
  double sqrt2nu;               // = sqrt(twonu);

};

#endif

// Author: J. Chen
// $Date: 2013/03/28 15:50:43$
