#ifndef _MULTI_IDX_
#define _MULTI_IDX_
// Defines a multi-index class. The main purpose is to get the next
// multi-index given a bound along each dimension. Once the number of
// dimensions and the bounds are set, they cannot be changed.
//
// In addition, the range of the multi-index is contained in a
// box. This allows transforming a multi-index to a linear index. Here
// we assume that the last dimension varies the fastest (row-major
// order in C/C++, as oppose to column-major order in Matlab).
//
// BoxLowerBound <= IdxLowerBound <= CurrentId < IdxUpperBound <= BoxUpperBound

#include "Utility.hpp"

//-------------------------------------------------------------------------
#define PASS_UPPER_LIMIT 0

template<class T>
class MultiIdx {

public:

  // Constructor, destructor
  MultiIdx(int ndim, T  idx_lower_bound, T  idx_upper_bound);
  MultiIdx(int ndim, T *idx_lower_bound, T *idx_upper_bound);
  MultiIdx(int ndim, T  idx_lower_bound, T  idx_upper_bound,
                     T  box_lower_bound, T  box_upper_bound);
  MultiIdx(int ndim, T *idx_lower_bound, T *idx_upper_bound,
                     T *box_lower_bound, T *box_upper_bound);
  MultiIdx(int ndim, T  idx_lower_bound, T  idx_upper_bound, T *idx);
  MultiIdx(int ndim, T *idx_lower_bound, T *idx_upper_bound, T *idx);
  MultiIdx(int ndim, T  idx_lower_bound, T  idx_upper_bound,
                     T  box_lower_bound, T  box_upper_bound, T *idx);
  MultiIdx(int ndim, T *idx_lower_bound, T *idx_upper_bound,
                     T *box_lower_bound, T *box_upper_bound, T *idx);
  MultiIdx(const MultiIdx &MI);
  MultiIdx& operator= (const MultiIdx &MI);
  ~MultiIdx();

  // Set
  void SetIdx(T *idx);

  // Get
  void GetCurrentIdx(T *idx);
  T GetOffset(void);

  // Major computational component
  void Next(T *idx);

private:

  // Needed by constructor and destructor
  void ComputeOffset(void);
  void DeepCopy(const MultiIdx &MI);
  void DeAllocMem(void);

  int NDim;
  T *IdxLowerBound;
  T *IdxUpperBound;
  T *IdxLen;
  T *CurrentIdx;

  T *BoxLowerBound;
  T *BoxUpperBound;
  T *BoxLen;
  T *CummulateLen;
  T Offset;

};

//-------------------------------------------------------------------------
#define MULTIIDX_BODY1 {                                                \
    NDim = ndim;                                                        \
    IdxLowerBound = NULL;                                               \
    New_1D_Array<T, int>(&IdxLowerBound, NDim);                         \
    Set_Array<T, int>(IdxLowerBound, idx_lower_bound, NDim);            \
    IdxUpperBound = NULL;                                               \
    New_1D_Array<T, int>(&IdxUpperBound, NDim);                         \
    Set_Array<T, int>(IdxUpperBound, idx_upper_bound, NDim);            \
    IdxLen = NULL;                                                      \
    New_1D_Array<T, int>(&IdxLen, NDim);                                \
    int i;                                                              \
    for (i = 0; i < NDim; i++) {                                        \
      IdxLen[i] = IdxUpperBound[i] - IdxLowerBound[i];                  \
      if (IdxLen[i] < 0) {                                              \
        throw std::string("MultiIdx::MultiIdx: Idx lower > upper!\n");  \
      }                                                                 \
    }                                                                   \
    BoxLowerBound = NULL;                                               \
    New_1D_Array<T, int>(&BoxLowerBound, NDim);                         \
    BoxUpperBound = NULL;                                               \
    New_1D_Array<T, int>(&BoxUpperBound, NDim);                         \
    BoxLen = NULL;                                                      \
    New_1D_Array<T, int>(&BoxLen, NDim);                                \
    CummulateLen = NULL;                                                \
    New_1D_Array<T, int>(&CummulateLen, NDim);                          \
  }

//-------------------------------------------------------------------------
#define MULTIIDX_BODY2a {                                               \
    Set_Array<T, int>(BoxLowerBound, IdxLowerBound, NDim);              \
    Set_Array<T, int>(BoxUpperBound, IdxUpperBound, NDim);              \
  }

//-------------------------------------------------------------------------
#define MULTIIDX_BODY2b {                                               \
    Set_Array<T, int>(BoxLowerBound, box_lower_bound, NDim);            \
    Set_Array<T, int>(BoxUpperBound, box_upper_bound, NDim);            \
    int i;                                                              \
    for (i = 0; i < NDim; i++) {                                        \
      if (IdxLowerBound[i] < BoxLowerBound[i]) {                        \
        throw std::string("MultiIdx::MultiIdx: Idx lower < box lower!\n"); \
      }                                                                 \
      if (IdxUpperBound[i] > BoxUpperBound[i]) {                        \
        throw std::string("MultiIdx::MultiIdx: Idx upper > box upper!\n"); \
      }                                                                 \
    }                                                                   \
  }

//-------------------------------------------------------------------------
#define MULTIIDX_BODY3 {                                                \
    int i;                                                              \
    CummulateLen[NDim-1] = 1;                                           \
    for (i = NDim-1; i >= 0; i--) {                                     \
      BoxLen[i] = BoxUpperBound[i] - BoxLowerBound[i];                  \
      if (i != 0) {                                                     \
        CummulateLen[i-1] = BoxLen[i] * CummulateLen[i];                \
      }                                                                 \
    }                                                                   \
  }

//-------------------------------------------------------------------------
template<class T>
MultiIdx<T>::
MultiIdx(int ndim, T idx_lower_bound, T idx_upper_bound) {
  MULTIIDX_BODY1;
  MULTIIDX_BODY2a;
  MULTIIDX_BODY3;
  CurrentIdx = NULL;
  Offset = -1;
}

//-------------------------------------------------------------------------
template<class T>
MultiIdx<T>::
MultiIdx(int ndim, T *idx_lower_bound, T *idx_upper_bound) {
  MULTIIDX_BODY1;
  MULTIIDX_BODY2a;
  MULTIIDX_BODY3;
  CurrentIdx = NULL;
  Offset = -1;
}

//-------------------------------------------------------------------------
template<class T>
MultiIdx<T>::
MultiIdx(int ndim, T idx_lower_bound, T idx_upper_bound,
                   T box_lower_bound, T box_upper_bound) {
  MULTIIDX_BODY1;
  MULTIIDX_BODY2b;
  MULTIIDX_BODY3;
  CurrentIdx = NULL;
  Offset = -1;
}

//-------------------------------------------------------------------------
template<class T>
MultiIdx<T>::
MultiIdx(int ndim, T *idx_lower_bound, T *idx_upper_bound,
                   T *box_lower_bound, T *box_upper_bound) {
  MULTIIDX_BODY1;
  MULTIIDX_BODY2b;
  MULTIIDX_BODY3;
  CurrentIdx = NULL;
  Offset = -1;
}

//-------------------------------------------------------------------------
template<class T>
MultiIdx<T>::
MultiIdx(int ndim, T idx_lower_bound, T idx_upper_bound, T *idx) {
  MULTIIDX_BODY1;
  MULTIIDX_BODY2a;
  MULTIIDX_BODY3;
  SetIdx(idx);
}

//-------------------------------------------------------------------------
template<class T>
MultiIdx<T>::
MultiIdx(int ndim, T *idx_lower_bound, T *idx_upper_bound, T *idx) {
  MULTIIDX_BODY1;
  MULTIIDX_BODY2a;
  MULTIIDX_BODY3;
  SetIdx(idx);
}

//-------------------------------------------------------------------------
template<class T>
MultiIdx<T>::
MultiIdx(int ndim, T idx_lower_bound, T idx_upper_bound,
                   T box_lower_bound, T box_upper_bound, T *idx) {
  MULTIIDX_BODY1;
  MULTIIDX_BODY2b;
  MULTIIDX_BODY3;
  SetIdx(idx);
}

//-------------------------------------------------------------------------
template<class T>
MultiIdx<T>::
MultiIdx(int ndim, T *idx_lower_bound, T *idx_upper_bound,
                   T *box_lower_bound, T *box_upper_bound, T *idx) {
  MULTIIDX_BODY1;
  MULTIIDX_BODY2b;
  MULTIIDX_BODY3;
  SetIdx(idx);
}

//-------------------------------------------------------------------------
template<class T>
MultiIdx<T>::
MultiIdx(const MultiIdx &MI) {
  IdxLowerBound = NULL;
  IdxUpperBound = NULL;
  IdxLen = NULL;
  CurrentIdx = NULL;
  BoxLowerBound = NULL;
  BoxUpperBound = NULL;
  BoxLen = NULL;
  CummulateLen = NULL;
  DeepCopy(MI);
}

//-------------------------------------------------------------------------
template<class T>
void MultiIdx<T>::
DeepCopy(const MultiIdx &MI) {
  NDim = MI.NDim;
  New_1D_Array<T, int>(&IdxLowerBound, NDim);
  Set_Array<T, int>(IdxLowerBound, MI.IdxLowerBound, NDim);
  New_1D_Array<T, int>(&IdxUpperBound, NDim);
  Set_Array<T, int>(IdxUpperBound, MI.IdxUpperBound, NDim);
  New_1D_Array<T, int>(&IdxLen, NDim);
  Set_Array<T, int>(IdxLen, MI.IdxLen, NDim);
  New_1D_Array<T, int>(&CurrentIdx, NDim);
  Set_Array<T, int>(CurrentIdx, MI.CurrentIdx, NDim);
  New_1D_Array<T, int>(&BoxLowerBound, NDim);
  Set_Array<T, int>(BoxLowerBound, MI.BoxLowerBound, NDim);
  New_1D_Array<T, int>(&BoxUpperBound, NDim);
  Set_Array<T, int>(BoxUpperBound, MI.BoxUpperBound, NDim);
  New_1D_Array<T, int>(&BoxLen, NDim);
  Set_Array<T, int>(BoxLen, MI.BoxLen, NDim);
  New_1D_Array<T, int>(&CummulateLen, NDim);
  Set_Array<T, int>(CummulateLen, MI.CummulateLen, NDim);
  Offset = MI.Offset;
}

//-------------------------------------------------------------------------
template<class T>
MultiIdx<T>& MultiIdx<T>::
operator= (const MultiIdx &MI) {
  if (this == &MI) {
    return *this;
  }
  DeAllocMem();
  DeepCopy(MI);
  return *this;
}

//-------------------------------------------------------------------------
template<class T>
MultiIdx<T>::
~MultiIdx() {
  DeAllocMem();
}

//-------------------------------------------------------------------------
template<class T>
void MultiIdx<T>::
DeAllocMem(void) {
  Delete_1D_Array<T>(&IdxLowerBound);
  Delete_1D_Array<T>(&IdxUpperBound);
  Delete_1D_Array<T>(&IdxLen);
  Delete_1D_Array<T>(&CurrentIdx);
  Delete_1D_Array<T>(&BoxLowerBound);
  Delete_1D_Array<T>(&BoxUpperBound);
  Delete_1D_Array<T>(&BoxLen);
  Delete_1D_Array<T>(&CummulateLen);
}

//-------------------------------------------------------------------------
template<class T>
void MultiIdx<T>::
SetIdx(T *idx) {
  CurrentIdx = NULL;
  New_1D_Array<T, int>(&CurrentIdx, NDim);
  Set_Array<T, int>(CurrentIdx, idx, NDim);
  int i;
  for (i = 0; i < NDim; i++) {
    if (CurrentIdx[i] >= IdxUpperBound[i] || CurrentIdx[i] < IdxLowerBound[i]) {
      throw std::string("MultiIdx::SetIdx: Index not in bound!\n");
    }
  }
  ComputeOffset();
}

//-------------------------------------------------------------------------
template<class T>
void MultiIdx<T>::
ComputeOffset(void) {
  int i;
  Offset = 0;
  for (i = NDim-1; i >= 0; i--) {
    Offset += CummulateLen[i] * (CurrentIdx[i] - BoxLowerBound[i]);
  }
}

//-------------------------------------------------------------------------
template<class T>
void MultiIdx<T>::
GetCurrentIdx(T *idx) {
  Set_Array<T, int>(idx, CurrentIdx, NDim);
}

//-------------------------------------------------------------------------
template<class T>
T MultiIdx<T>::
GetOffset(void) {
  return Offset;
}

//-------------------------------------------------------------------------
template<class T>
void MultiIdx<T>::
Next(T *idx) {
  int i;
  for (i = NDim-1; i >= 0; i--) {
    if (CurrentIdx[i] < IdxUpperBound[i] - 1) {
      CurrentIdx[i]++;
      Offset += CummulateLen[i];
      break;
    }
    else {
      CurrentIdx[i] = IdxLowerBound[i];
      Offset -= (IdxLen[i]-1) * CummulateLen[i];
    }
  }
  Set_Array<T, int>(idx, CurrentIdx, NDim);
}

#endif

// Author: J. Chen
// $Date: 2013/03/28 15:50:43$
