#include "MultiIdx.hpp"
#include "RegularGrid.hpp"

//--------------------------------------------------------------------------
RegularGrid::RegularGrid() {
  Init();
}

//--------------------------------------------------------------------------
void RegularGrid::Init(void) {
  NDim = 0;
  Dim = NULL;
  mType = UNKNOWN_GRID_TYPE;
  mSpec.Range = NULL;
  NProc = NULL;
  DimStart = NULL;
  DimLen = NULL;
  bElimAllReduce = true;
}

//--------------------------------------------------------------------------
RegularGrid::RegularGrid(MPI_Comm &comm) {
  Init();
  Comm = comm;
}

//--------------------------------------------------------------------------
RegularGrid::RegularGrid(const RegularGrid &G) {
  Dim = NULL;
  mSpec.Range = NULL;
  NProc = NULL;
  DimStart = NULL;
  DimLen = NULL;
  DeepCopy(G);
}

//--------------------------------------------------------------------------
void RegularGrid::DeepCopy(const RegularGrid &G) {
  NDim = G.NDim;
  New_1D_Array<long, int>(&Dim, NDim);
  Set_Array<long, int>(Dim, G.Dim, NDim);
  mType = G.mType;
  switch (mType) {
  case FIXED_DOMAIN:
    New_1D_Array<double, int>(&(mSpec.Range), NDim);
    Set_Array<double, int>(mSpec.Range, G.mSpec.Range, NDim);
    break;
  case FIXED_SPACING:
    New_1D_Array<double, int>(&(mSpec.Delta), NDim);
    Set_Array<double, int>(mSpec.Range, G.mSpec.Delta, NDim);
    break;
  case UNKNOWN_GRID_TYPE:
    throw std::string("RegularGrid::DeepCopy: Unknown grid type!\n");
    break;
  }
  New_1D_Array<int, int>(&NProc, NDim);
  Set_Array<int, int>(NProc, G.NProc, NDim);
  New_1D_Array<long, int>(&DimStart, NDim);
  Set_Array<long, int>(DimStart, G.DimStart, NDim);
  New_1D_Array<long, int>(&DimLen, NDim);
  Set_Array<long, int>(DimLen, G.DimLen, NDim);
  Comm = G.Comm;
  mToepSolver = G.mToepSolver;
  bElimAllReduce = G.bElimAllReduce;
}

//--------------------------------------------------------------------------
RegularGrid& RegularGrid::operator= (const RegularGrid &G) {
  if (this == &G) {
    return *this;
  }
  DeAllocMem();
  DeepCopy(G);
  return *this;
}

//--------------------------------------------------------------------------
void RegularGrid::DeAllocMem(void) {
  Delete_1D_Array<long>(&Dim);
  switch (mType) {
  case FIXED_DOMAIN:
    Delete_1D_Array<double>(&(mSpec.Range));
    break;
  case FIXED_SPACING:
    Delete_1D_Array<double>(&(mSpec.Delta));
    break;
  case UNKNOWN_GRID_TYPE:
    break;
  }
  Delete_1D_Array<int>(&NProc);
  Delete_1D_Array<long>(&DimStart);
  Delete_1D_Array<long>(&DimLen);
}

//--------------------------------------------------------------------------
RegularGrid::~RegularGrid() {
  DeAllocMem();
}

//--------------------------------------------------------------------------
#define SET_GRID_BODY {                                                 \
    DeAllocMem();                                                       \
                                                                        \
    NDim = ndim;                                                        \
    New_1D_Array<long, int>(&Dim, NDim);                                \
    Set_Array<long, int>(Dim, dim, NDim);                               \
    mType = type;                                                       \
    switch (mType) {                                                    \
    case FIXED_DOMAIN:                                                  \
      New_1D_Array<double, int>(&(mSpec.Range), NDim);                  \
      Set_Array<double, int>(mSpec.Range, spval, NDim);                 \
      break;                                                            \
    case FIXED_SPACING:                                                 \
      New_1D_Array<double, int>(&(mSpec.Delta), NDim);                  \
      Set_Array<double, int>(mSpec.Delta, spval, NDim);                 \
      break;                                                            \
    case UNKNOWN_GRID_TYPE:                                             \
      break;                                                            \
    }                                                                   \
                                                                        \
    New_1D_Array<int, int>(&NProc, NDim);                               \
    Set_Array<int, int>(NProc, nproc, NDim);                            \
    int total_nproc;                                                    \
    MPI_Comm_size(Comm, &total_nproc);                                  \
    int total_nproc2 = 1;                                               \
    int i;                                                              \
    for (i = 0; i < NDim; i++) {                                        \
      total_nproc2 *= NProc[i];                                         \
    }                                                                   \
    if (total_nproc != total_nproc2) {                                  \
      throw std::string("RegularGrid::SetGrid: Prod(nproc) != #processes!\n"); \
    }                                                                   \
                                                                        \
    bElimAllReduce = elim_allreduce;                                    \
    mToepSolver.SetDimAndOpt(NDim, Dim, NProc, bElimAllReduce, Comm);   \
    New_1D_Array<long, int>(&DimStart, NDim);                           \
    New_1D_Array<long, int>(&DimLen, NDim);                             \
    mToepSolver.GetPartitioning(DimStart, DimLen);                      \
  }

//--------------------------------------------------------------------------
void RegularGrid::SetGrid(int ndim, long dim, GRID_TYPE type, double spval,
                          int *nproc, bool elim_allreduce) {
  SET_GRID_BODY;
}

//--------------------------------------------------------------------------
void RegularGrid::SetGrid(int ndim, long dim, GRID_TYPE type, double *spval,
                          int *nproc, bool elim_allreduce) {
  SET_GRID_BODY;
}

//--------------------------------------------------------------------------
void RegularGrid::SetGrid(int ndim, long *dim, GRID_TYPE type, double spval,
                          int *nproc, bool elim_allreduce) {
  SET_GRID_BODY;
}

//--------------------------------------------------------------------------
void RegularGrid::SetGrid(int ndim, long *dim, GRID_TYPE type, double *spval,
                          int *nproc, bool elim_allreduce) {
  SET_GRID_BODY;
}

//--------------------------------------------------------------------------
int RegularGrid::GetNDim(void) {
  return NDim;
}

//--------------------------------------------------------------------------
void RegularGrid::GetDim(long *dim) {
  Set_Array<long, int>(dim, Dim, NDim);
}

//--------------------------------------------------------------------------
long RegularGrid::GetPartitioning(long *dim_start, long *dim_len) {
  if (dim_start) {
    Set_Array<long, int>(dim_start, DimStart, NDim);
  }
  if (dim_len) {
    Set_Array<long, int>(dim_len, DimLen, NDim);
  }
  long total_len = 1;
  int i;
  for (i = 0; i < NDim; i++) {
    total_len *= DimLen[i];
  }
  return total_len;
}

//--------------------------------------------------------------------------
Stencil::Stencil() {
  NDim = 0;
  Dim = NULL;
  NumLapFilter = 0;
  Coef = NULL;
  Displs = NULL;
  Len = 0;
}

//--------------------------------------------------------------------------
Stencil::Stencil(const Stencil &S) {
  Dim = NULL;
  Coef = NULL;
  Displs = NULL;
  DeepCopy(S);
}

//--------------------------------------------------------------------------
void Stencil::DeepCopy(const Stencil &S) {
  NDim = S.NDim;
  New_1D_Array<long, int>(&Dim, NDim);
  Set_Array<long, int>(Dim, S.Dim, NDim);
  NumLapFilter = S.NumLapFilter;
  Len = S.Len;
  New_1D_Array<double, int>(&Coef, Len);
  Set_Array<double, int>(Coef, S.Coef, Len);
  New_1D_Array<long, int>(&Displs, Len);
  Set_Array<long, int>(Displs, S.Displs, Len);
}

//--------------------------------------------------------------------------
Stencil& Stencil::operator= (const Stencil &S) {
  if (this == &S) {
    return *this;
  }
  DeAllocMem();
  DeepCopy(S);
  return *this;
}

//--------------------------------------------------------------------------
Stencil::~Stencil() {
  DeAllocMem();
}

//--------------------------------------------------------------------------
void Stencil::DeAllocMem(void) {
  Delete_1D_Array<long>(&Dim);
  Delete_1D_Array<double>(&Coef);
  Delete_1D_Array<long>(&Displs);
}

//--------------------------------------------------------------------------
void Stencil::SetFilteringInfo(int ndim, long *dim, int num_filter) {
  DeAllocMem();

  int i, j, k;

  NDim = ndim;
  New_1D_Array<long, int>(&Dim, NDim);
  Set_Array<long, int>(Dim, dim, NDim);
  NumLapFilter = num_filter;

  // The following codes compute Coef, Displs and Len

  // 1. Compute the displacements of all the points to the center of a
  // standard 2nd order stencil:
  //                                 +
  //                                 | +
  //                                 |/
  //                             +---o---+
  //                                /|
  //                               + |
  //                                 +
  // The displacements are computed w.r.t. a cube of size
  // (2*NumLapFilter+1)^NDim.
  int *local_displs = NULL;
  New_1D_Array<int, int>(&local_displs, NDim*2);
  local_displs[2*NDim-2] = -1;
  local_displs[2*NDim-1] = 1;
  int cube_len = 2*NumLapFilter + 1;
  for (i = NDim-2; i >= 0; i--) {
    local_displs[2*i] = -cube_len;
    local_displs[2*i+1] = cube_len;
    cube_len *= (2*NumLapFilter + 1);
  }

  // 2. Compute the displacements of all the points inside the cube,
  // to the center of the cube. This cube contains the full
  // stencil. The displacements are computed w.r.t. the data tank.
  long *tank_len = NULL;
  New_1D_Array<long, int>(&tank_len, NDim);
  tank_len[NDim-1] = 1;
  for (i = NDim-2; i >= 0; i--){
    tank_len[i] = tank_len[i+1] * Dim[i+1];
  }
  int *idx = NULL;
  New_1D_Array<int, int>(&idx, NDim);
  Set_Array<int, int>(idx, -NumLapFilter, NDim);
  MultiIdx<int> midx(NDim, -NumLapFilter, NumLapFilter+1, idx);

  long *grid_displs = NULL;
  New_1D_Array<long, int>(&grid_displs, cube_len);
  Set_Array<long, int>(grid_displs, (long)0, cube_len);
  for (j = 0; j < cube_len; j++) {
    for (i = 0; i < NDim; i++) {
      grid_displs[j] += (tank_len[i] * idx[i]);
    }
    midx.Next(idx);
  }

  // 3. Fill the stencil the zero-th time
  double *sten1 = NULL;
  double *sten2 = NULL;
  New_1D_Array<double, int>(&sten1, cube_len);
  New_1D_Array<double, int>(&sten2, cube_len);
  int center = cube_len/2;
  sten1[center] = 1.0;

  // 4. Fill the stencil a number of times. Always assume that the
  // stencil that ptr_sten1 points to has been filled
  double *ptr_sten1 = sten1;
  double *ptr_sten2 = sten2;
  for (j = 0; j < NumLapFilter; j++) {
    Set_Array<double, int>(ptr_sten2, 0.0, cube_len);
    for (i = 0; i < cube_len; i++) {
      if ((*ptr_sten1) != 0) {
        *ptr_sten2 += 2.0*NDim * (*ptr_sten1);
        for (k = 0; k < NDim*2; k++) {
          *(ptr_sten2 + local_displs[k]) -= 1.0 * (*ptr_sten1);
        }
      }
      ptr_sten1++;
      ptr_sten2++;
    }
    ptr_sten1 -= cube_len;
    ptr_sten2 -= cube_len;
    Swap<double*>(ptr_sten1, ptr_sten2);
  }

  // 5. Compute Len; copy sten1 to Coef and grid_displs to Displs
  Len = 0;
  for (i = 0; i < cube_len; i++) {
    if (ptr_sten1[i] != 0.0) {
      Len++;
    }
  }
  New_1D_Array<double, int>(&Coef, Len);
  New_1D_Array<long, int>(&Displs, Len);
  double *ptr_Coef = Coef;
  long *ptr_Displs = Displs;
  for (i = 0; i < cube_len; i++) {
    if (ptr_sten1[i] != 0.0) {
      *ptr_Coef++ = ptr_sten1[i];
      *ptr_Displs++ = grid_displs[i];
    }
  }

  // 6. Clean up
  Delete_1D_Array<int>(&local_displs);
  Delete_1D_Array<long>(&tank_len);
  Delete_1D_Array<int>(&idx);
  Delete_1D_Array<long>(&grid_displs);
  Delete_1D_Array<double>(&sten1);
  Delete_1D_Array<double>(&sten2);
}

//--------------------------------------------------------------------------
long Stencil::Filtering(double **data) {
  int i;
  long k;

  double *data_in = *data;
  double *data_out = NULL;

  // Compute the length of the filtered data
  long size_out = 1;
  for (i = 0; i < NDim; i++) {
    size_out *= (Dim[i] - 2*NumLapFilter);
  }

  // Allocate the filtered data (initialized zero)
  New_1D_Array<double, long>(&data_out, size_out);

  // Setup a multi-idx
  long *lower = NULL;
  long *upper = NULL;
  long *idx = NULL;
  New_1D_Array<long, int>(&lower, NDim);
  New_1D_Array<long, int>(&upper, NDim);
  New_1D_Array<long, int>(&idx, NDim);
  for (i = 0; i < NDim; i++) {
    lower[i] = NumLapFilter;
    upper[i] = Dim[i] - NumLapFilter;
    idx[i] = NumLapFilter;
  }
  MultiIdx<long> midx(NDim, lower, upper, idx);

  // The tank_len array
  long *tank_len = NULL;
  New_1D_Array<long, int>(&tank_len, NDim);
  tank_len[NDim-1] = 1;
  for (i = NDim-2; i >= 0; i--){
    tank_len[i] = tank_len[i+1] * Dim[i+1];
  }

  // Fill the array
  double *ptr_data_out = data_out;
  for (k = 0; k < size_out; k++) {
    // Find the corresponding location in the original data
    long loc = 0;
    for (i = 0; i < NDim; i++) {
      loc += tank_len[i] * idx[i];
    }

    // Apply stencil
    for (i = 0; i < Len; i++) {
      *ptr_data_out += ( Coef[i] * data_in[ loc+Displs[i] ] );
    }
    ptr_data_out++;

    // Update idx
    midx.Next(idx);
  }

  // Cleanup
  Delete_1D_Array<double>(&data_in);
  Delete_1D_Array<long>(&lower);
  Delete_1D_Array<long>(&upper);
  Delete_1D_Array<long>(&idx);
  Delete_1D_Array<long>(&tank_len);

  *data = data_out;
  return size_out;
}

//--------------------------------------------------------------------------
RegularGridData::RegularGridData() : RegularGrid(), Data() {
  Init();
}

//--------------------------------------------------------------------------
void RegularGridData::Init(void) {
}

//--------------------------------------------------------------------------
RegularGridData::RegularGridData(MPI_Comm &comm) :
  RegularGrid(comm), Data(comm) {
  Init();
  Comm = comm;
}

//--------------------------------------------------------------------------
RegularGridData::RegularGridData(const RegularGridData &D) :
  RegularGrid(D), Data(D) {
  DeepCopy(D);
}

//--------------------------------------------------------------------------
void RegularGridData::DeepCopy(const RegularGridData &D) {
  Comm = D.Comm;
}

//--------------------------------------------------------------------------
RegularGridData& RegularGridData::operator= (const RegularGridData &D) {
  if (this == &D) {
    return *this;
  }
  RegularGrid::operator=(D);
  Data::operator=(D);
  DeAllocMem();
  DeepCopy(D);
  return *this;
}

//--------------------------------------------------------------------------
void RegularGridData::DeAllocMem(void) {
}

//--------------------------------------------------------------------------
RegularGridData::~RegularGridData() {
  DeAllocMem();
}

//--------------------------------------------------------------------------
void RegularGridData::SetGrid(RegularGrid &G) {
  RegularGrid::DeAllocMem();
  RegularGrid::DeepCopy(G);
}

//--------------------------------------------------------------------------
void RegularGridData::SetData(double  val) {
  Len = CalLen();
  Data::SetVal(val, Len);
}

//--------------------------------------------------------------------------
void RegularGridData::SetData(double *val) {
  Len = CalLen();
  Data::SetVal(val, Len);
}

//--------------------------------------------------------------------------
void RegularGridData::SetSymBernoulli(void) {
  Len = CalLen();
  Data::SetSymBernoulli(Len);
}

//--------------------------------------------------------------------------                                                                   
void RegularGridData::SetSymBernoulli(unsigned int seed) {
  Len = CalLen();
  Data::SetSymBernoulli(Len, seed);
}

//--------------------------------------------------------------------------
long RegularGridData::CalLen(void) {
  int i;
  long total_len = 1;
  for (i = 0; i < NDim; i++) {
    total_len *= DimLen[i];
  }
  return total_len;
}

//--------------------------------------------------------------------------
void RegularGridData::ApplyLapFilter(int num_filter) {
  if (num_filter == 0) {
    return;
  }

  // The grid after filtering
  RegularGrid AFRegularGrid(Comm);
  ComputeGridAfterFiltering(AFRegularGrid, num_filter);

  // Pad data
  long *APDimLen = NULL;
  New_1D_Array<long, int>(&APDimLen, NDim);
  PadData(&Val, AFRegularGrid, num_filter, APDimLen);

  // Local filtering
  Stencil mStencil;
  mStencil.SetFilteringInfo(NDim, APDimLen, num_filter);
  Len = mStencil.Filtering(&Val);

  // Reset my grid
  SetGrid(AFRegularGrid);

  // Cleanup
  Delete_1D_Array<long>(&APDimLen);
}

//--------------------------------------------------------------------------
void RegularGridData::ComputeGridAfterFiltering(RegularGrid &G, int num_filter){
  int i;
  long *AFDim = NULL;
  New_1D_Array<long, int>(&AFDim, NDim);
  for (i = 0; i < NDim; i++) {
    AFDim[i] = Dim[i] - 2*num_filter;
  }
  double *AFRangeOrDelta = NULL;
  New_1D_Array<double, int>(&AFRangeOrDelta, NDim);
  switch (mType) {
    case FIXED_DOMAIN:
      for (i = 0; i < NDim; i++) {
        AFRangeOrDelta[i] = mSpec.Range[i]
          * (Dim[i]-1-2*num_filter) / (Dim[i]-1);
      }
      break;
    case FIXED_SPACING:
      Set_Array<double, int>(AFRangeOrDelta, mSpec.Delta, NDim);
      break;
    case UNKNOWN_GRID_TYPE:
      break;
  }
  G.SetGrid(NDim, AFDim, mType, AFRangeOrDelta, NProc, bElimAllReduce);
  Delete_1D_Array<long>(&AFDim);
  Delete_1D_Array<double>(&AFRangeOrDelta);
}

//--------------------------------------------------------------------------
void RegularGridData::PadData(double **data, RegularGrid &grid_af,
                              int num_filter, long *dim_len_ap) {
  PadDataCommon(data, grid_af, num_filter, dim_len_ap);
}

//--------------------------------------------------------------------------
void RegularGridData::PadDataCommon(double **data, RegularGrid &grid_af,
                                    int num_filter, long *dim_len_ap) {
  // data1 <- data
  // for each dim {
  //   data1 -> data2: rearrange data layout
  //   data2 -> data3: alltoall
  //   data3 -> data4: rearrange data layout
  //   delete data1, data2, data3
  //   swap(data1, data4)
  // }
  // data <- data1
  double *data1 = NULL, *data2 = NULL, *data3 = NULL, *data4 = NULL;
  data1 = *data;

  // Flow chart:
  // along each dimension {
  //   get to know the data distribution of my group of procs
  long *all_start = NULL, *all_len = NULL;
  //   determine my data layout after alltoall (what I need)
  int *recvcnts = NULL, *rdispls = NULL;
  //   and at the same time, prepare to determine data layout before alltoall
  //   (what my peer should send to me)
  int *peer_start = NULL, *peer_len = NULL;
  //   still preparing.. (switching role, I know what to send to peer)
  int *send_start = NULL, *send_len = NULL;
  //   finally, compute data layout before alltoall
  int *sendcnts = NULL, *sdispls = NULL;
  //   lengths of different dimensions
  long len_before_current_dim;
  long len_after_current_dim;
  long len_except_current_dim;
  long data2_len_current_dim;
  long data2_len_total;
  long data3_len_current_dim;
  long data3_len_total;
  // }

  // Initialize dim_start_ap and dim_len_ap
  long *dim_start_ap = NULL;
  New_1D_Array<long, int>(&dim_start_ap, NDim);
  grid_af.GetPartitioning(dim_start_ap, dim_len_ap);

  // Do padding along each dimension
  int current_dim;
  int i;
  for (current_dim = 0; current_dim < NDim; current_dim++) {

    // Update dim_len_ap
    dim_len_ap[current_dim] += 2*num_filter;

    // If only one processor along this dimension, no need to pad
    if (NProc[current_dim] == 1) {
      continue;
    }

    // Construct a grid communicator
    MPI_Comm grid_comm;
    long color = 0;
    long multiplier = 1;
    for (i = 0; i < current_dim; i++) {
      color += multiplier * DimStart[i];
      multiplier *= DimLen[i];
    }
    multiplier *= Dim[i];
    i++;
    for (; i < NDim; i++) {
      color += multiplier * DimStart[i];
      multiplier *= DimLen[i];
    }
    long key = DimStart[current_dim];
    MPI_Comm_split(Comm, (int)color, (int)key, &grid_comm);
    int grid_comm_size;
    MPI_Comm_size(grid_comm, &grid_comm_size);

    // Compute all_start and all_len
    long my_start = DimStart[current_dim];
    long my_len = DimLen[current_dim];
    New_1D_Array<long, int>(&all_start, grid_comm_size);
    New_1D_Array<long, int>(&all_len, grid_comm_size);
    MPI_Allgather(&my_start, 1, MPI_LONG, all_start, 1, MPI_LONG, grid_comm);
    MPI_Allgather(&my_len, 1, MPI_LONG, all_len, 1, MPI_LONG, grid_comm);

    // Modify my_start and my_len;
    // Compute recvcnts, rdispls, peer_start, and peer_len
    my_start = dim_start_ap[current_dim];
    my_len = dim_len_ap[current_dim];
    New_1D_Array<int, int>(&recvcnts, grid_comm_size);
    New_1D_Array<int, int>(&rdispls, grid_comm_size);
    New_1D_Array<int, int>(&peer_start, grid_comm_size);
    New_1D_Array<int, int>(&peer_len, grid_comm_size);
    FillRecvArrays(my_start, my_len, all_start, all_len, grid_comm_size,
                   recvcnts, rdispls, peer_start, peer_len);

    // Compute send_start and send_len
    New_1D_Array<int, int>(&send_start, grid_comm_size);
    New_1D_Array<int, int>(&send_len, grid_comm_size);
    MPI_Alltoall(peer_start, 1, MPI_INT, send_start, 1, MPI_INT, grid_comm);
    MPI_Alltoall(peer_len, 1, MPI_INT, send_len, 1, MPI_INT, grid_comm);

    // Compute sendcnts and sdispls
    New_1D_Array<int, int>(&sendcnts, grid_comm_size);
    New_1D_Array<int, int>(&sdispls, grid_comm_size);
    sendcnts[0] = send_len[0];
    sdispls[0] = 0;
    for (i = 1; i < grid_comm_size; i++) {
      sendcnts[i] = send_len[i];
      sdispls[i] = sdispls[i-1] + sendcnts[i-1];
    }

    // Compute len's
    len_before_current_dim = 1;
    for (i = 0; i < current_dim; i++) {
      len_before_current_dim *= dim_len_ap[i];
    }
    len_after_current_dim = 1;
    for (i = current_dim+1; i < NDim; i++) {
      len_after_current_dim *= DimLen[i];
    }
    len_except_current_dim = len_before_current_dim * len_after_current_dim;

    // Compute data2_len's and allocate memory for data2
    data2_len_current_dim = sdispls[grid_comm_size-1]
      + sendcnts[grid_comm_size-1];
    data2_len_total = data2_len_current_dim * len_except_current_dim;
    New_1D_Array<double, long>(&data2, data2_len_total);

    // Rearrange data layout from data1 to data2
    for (i = 0; i < grid_comm_size; i++) {
      CopyData(data1, send_start[i], DimLen[current_dim],
               data2, sdispls[i], data2_len_current_dim,
               send_len[i], len_before_current_dim, len_after_current_dim);
    }

    // Compute data3_len's and allocate memory for data3
    data3_len_current_dim = rdispls[grid_comm_size-1]
      + recvcnts[grid_comm_size-1];
    data3_len_total = data3_len_current_dim * len_except_current_dim;
    New_1D_Array<double, long>(&data3, data3_len_total);

    // alltoall
    for (i = 0; i < grid_comm_size; i++) {
      sendcnts[i] *= (int)len_except_current_dim;
      sdispls[i] *= (int)len_except_current_dim;
      recvcnts[i] *= (int)len_except_current_dim;
      rdispls[i] *= (int)len_except_current_dim;
    }
    MPI_Alltoallv(data2, sendcnts, sdispls, MPI_DOUBLE,
                  data3, recvcnts, rdispls, MPI_DOUBLE, grid_comm);
    for (i = 0; i < grid_comm_size; i++) {
      recvcnts[i] /= (int)len_except_current_dim;
      rdispls[i] /= (int)len_except_current_dim;
    }

    // Allocate memory for data4
    New_1D_Array<double, long>(&data4, data3_len_total);

    // Rearrange data layout from data3 to data4
    RearrangeData(data3, data4, recvcnts, grid_comm_size, data3_len_current_dim,
                  len_before_current_dim, len_after_current_dim);

    // Delete memory
    Delete_1D_Array<double>(&data1);
    Delete_1D_Array<double>(&data2);
    Delete_1D_Array<double>(&data3);
    Delete_1D_Array<long>(&all_start);
    Delete_1D_Array<long>(&all_len);
    Delete_1D_Array<int>(&recvcnts);
    Delete_1D_Array<int>(&rdispls);
    Delete_1D_Array<int>(&peer_start);
    Delete_1D_Array<int>(&peer_len);
    Delete_1D_Array<int>(&send_start);
    Delete_1D_Array<int>(&send_len);
    Delete_1D_Array<int>(&sendcnts);
    Delete_1D_Array<int>(&sdispls);

    // Swap pointer
    Swap<double*>(data1, data4);

    // Free the communicator
    MPI_Comm_free(&grid_comm);
  }

  // Cleanup
  Delete_1D_Array<long>(&dim_start_ap);

  // Assign pointer
  *data = data1;
}

//--------------------------------------------------------------------------
void RegularGridData::FillRecvArrays(long my_start, long my_len,
                                     long *grp_starts, long *grp_lens,
                                     int grp_size, int *recvcnts, int *rdispls,
                                     int *peer_start, int *peer_len) {
  long current_my_start = my_start;
  long my_end = my_start + my_len;
  long current_grp_start = grp_starts[0];
  int idx = 0;

  while (current_grp_start < current_my_start && idx < grp_size) {
    recvcnts[idx] = 0;
    rdispls[idx] = 0;
    peer_start[idx] = 0;
    peer_len[idx] = 0;
    current_grp_start += grp_lens[idx];
    idx++;
  }
  if (idx == grp_size) {
    return;
  }

  if (current_grp_start >= my_end) {
    recvcnts[idx-1] = my_len;
    peer_start[idx-1] = my_start - (current_grp_start - grp_lens[idx-1]);
    peer_len[idx-1] = my_len;
    while (idx < grp_size) {
      recvcnts[idx] = 0;
      rdispls[idx] = my_len;
      peer_start[idx] = 0;
      peer_len[idx] = 0;
      idx++;
    }
    return;
  }
  else if (current_grp_start > current_my_start) {
    recvcnts[idx-1] = current_grp_start - current_my_start;
    peer_start[idx-1] = my_start - (current_grp_start - grp_lens[idx-1]);
    peer_len[idx-1] = recvcnts[idx-1];
    current_my_start = current_grp_start;
  }

  while (idx < grp_size && current_grp_start + grp_lens[idx] <= my_end) {
    recvcnts[idx] = grp_lens[idx];
    rdispls[idx] = current_my_start - my_start;
    peer_start[idx] = 0;
    peer_len[idx] = recvcnts[idx];
    current_grp_start += grp_lens[idx];
    current_my_start = current_grp_start;
    idx++;
  }
  if (idx == grp_size) {
    return;
  }

  recvcnts[idx] = my_end - current_my_start;
  rdispls[idx] = current_my_start - my_start;
  peer_start[idx] = 0;
  peer_len[idx] = recvcnts[idx];
  idx++;
  while (idx < grp_size) {
    recvcnts[idx] = 0;
    rdispls[idx] = my_len;
    peer_start[idx] = 0;
    peer_len[idx] = 0;
    idx++;
  }
}

//--------------------------------------------------------------------------
void RegularGridData::CopyData(double *data_in, int in_start, long in_dim_len,
                               double *data_out, int out_start,long out_dim_len,
                               int cp_dim_len,
                               long len_before_current_dim,
                               long len_after_current_dim) {
  long k;
  double *ptr_data_in = data_in + len_after_current_dim * in_start;
  double *ptr_data_out = data_out + len_before_current_dim
    * len_after_current_dim * out_start;
  long inc_data_in = len_after_current_dim * in_dim_len;
  long cp_len = len_after_current_dim * cp_dim_len;
  for (k = 0; k < len_before_current_dim; k++) {
    Set_Array<double, long>(ptr_data_out, ptr_data_in, cp_len);
    ptr_data_in += inc_data_in;
    ptr_data_out += cp_len;
  }
}

//--------------------------------------------------------------------------
void RegularGridData::RearrangeData(double *data_in, double *data_out,
                                    int *cp_dim_len, int grid_comm_size,
                                    long len_current_dim,
                                    long len_before_current_dim,
                                    long len_after_current_dim) {
  int i;
  long k;
  double *ptr_data_in = data_in;
  double *ptr_data_out = data_out;
  long inc_data_out = len_current_dim * len_after_current_dim;
  for (i = 0; i < grid_comm_size; i++) {
    long cp_len = len_after_current_dim * cp_dim_len[i];
    double *ptr_data_out2 = ptr_data_out;
    for (k = 0; k < len_before_current_dim; k++) {
      Set_Array<double, long>(ptr_data_out2, ptr_data_in, cp_len);
      ptr_data_in += cp_len;
      ptr_data_out2 += inc_data_out;
    }
    ptr_data_out += len_after_current_dim * cp_dim_len[i];
  }  
}

// Author: J. Chen
// $Date: 2013/03/28 15:50:43$
