#ifndef _REGULAR_GRID_
#define _REGULAR_GRID_
// Defines a d-dimensional regular grid and several data structures
// based on the regular grid. This file contains the following
// classes:
//  class RegularGrid
//  class Stencil
//  class RegularGridData: public RegularGrid, public Data
//  template<class Kernel> class RegularGridCovarMat: public RegularGridData
//  template<class Kernel> class RegularGridCovarVec: public RegularGridData
//
// RegularGridCovarVec is the vector of covariances between a new
// point and the grid points.

#include <mpi.h>
#include <complex>
#include "Data.hpp"
#include "Utility.hpp"
#include "MultiIdx.hpp"
#include "ToepSolverWrapper.hpp"

//--------------------------------------------------------------------------
// Grid type
typedef enum { UNKNOWN_GRID_TYPE, FIXED_DOMAIN, FIXED_SPACING } GRID_TYPE;

// Grid specification w.r.t. each type
typedef union { double *Range; double *Delta; } GRID_SPEC;

class RegularGrid {

public:

  // Constructor, destructor
  RegularGrid();
  RegularGrid(MPI_Comm &comm);
  RegularGrid(const RegularGrid &G);
  RegularGrid& operator= (const RegularGrid &G);
  ~RegularGrid();

  // Set. Note: Data partitioning is computed inside SetGrid
  void SetGrid(int ndim, long  dim, GRID_TYPE type, double  spval,
               int *nproc, bool elim_allreduce);
  void SetGrid(int ndim, long  dim, GRID_TYPE type, double *spval,
               int *nproc, bool elim_allreduce);
  void SetGrid(int ndim, long *dim, GRID_TYPE type, double  spval,
               int *nproc, bool elim_allreduce);
  void SetGrid(int ndim, long *dim, GRID_TYPE type, double *spval,
               int *nproc, bool elim_allreduce);

  // Get
  int GetNDim(void);
  void GetDim(long *dim);
  long GetPartitioning(long *dim_start, long *dim_len); // Return prod(dim_len)

protected:

  // Needed by constructor and destructor
  void Init(void);
  void DeepCopy(const RegularGrid &G);
  void DeAllocMem(void);

  int NDim;                // Number of dimensions
  long *Dim;               // Grid size of each dimension
  GRID_TYPE mType;         // Grid type
  GRID_SPEC mSpec;         // Grid spec
  int *NProc;              // Number of processes along each dimension
  long *DimStart;          // Partitioning
  long *DimLen;            // Partitioning
  MPI_Comm Comm;

  ToepSolver mToepSolver;    // Toeplitz solver attached to the grid
  bool bElimAllReduce;       // Whether use a solver without AllReduce

private:

};

//--------------------------------------------------------------------------
class Stencil {

public:

  // Constructor, destructor
  Stencil();
  Stencil(const Stencil &S);
  Stencil& operator= (const Stencil &S);
  ~Stencil();

  // Set
  void SetFilteringInfo(int ndim, long *dim, int num_filter);

  // Major computational component
  long                          // Return the length of filtered data
  Filtering(double **data);     // data serves as both input and output

private:

  // Needed by constructor and destructor
  void DeepCopy(const Stencil &S);
  void DeAllocMem(void);

  int NDim;                     // Number of dimensions
  long *Dim;                    // Dimension BEFORE FILTERING
  int NumLapFilter;             // Number of Laplacian filtering

  double *Coef;                 // Coefficients of the stencil
  long *Displs;                 // Displacements of the stencil points
  int Len;                      // Length of the above two arrays

};

//--------------------------------------------------------------------------
// To fully "intialize" this class, need to
// Call SetGrid(RegularGrid &G);
// Call SetData(double *val);
class RegularGridData: public RegularGrid, public Data {

public:

  // Constructor, destructor
  RegularGridData();
  RegularGridData(MPI_Comm &comm);
  RegularGridData(const RegularGridData &D);
  RegularGridData& operator= (const RegularGridData &D);
  ~RegularGridData();

  // Set
  void SetGrid(RegularGrid &G);
  void SetData(double  val);
  void SetData(double *val);
  void SetSymBernoulli(void);
  void SetSymBernoulli(unsigned int seed);

  // Major computational component
  void ApplyLapFilter(int num_filter);

protected:

  // Needed by constructor and destructor
  void Init(void);
  void DeepCopy(const RegularGridData &D);
  void DeAllocMem(void);

  // Needed by SetData(), SetSymBernoulii() and other places
  long CalLen(void);

  // Needed by ApplyLapFilter()
  void ComputeGridAfterFiltering(RegularGrid &G, int num_filter);
  void PadData(double **data, RegularGrid &grid_af, int num_filter,
               long *dim_len_ap);
  void PadDataCommon(double **data, RegularGrid &grid_af, int num_filter,
                     long *dim_len_ap);
  void FillRecvArrays(long my_start, long my_len, long *grp_starts,
                      long *grp_lens, int grp_size, int *recvcnts, int *rdispls,
                      int *peer_start, int *peer_len);
  void CopyData(double *data_in, int in_start, long in_dim_len,
                double *data_out, int out_start,long out_dim_len,
                int cp_dim_len, long len_before_current_dim,
                long len_after_current_dim);
  void RearrangeData(double *data_in, double *data_out, int *cp_dim_len,
                     int grid_comm_size, long len_current_dim,
                     long len_before_current_dim, long len_after_current_dim);

  MPI_Comm Comm;

private:

};

//--------------------------------------------------------------------------
// To fully "intialize" this class, need to
// Call SetGrid(RegularGrid &G);
// Call UseKernel(Kernel &kernel, bool need_partials[3]);
template<class Kernel>
class RegularGridCovarMat: public RegularGridData {

public:

  // Constructor, destructor
  RegularGridCovarMat();
  RegularGridCovarMat(MPI_Comm &comm);
  RegularGridCovarMat(const RegularGridCovarMat &C);
  RegularGridCovarMat& operator= (const RegularGridCovarMat &C);
  ~RegularGridCovarMat();

  // Set
  void SetToepSolverParam(ToepSolverParam &P);

  // Get
  int GetNumParam(void);
  Kernel GetKernel(void);

  // Compare
  bool Compare(RegularGridCovarMat &C);

  // Major computational component
  void UseKernel(Kernel &kernel, bool need_partials[3]);
  void ApplyLapFilter(int num_filter);
  // y = K*x
  void KMult(RegularGridData &x, RegularGridData &y);
  // y = dK[i]*x
  void dKMult(int i, RegularGridData &x, RegularGridData &y);
  // y = ddK[i][j]*x
  void ddKMult(int i, int j, RegularGridData &x, RegularGridData &y);
  // x = inv(K)*b
  void KSolve(RegularGridData &b, RegularGridData &x);

protected:

  // Needed by constructor and destructor
  void Init(void);
  void DeepCopy(const RegularGridCovarMat &C);
  void DeAllocMem(void);
  void DeAllocKs(void);
  void DeAllocEigs(void);

  // Needed by ApplyLapFilter()
  void PadData(double **data, RegularGrid &grid_af, int num_filter,
               long *dim_len_ap, bool reset_dim = false);
  void ComputeEigs(void);

  Kernel mKernel;                   // Kernel
  ToepSolverParam mToepSolverParam; // Solver parameter

  bool bNeedPartials[3];   // [i] asks whether order-i partials are needed
  int NumParam;            // Number of parameters in K
  long LenK;               // Length of K
  double *K;               // K
  double **dK;             // first partials of K
  double ***ddK;           // second partials of K

  long LenKEmbedEig;             // Length of buffer to hold KEmbedEig
  double_complex *KEmbedEig;     // Eigvals of embed of K
  double_complex **dKEmbedEig;   // Eigvals of embed of first partials of K
  double_complex ***ddKEmbedEig; // Eigvals of embed of second partials of K

  long LenPrecondEig;            // Length of of buffer to hold PrecondEig
  double_complex *PrecondEig;    // Eigenvalues of preconditioner

private:

};

//--------------------------------------------------------------------------
// To fully "intialize" this class, need to
// Call SetGrid(RegularGrid &G);
// Call SetVal(Kernel &kernel, double *coord);
template<class Kernel>
class RegularGridCovarVec: public RegularGridData {

public:

  // Constructor, destructor
  RegularGridCovarVec();
  RegularGridCovarVec(MPI_Comm &comm);
  RegularGridCovarVec(const RegularGridCovarVec &V);
  RegularGridCovarVec& operator= (const RegularGridCovarVec &V);
  ~RegularGridCovarVec();

  // Major computational component
  void SetVal(Kernel &kernel, double *coord);

protected:

  // Needed by constructor and destructor
  void Init(void);
  void DeepCopy(const RegularGridCovarVec &V);
  void DeAllocMem(void);

private:

};

//--------------------------------------------------------------------------
template<class Kernel>
RegularGridCovarMat<Kernel>::
RegularGridCovarMat() : RegularGridData() {
  Init();
}

//--------------------------------------------------------------------------
template<class Kernel>
void RegularGridCovarMat<Kernel>::
Init(void) {
  bNeedPartials[0] = false;
  bNeedPartials[1] = false;
  bNeedPartials[2] = false;
  NumParam = 0;
  LenK = 0;
  K = NULL;
  dK = NULL;
  ddK = NULL;
  LenKEmbedEig = 0;
  KEmbedEig = NULL;
  dKEmbedEig = NULL;
  ddKEmbedEig = NULL;
  LenPrecondEig = 0;
  PrecondEig = NULL;
}

//--------------------------------------------------------------------------
template<class Kernel>
RegularGridCovarMat<Kernel>::
RegularGridCovarMat(MPI_Comm &comm) : RegularGridData(comm) {
  Init();
}

//--------------------------------------------------------------------------
template<class Kernel>
RegularGridCovarMat<Kernel>::
RegularGridCovarMat(const RegularGridCovarMat &C) : RegularGridData(C) {
  K = NULL;
  dK = NULL;
  ddK = NULL;
  KEmbedEig = NULL;
  dKEmbedEig = NULL;
  ddKEmbedEig = NULL;
  PrecondEig = NULL;
  DeepCopy(C);
}

//--------------------------------------------------------------------------
template<class Kernel>
void RegularGridCovarMat<Kernel>::
DeepCopy(const RegularGridCovarMat &C) {
  int i, j;
  mKernel = C.mKernel;
  mToepSolverParam = C.mToepSolverParam;
  memcpy(bNeedPartials, C.bNeedPartials, 3*sizeof(bool));
  NumParam = C.NumParam;

  LenK = C.LenK;
  New_1D_Array<double, long>(&K, LenK);
  Set_Array<double, long>(K, C.K, LenK);
  if (bNeedPartials[1]) {
    New_2D_Array<double, int, long>(&dK, NumParam, LenK);
    for (i = 0; i < NumParam; i++) {
      Set_Array<double, long>(dK[i], C.dK[i], LenK);
    }
  }
  if (bNeedPartials[2]) {
    New_3D_Array_Sym2D<double, int, long>(&ddK, NumParam, LenK);
    for (i = 0; i < NumParam; i++) {
      for (j = 0; j <= i; j++) {
        Set_Array<double, long>(ddK[i][j], C.ddK[i][j], LenK);
      }
    }
  }

  LenKEmbedEig = C.LenKEmbedEig;
  New_1D_Array<double_complex, long>(&KEmbedEig, LenKEmbedEig);
  Set_Array<double_complex, long>(KEmbedEig, C.KEmbedEig, LenKEmbedEig);
  if (bNeedPartials[1]) {
    New_2D_Array<double_complex, int, long>(&dKEmbedEig, NumParam,
                                            LenKEmbedEig);
    for (i = 0; i < NumParam; i++) {
      Set_Array<double_complex, long>(dKEmbedEig[i], C.dKEmbedEig[i],
                                      LenKEmbedEig);
    }
  }
  if (bNeedPartials[2]) {
    New_3D_Array_Sym2D<double_complex, int, long>(&ddKEmbedEig, NumParam,
                                                  LenKEmbedEig);
    for (i = 0; i < NumParam; i++) {
      for (j = 0; j <= i; j++) {
        Set_Array<double_complex, long>(ddKEmbedEig[i][j], C.ddKEmbedEig[i][j],
                                        LenKEmbedEig);
      }
    }
  }

  LenPrecondEig = C.LenPrecondEig;
  New_1D_Array<double_complex, long>(&PrecondEig, LenPrecondEig);
  Set_Array<double_complex, long>(PrecondEig, C.PrecondEig, LenPrecondEig);
}

//--------------------------------------------------------------------------
template<class Kernel>
RegularGridCovarMat<Kernel>& RegularGridCovarMat<Kernel>::
operator= (const RegularGridCovarMat &C) {
  if (this == &C) {
    return *this;
  }
  RegularGridData::operator=(C);
  DeAllocMem();
  DeepCopy(C);
  return *this;
}

//--------------------------------------------------------------------------
template<class Kernel>
void RegularGridCovarMat<Kernel>::
DeAllocMem(void) {
  DeAllocKs();
  DeAllocEigs();
}

//--------------------------------------------------------------------------
template<class Kernel>
void RegularGridCovarMat<Kernel>::
DeAllocKs(void) {
  Delete_1D_Array<double>(&K);
  Delete_2D_Array<double, int>(&dK, NumParam);
  Delete_3D_Array_Sym2D<double, int>(&ddK, NumParam);
}

//--------------------------------------------------------------------------
template<class Kernel>
void RegularGridCovarMat<Kernel>::
DeAllocEigs(void) {
  Delete_1D_Array<double_complex>(&KEmbedEig);
  Delete_2D_Array<double_complex, int>(&dKEmbedEig, NumParam);
  Delete_3D_Array_Sym2D<double_complex, int>(&ddKEmbedEig, NumParam);
  Delete_1D_Array<double_complex>(&PrecondEig);
}

//--------------------------------------------------------------------------
template<class Kernel>
RegularGridCovarMat<Kernel>::
~RegularGridCovarMat() {
  DeAllocMem();
}

//--------------------------------------------------------------------------
template<class Kernel>
void RegularGridCovarMat<Kernel>::
SetToepSolverParam(ToepSolverParam &P) {
  mToepSolverParam = P;
}

//--------------------------------------------------------------------------
template<class Kernel>
int RegularGridCovarMat<Kernel>::
GetNumParam(void) {
  return NumParam;
}

//--------------------------------------------------------------------------
template<class Kernel>
Kernel RegularGridCovarMat<Kernel>::
GetKernel(void) {
  return mKernel;
}

//--------------------------------------------------------------------------
template<class Kernel>
bool RegularGridCovarMat<Kernel>::
Compare(RegularGridCovarMat &C) {
  // Compare kernel
  Kernel kernel = C.GetKernel();
  if (mKernel.Compare(kernel) == false) { return false; }

  // Compare grid
  if (NDim != C.NDim) { return false; }
  int i;
  for (i = 0; i < NDim; i++) {
    if (Dim[i] != C.Dim[i]) { return false; }
  }
  if (mType != C.mType) { return false; }
  switch (mType) {
    case FIXED_DOMAIN:
      for (i = 0; i < NDim; i++) {
        if (mSpec.Range[i] != C.mSpec.Range[i]) { return false; }
      }
      break;
    case FIXED_SPACING:
      for (i = 0; i < NDim; i++) {
        if (mSpec.Delta[i] != C.mSpec.Delta[i]) { return false; }
      }
      break;
    case UNKNOWN_GRID_TYPE:
      break;
  }
  for (i = 0; i < NDim; i++) {
    if (NProc[i] != C.NProc[i]) { return false; }
  }
  for (i = 0; i < NDim; i++) {
    if (DimStart[i] != C.DimStart[i]) { return false; }
  }
  for (i = 0; i < NDim; i++) {
    if (DimLen[i] != C.DimLen[i]) { return false; }
  }

  // Note sure if need to compare other member variables..

  return true;
}

//--------------------------------------------------------------------------
template<class Kernel>
void RegularGridCovarMat<Kernel>::
UseKernel(Kernel &kernel, bool need_partials[3]) {
  int i, j;
  long k;

  DeAllocKs();
  DeAllocEigs();

  // Store a copy of the kernel
  mKernel = kernel;

  // Set bNeedPartials
  Set_Array<bool, int>(bNeedPartials, need_partials, 3);

  // Get NumParam
  NumParam = mKernel.GetLenTheta();

  // Compute grid_dist
  double **grid_dist = NULL;
  New_2D_Array<double, int>(&grid_dist, NDim, Dim);
  double spacing = 0.0;
  for (i = 0; i < NDim; i++) {
    switch (mType) {
    case FIXED_DOMAIN:
      spacing = mSpec.Range[i]/(Dim[i]-1);
      break;
    case FIXED_SPACING:
      spacing = mSpec.Delta[i];
      break;
    case UNKNOWN_GRID_TYPE:
      throw std::string("RegularGridCovarMat::UseKernel: Unknown grid type!\n");
      break;
    }
    for (k = 0; k < Dim[i]; k++) {
      grid_dist[i][k] = spacing * k;
    }
  }

  // Compute K, dK, ddK
  LenK = RegularGridData::CalLen();
  New_1D_Array<double, long>(&K, LenK);
  if (bNeedPartials[1]) {
    New_2D_Array<double, int, long>(&dK, NumParam, LenK);
  }
  if (bNeedPartials[2]) {
    New_3D_Array_Sym2D<double, int, long>(&ddK, NumParam, LenK);
  }

  double *x = NULL;
  New_1D_Array<double, int>(&x, NDim);
  long *mIdx = NULL;
  New_1D_Array<long, int>(&mIdx, NDim);
  Set_Array<long, int>(mIdx, DimStart, NDim);
  double *p = NULL;
  double **pp = NULL;
  if (bNeedPartials[1]) {
    New_1D_Array<double, int>(&p, NumParam);
  }
  if (bNeedPartials[2]) {
    New_2D_Array_Sym<double, int>(&pp, NumParam);
  }
  long *upper = NULL;
  New_1D_Array<long, int>(&upper, NDim);
  for (i = 0; i < NDim; i++) {
    upper[i] = DimStart[i] + DimLen[i];
  }
  MultiIdx<long> mMultiIdx(NDim, DimStart, upper, DimStart);

  for (k = 0; k < LenK; k++) {
    for (i = 0; i < NDim; i++) {
      x[i] = grid_dist[i][mIdx[i]];
    }
    K[k] = mKernel.GetValues(x, p, pp);
    if (bNeedPartials[1]) {
      for (i = 0; i < NumParam; i++) {
        dK[i][k] = p[i];
      }
    }
    if (bNeedPartials[2]) {
      for (i = 0; i < NumParam; i++) {
        for (j = 0; j <= i; j++) {
          ddK[i][j][k] = pp[i][j];
        }
      }
    }
    mMultiIdx.Next(mIdx);
  }

  // Compute KEmbedEig, dKEmbedEig, ddKEmbedEig and PrecondEig
  ComputeEigs();

  // Clean up
  Delete_2D_Array<double, int>(&grid_dist, NDim);
  Delete_1D_Array<double>(&x);
  Delete_1D_Array<long>(&mIdx);
  Delete_1D_Array<double>(&p);
  Delete_2D_Array<double, int>(&pp, NumParam);
  Delete_1D_Array<long>(&upper);
}

//--------------------------------------------------------------------------
template<class Kernel>
void RegularGridCovarMat<Kernel>::
ApplyLapFilter(int num_filter) {
  if (num_filter == 0) {
    return;
  }

  int i, j;

  // Check conditions
  int my_err = 0, err;
  for (i = 0; i < NDim; i++) {
    if (DimLen[i] <= num_filter) {
      my_err = 1;
      break;
    }
  }
  MPI_Allreduce(&my_err, &err, 1, MPI_INT, MPI_SUM, Comm);
  if (err > 0) {
    throw std::string("RegularGridCovarMat::ApplyLapFilter: Dim too small!\n");
  }

  // The grid after filtering
  RegularGrid AFRegularGrid(Comm);
  RegularGridData::ComputeGridAfterFiltering(AFRegularGrid, num_filter);

  // Pad data
  long *APDimLen = NULL;
  New_1D_Array<long, int>(&APDimLen, NDim);
  PadData(&K, AFRegularGrid, num_filter, APDimLen, true);
  if (bNeedPartials[1]) {
    for (i = 0; i < NumParam; i++) {
      PadData(&(dK[i]), AFRegularGrid, num_filter, APDimLen, true);
    }
  }
  if (bNeedPartials[2]) {
    for (i = 0; i < NumParam; i++) {
      for (j = 0; j <= i; j++) {
        if (i != NumParam-1 || j != i) {
          PadData(&(ddK[i][j]), AFRegularGrid, num_filter, APDimLen, true);
        }
        else {
          PadData(&(ddK[i][j]), AFRegularGrid, num_filter, APDimLen, false);
        }
      }
    }
  }

  // Local filtering
  Stencil mStencil;
  mStencil.SetFilteringInfo(NDim, APDimLen, 2*num_filter);
  LenK = mStencil.Filtering(&K);
  if (bNeedPartials[1]) {
    for (i = 0; i < NumParam; i++) {
      mStencil.Filtering(&(dK[i]));
    }
  }
  if (bNeedPartials[2]) {
    for (i = 0; i < NumParam; i++) {
      for (j = 0; j <= i; j++) {
        mStencil.Filtering(&(ddK[i][j]));
      }
    }
  }

  // Reset my grid. Must do this before recomputing the eigenvalues.
  RegularGridData::SetGrid(AFRegularGrid);

  // Recompute KEmbedEig, dKEmbedEig, ddKEmbedEig and PrecondEig
  DeAllocEigs();
  ComputeEigs();

  // Cleanup
  Delete_1D_Array<long>(&APDimLen);
}

//--------------------------------------------------------------------------
template<class Kernel>
void RegularGridCovarMat<Kernel>::
PadData(double **data, RegularGrid &grid_af, int num_filter, long *dim_len_ap,
        bool reset_dim) {

  // Dimensions before padding
  long *DimStartCopy = NULL, *DimLenCopy = NULL;

  if (reset_dim) {
    New_1D_Array<long, int>(&DimStartCopy, NDim);
    New_1D_Array<long, int>(&DimLenCopy, NDim);
    Set_Array<long, int>(DimStartCopy, DimStart, NDim);
    Set_Array<long, int>(DimLenCopy, DimLen, NDim);
  }

  // Do a local padding to expand towards the lower end of each dim
  int current_dim;
  for (current_dim = 0; current_dim < NDim; current_dim++) {
    if (DimStart[current_dim] != 0) {
      // Modify DimStart
      DimStart[current_dim] += 2*num_filter;
    }
    else {
      int i;
      long len_before_current_dim = 1;
      for (i = 0; i < current_dim; i++) {
        len_before_current_dim *= DimLen[i];
      }
      long len_after_current_dim = 1;
      for (i = current_dim+1; i < NDim; i++) {
        len_after_current_dim *= DimLen[i];
      }
      long len_out = len_before_current_dim * len_after_current_dim
        * (DimLen[current_dim] + 2*num_filter);

      double *data_in = *data;
      double *data_out = NULL;
      New_1D_Array<double, long>(&data_out, len_out);

      long k;
      double *ptr_data_in = data_in;
      double *ptr_data_out = data_out;
      long inc_data_in = len_after_current_dim * DimLen[current_dim];
      long inc_data_diff = len_after_current_dim * 2*num_filter;
      long inc_data_out = inc_data_in + inc_data_diff;
      for (k = 0; k < len_before_current_dim; k++) {
        for (i = 0; i < 2*num_filter; i++) {
          Set_Array<double, long>(ptr_data_out + len_after_current_dim*i,
                                  ptr_data_in + len_after_current_dim
                                  *(2*num_filter-i),
                                  len_after_current_dim);
        }
        Set_Array<double, long>(ptr_data_out + inc_data_diff,
                                ptr_data_in, inc_data_in);
        ptr_data_in += inc_data_in;
        ptr_data_out += inc_data_out;
      }

      Delete_1D_Array<double>(&data_in);
      *data = data_out;

      // Modify DimLen
      DimLen[current_dim] += 2*num_filter;
    }
  }

  // Call RegularGridData::PadDataCommon
  RegularGridData::PadDataCommon(data, grid_af, 2*num_filter, dim_len_ap);

  // Handle reset_dim
  if (reset_dim) {
    Set_Array<long, int>(DimStart, DimStartCopy, NDim);
    Set_Array<long, int>(DimLen, DimLenCopy, NDim);
  }
  Delete_1D_Array<long>(&DimStartCopy);
  Delete_1D_Array<long>(&DimLenCopy);
}

//--------------------------------------------------------------------------
template<class Kernel>
void RegularGridCovarMat<Kernel>::
ComputeEigs(void) {
  int i, j;
  mToepSolver.GetLens(LenKEmbedEig, LenPrecondEig);
  New_1D_Array<double_complex, long>(&KEmbedEig, LenKEmbedEig);
  New_1D_Array<double_complex, long>(&PrecondEig, LenPrecondEig);
  mToepSolver.Planning(K, KEmbedEig, PrecondEig);
  if (bNeedPartials[1]) {
    New_2D_Array<double_complex, int, long>(&dKEmbedEig, NumParam,
                                            LenKEmbedEig);
    for (i = 0; i < NumParam; i++) {
      mToepSolver.Planning(dK[i], dKEmbedEig[i], NULL);
    }
  }
  if (bNeedPartials[2]) {
    New_3D_Array_Sym2D<double_complex, int, long>(&ddKEmbedEig, NumParam,
                                                  LenKEmbedEig);
    for (i = 0; i < NumParam; i++) {
      for (j = 0; j <= i; j++) {
        mToepSolver.Planning(ddK[i][j], ddKEmbedEig[i][j], NULL);
      }
    }
  }
}

//--------------------------------------------------------------------------
template<class Kernel>
void RegularGridCovarMat<Kernel>::
KMult(RegularGridData &x, RegularGridData &y) {
  mToepSolver.MatMult(KEmbedEig, x.GetValPtr(), y.GetValPtr());
}

//--------------------------------------------------------------------------
template<class Kernel>
void RegularGridCovarMat<Kernel>::
dKMult(int i, RegularGridData &x, RegularGridData &y) {
  mToepSolver.MatMult(dKEmbedEig[i], x.GetValPtr(), y.GetValPtr());
}

//--------------------------------------------------------------------------
template<class Kernel>
void RegularGridCovarMat<Kernel>::
ddKMult(int i, int j, RegularGridData &x, RegularGridData &y) {
  if (i < j) {
    Swap<int>(i,j);
  }
  mToepSolver.MatMult(ddKEmbedEig[i][j], x.GetValPtr(), y.GetValPtr());
}

//--------------------------------------------------------------------------
template<class Kernel>
void RegularGridCovarMat<Kernel>::
KSolve(RegularGridData &b, RegularGridData &x) {
  x.SetData(0.0);
  mToepSolver.MatSolve(KEmbedEig, PrecondEig, b.GetValPtr(), x.GetValPtr(),
                       mToepSolverParam.RTol, mToepSolverParam.MaxIt);
}

//--------------------------------------------------------------------------
template<class Kernel>
RegularGridCovarVec<Kernel>::
RegularGridCovarVec() : RegularGridData() {
  Init();
}

//--------------------------------------------------------------------------
template<class Kernel>
void RegularGridCovarVec<Kernel>::
Init(void) {
}

//--------------------------------------------------------------------------
template<class Kernel>
RegularGridCovarVec<Kernel>::
RegularGridCovarVec(MPI_Comm &comm) : RegularGridData(comm) {
  Init();
}

//--------------------------------------------------------------------------
template<class Kernel>
RegularGridCovarVec<Kernel>::
RegularGridCovarVec(const RegularGridCovarVec &V) : RegularGridData(V) {
  DeepCopy(V);
}

//--------------------------------------------------------------------------
template<class Kernel>
void RegularGridCovarVec<Kernel>::
DeepCopy(const RegularGridCovarVec &V) {
}

//--------------------------------------------------------------------------
template<class Kernel>
RegularGridCovarVec<Kernel>& RegularGridCovarVec<Kernel>::
operator= (const RegularGridCovarVec &V) {
  if (this == &V) {
    return *this;
  }
  RegularGridData::operator=(V);
  DeAllocMem();
  DeepCopy(V);
  return *this;
}

//--------------------------------------------------------------------------
template<class Kernel>
void RegularGridCovarVec<Kernel>::
DeAllocMem(void) {
}

//--------------------------------------------------------------------------
template<class Kernel>
RegularGridCovarVec<Kernel>::
~RegularGridCovarVec() {
  DeAllocMem();
}

//--------------------------------------------------------------------------
template<class Kernel>
void RegularGridCovarVec<Kernel>::
SetVal(Kernel &kernel, double *coord) {
  Delete_1D_Array<double>(&Val);
  Len = CalLen();
  New_1D_Array<double, long>(&Val, Len);

  // Compute grid_dist
  int i,k;
  double **grid_dist = NULL;
  New_2D_Array<double, int>(&grid_dist, NDim, Dim);
  double spacing = 0.0;
  for (i = 0; i < NDim; i++) {
    switch (mType) {
    case FIXED_DOMAIN:
      spacing = mSpec.Range[i]/(Dim[i]-1);
      break;
    case FIXED_SPACING:
      spacing = mSpec.Delta[i];
      break;
    case UNKNOWN_GRID_TYPE:
      throw std::string("RegularGridCovarVec::SetVal: Unknown grid type!\n");
      break;
    }
    for (k = 0; k < Dim[i]; k++) {
      grid_dist[i][k] = spacing * k;
    }
  }

  // Compute Val
  double *x = NULL;
  New_1D_Array<double, int>(&x, NDim);
  long *mIdx = NULL;
  New_1D_Array<long, int>(&mIdx, NDim);
  Set_Array<long, int>(mIdx, DimStart, NDim);
  long *upper = NULL;
  New_1D_Array<long, int>(&upper, NDim);
  for (i = 0; i < NDim; i++) {
    upper[i] = DimStart[i] + DimLen[i];
  }
  MultiIdx<long> mMultiIdx(NDim, DimStart, upper, DimStart);

  for (k = 0; k < Len; k++) {
    for (i = 0; i < NDim; i++) {
      x[i] = grid_dist[i][mIdx[i]] - coord[i];
    }
    Val[k] = kernel.GetValues(x, NULL, NULL);
    mMultiIdx.Next(mIdx);
  }
}

#endif

// Author: J. Chen
// $Date: 2013/03/28 15:50:43$
