#ifndef _SCORE_
#define _SCORE_
// Computes major quantities of MLE.

#include <unistd.h>
#include "Utility.hpp"

//--------------------------------------------------------------------------
// Scores:
//
//   F[i] = z' * inv(K) * dK_i * inv(K) * z / 2 - tr( inv(K) * dK_i ) / 2
//
// Jacobian:
//
//   J[i][j] = - z' * inv(K) * dK_i * inv(K) * dK_j * inv(K) * z
//             + z' * inv(K) * ddK_{ij} * inv(K) * z / 2
//             + tr( inv(K) * dK_j * inv(K) * dK_i ) / 2
//             - tr( inv(K) * ddK_{ij} ) / 2
//
// Fisher information matrix:
//
//   I[i][j] = tr( inv(K) * dK_i * inv(K) * dK_j ) / 2
//
// Variance of standard MLE:
//
//   invI = inv(I);
//
// Variance of MLE with SAA:
//
//   JJ[i][j] = cov( u' * inv(K) * dK_i * u, u' * inv(K) * dK_j * u )
//   invG = inv(I) + inv(I) * JJ * inv(I) / 4N
//
// Sandwiched information matrix:
//
//   G = inv(invG)
//--------------------------------------------------------------------------
template<class TData, class CovarMat>
class ApproxScoreFunc {

public:

  // Constructor, destructor
  ApproxScoreFunc();
  ApproxScoreFunc(const ApproxScoreFunc &A);
  ApproxScoreFunc& operator= (const ApproxScoreFunc &A);
  ~ApproxScoreFunc();

  // Set
  void SetGPSample(TData *gp_sample);
  void SetCovarMat(CovarMat *covar_mat);
  void SetNumRandVec(int num_rand_vec);

  // Major computational component
  void GetScores(double *score, bool need_jacob = false);
  void GetJacobian(double **jacob);
  void GetVariance(double **fisher, double **var_mle, // inverse of fisher
                   double **sandwich, double **var_mle_saa // inverse of sand
                   );

private:

  // Needed by constructor and destructor
  void DeepCopy(const ApproxScoreFunc &A);
  void DeAllocMem(void);
  void DeAllocFJIG(void);

  // Needed by GetScores() and GetJacobian()
  void ComputeScoresAndPerhapsJacobian(bool need_jacob);

  // Matrix mult and inv
  void MatMat(double **A, double **B, double **AB, int n);
  void InvMat(double **A, double **invA, int n);

  TData *mGPSample;             // GP sample. Pass in pointer from outside
  CovarMat *mCovarMat;          // Covariance matrix
  int NumRandVec;               // Number of random vectors
  unsigned int *Seed;           // A seed for each random vector

  bool bFCurrent;               // Is F current?
  bool bJCurrent;               // Is J current?
  bool bNeedJacobian;           // Need J?

  int LenTheta;                 // Length of Theta
  double *F;                    // Scores
  double **J;                   // Jacobian
  double **I;                   // Fisher information matrix
  double **invI;                // Variance of standard MLE
  double **G;                   // Sandwiched information matrix
  double **invG;                // Variance of MLE with SAA
  double **JJ;

};

//--------------------------------------------------------------------------
template<class TData, class CovarMat>
ApproxScoreFunc<TData, CovarMat>::
ApproxScoreFunc() {
  mGPSample = NULL;
  mCovarMat = NULL;
  NumRandVec = 0;
  Seed = NULL;
  bFCurrent = false;
  bJCurrent = false;
  bNeedJacobian = false;
  LenTheta = 0;
  F = NULL;
  J = NULL;
  I = NULL;
  invI = NULL;
  G = NULL;
  invG = NULL;
  JJ = NULL;
}

//--------------------------------------------------------------------------
template<class TData, class CovarMat>
ApproxScoreFunc<TData, CovarMat>::
ApproxScoreFunc(const ApproxScoreFunc &A) {
  Seed = NULL;
  F = NULL;
  J = NULL;
  I = NULL;
  invI = NULL;
  G = NULL;
  invG = NULL;
  JJ = NULL;
  DeepCopy(A);
}

//--------------------------------------------------------------------------
template<class TData, class CovarMat>
void ApproxScoreFunc<TData, CovarMat>::
DeepCopy(const ApproxScoreFunc &A) {
  int i;
  mGPSample = A.mGPSample;
  mCovarMat = A.mCovarMat;
  NumRandVec = A.NumRandVec;
  New_1D_Array<unsigned int, int>(&Seed, NumRandVec);
  Set_Array<unsigned int, int>(Seed, A.Seed, NumRandVec);
  bFCurrent = A.bFCurrent;
  bJCurrent = A.bJCurrent;
  bNeedJacobian = A.NeedJacobian;
  LenTheta = A.LenTheta;
  New_1D_Array<double, int>(&F, LenTheta);
  Set_Array<double, int>(F, A.F, LenTheta);
  New_2D_Array<double, int, int>(&J, LenTheta, LenTheta);
  for (i = 0; i < LenTheta; i++) {
    Set_Array<double, int>(J[i], A.J[i], LenTheta);
  }
  New_2D_Array<double, int, int>(&I, LenTheta, LenTheta);
  for (i = 0; i < LenTheta; i++) {
    Set_Array<double, int>(I[i], A.I[i], LenTheta);
  }
  New_2D_Array<double, int, int>(&invI, LenTheta, LenTheta);
  for (i = 0; i < LenTheta; i++) {
    Set_Array<double, int>(invI[i], A.invI[i], LenTheta);
  }
  New_2D_Array<double, int, int>(&G, LenTheta, LenTheta);
  for (i = 0; i < LenTheta; i++) {
    Set_Array<double, int>(G[i], A.G[i], LenTheta);
  }
  New_2D_Array<double, int, int>(&invG, LenTheta, LenTheta);
  for (i = 0; i < LenTheta; i++) {
    Set_Array<double, int>(invG[i], A.invG[i], LenTheta);
  }
  New_2D_Array<double, int, int>(&JJ, LenTheta, LenTheta);
  for (i = 0; i < LenTheta; i++) {
    Set_Array<double, int>(JJ[i], A.JJ[i], LenTheta);
  }
}

//--------------------------------------------------------------------------
template<class TData, class CovarMat>
ApproxScoreFunc<TData, CovarMat>&
ApproxScoreFunc<TData, CovarMat>::
operator= (const ApproxScoreFunc &A) {
  if (this == &A) {
    return *this;
  }
  DeAllocMem();
  DeepCopy(A);
  return *this;
}

//--------------------------------------------------------------------------
template<class TData, class CovarMat>
void ApproxScoreFunc<TData, CovarMat>::
DeAllocMem(void) {
  Delete_1D_Array<unsigned int>(&Seed);
  DeAllocFJIG();
}

//--------------------------------------------------------------------------
template<class TData, class CovarMat>
void ApproxScoreFunc<TData, CovarMat>::
DeAllocFJIG(void) {
  Delete_1D_Array<double>(&F);
  Delete_2D_Array<double, int>(&J, LenTheta);
  Delete_2D_Array<double, int>(&I, LenTheta);
  Delete_2D_Array<double, int>(&invI, LenTheta);
  Delete_2D_Array<double, int>(&G, LenTheta);
  Delete_2D_Array<double, int>(&invG, LenTheta);
  Delete_2D_Array<double, int>(&JJ, LenTheta);
}

//--------------------------------------------------------------------------
template<class TData, class CovarMat>
ApproxScoreFunc<TData, CovarMat>::
~ApproxScoreFunc() {
  DeAllocMem();
}

//--------------------------------------------------------------------------
template<class TData, class CovarMat>
void ApproxScoreFunc<TData, CovarMat>::
SetGPSample(TData *gp_sample) {
  mGPSample = gp_sample;
}

//--------------------------------------------------------------------------
template<class TData, class CovarMat>
void ApproxScoreFunc<TData, CovarMat>::
SetCovarMat(CovarMat *covar_mat) {
  mCovarMat = covar_mat;
  bFCurrent = false;
  bJCurrent = false;
}

//--------------------------------------------------------------------------
template<class TData, class CovarMat>
void ApproxScoreFunc<TData, CovarMat>::
SetNumRandVec(int num_rand_vec) {
  NumRandVec = num_rand_vec;
  // Compute seeds
  Delete_1D_Array<unsigned int>(&Seed);
  New_1D_Array<unsigned int, int>(&Seed, NumRandVec);
  int i;
  long pid;
  srand(time(NULL));
  for (i = 0; i < NumRandVec; i++) {
    pid = getpid();
    Seed[i] = abs(((rand()*181)*((pid-83)*359))%104729);
  }
}

//--------------------------------------------------------------------------
template<class TData, class CovarMat>
void ApproxScoreFunc<TData, CovarMat>::
GetScores(double *score, bool need_jacob) {
  if (!score) {
    return;
  }
  bNeedJacobian = need_jacob;
  if (!bFCurrent) {
    ComputeScoresAndPerhapsJacobian(bNeedJacobian);
  }
  Set_Array<double, int>(score, F, LenTheta);
}

//--------------------------------------------------------------------------
template<class TData, class CovarMat>
void ApproxScoreFunc<TData, CovarMat>::
GetJacobian(double **jacob) {
  if (!jacob) {
    return;
  }
  if (!bJCurrent) {
    ComputeScoresAndPerhapsJacobian(true);
  }
  int i;
  for (i = 0; i < LenTheta; i++) {
    Set_Array<double, int>(jacob[i], J[i], LenTheta);
  }
}

//--------------------------------------------------------------------------
template<class TData, class CovarMat>
void ApproxScoreFunc<TData, CovarMat>::
GetVariance(double **fisher, double **var_mle,
            double **sandwich, double **var_mle_saa) {
  if (!fisher || !sandwich || !var_mle || !var_mle_saa) {
    return;
  }
  if (!bJCurrent) {
    ComputeScoresAndPerhapsJacobian(true);
  }
  int i;
  for (i = 0; i < LenTheta; i++) {
    Set_Array<double, int>(fisher[i], I[i], LenTheta);
    Set_Array<double, int>(var_mle[i], invI[i], LenTheta);
    Set_Array<double, int>(sandwich[i], G[i], LenTheta);
    Set_Array<double, int>(var_mle_saa[i], invG[i], LenTheta);
  }
}

//--------------------------------------------------------------------------
template<class TData, class CovarMat>
void ApproxScoreFunc<TData, CovarMat>::
ComputeScoresAndPerhapsJacobian(bool need_jacob) {
  int i, j, k;
  bNeedJacobian = need_jacob;

  DeAllocFJIG();
  LenTheta = mCovarMat->GetNumParam();
  New_1D_Array<double, int>(&F, LenTheta);
  New_2D_Array<double, int, int>(&J, LenTheta, LenTheta);
  New_2D_Array<double, int, int>(&I, LenTheta, LenTheta);
  New_2D_Array<double, int, int>(&invI, LenTheta, LenTheta);
  New_2D_Array<double, int, int>(&G, LenTheta, LenTheta);
  New_2D_Array<double, int, int>(&invG, LenTheta, LenTheta);
  New_2D_Array<double, int, int>(&JJ, LenTheta, LenTheta);

  // A bunch of temporary variables
  TData u = *mGPSample;
  TData invKz = u;
  TData invKu = u;
  TData *dKinvKz = NULL;
  New_1D_Array<TData, int>(&dKinvKz, LenTheta);
  TData *dKu = NULL;
  New_1D_Array<TData, int>(&dKu, LenTheta);
  // The followings are not used when need_jacob is false
  TData *invKdKinvKz = NULL;
  New_1D_Array<TData, int>(&invKdKinvKz, LenTheta);
  TData **ddKinvKz = NULL;
  New_2D_Array_Sym<TData, int>(&ddKinvKz, LenTheta);
  TData *dKinvKu = NULL;
  New_1D_Array<TData, int>(&dKinvKu, LenTheta);
  TData *invKdKu = NULL;
  New_1D_Array<TData, int>(&invKdKu, LenTheta);
  TData **ddKu = NULL;
  New_2D_Array_Sym<TData, int>(&ddKu, LenTheta);
  double **uinvKdKu = NULL;
  New_2D_Array<double, int, int>(&uinvKdKu, LenTheta, NumRandVec);
  double **tmpJ = NULL;
  New_2D_Array<double, int, int>(&tmpJ, LenTheta, LenTheta);
  double **tmpJ2 = NULL;
  New_2D_Array<double, int, int>(&tmpJ2, LenTheta, LenTheta);

  // Cumbersome initialization of the variables
  // Only use the Comm and Len, but not Val, of u
  for (i = 0; i < LenTheta; i++) {
    dKinvKz[i] = u;
    dKu[i] = u;
  }
  if (bNeedJacobian) {
    for (i = 0; i < LenTheta; i++) {
      invKdKinvKz[i] = u;
      dKinvKu[i] = u;
      invKdKu[i] = u;
    }
    for (i = 0; i < LenTheta; i++) {
      for (j = 0; j <= i; j++) {
        ddKinvKz[i][j] = u;
        ddKu[i][j] = u;
      }
    }
  }

  // The calculation
  for (k = 0; k < NumRandVec; k++) {
    // Generate random vector
    u.SetSymBernoulli(Seed[k]);

    // invKu = inv(K) * u
    mCovarMat->KSolve(u, invKu);

    // dKu[i] = dK_i * u
    for (i = 0; i < LenTheta; i++) {
      mCovarMat->dKMult(i, u, dKu[i]);
    }

    // accumulate -tr( inv(K) * dK_i )
    for (i = 0; i < LenTheta; i++) {
      F[i] -= invKu.InProd(dKu[i]);
    }

    if (bNeedJacobian) {
      // dKinvKu[i] = dK_i * inv(K) * u
      for (i = 0; i < LenTheta; i++) {
        mCovarMat->dKMult(i, invKu, dKinvKu[i]);
      }

      // invKdKu[i] = inv(K) * dK_i * u
      for (i = 0; i < LenTheta; i++) {
        mCovarMat->KSolve(dKu[i], invKdKu[i]);
      }

      // uinvKdKu[i][k] = u' * inv(K) * dK_i * u
      for (i = 0; i < LenTheta; i++) {
        uinvKdKu[i][k] = u.InProd(invKdKu[i]);
      }

      // accumulate tr( inv(K) * dK_j * inv(K) * dK_i )
      for (i = 0; i < LenTheta; i++) {
        for (j = 0; j < LenTheta; j++) {
          double tmp = dKinvKu[j].InProd(invKdKu[i]);
          J[i][j] += tmp;
          I[i][j] += tmp;
        }
      }

      // ddKu[i][j] = ddK_{ij} * u
      for (i = 0; i < LenTheta; i++) {
        for (j = 0; j <= i; j++) {
          mCovarMat->ddKMult(i, j, u, ddKu[i][j]);
        }
      }

      // accumulate -tr( inv(K) * ddK_{ij} )
      for (i = 0; i < LenTheta; i++) {
        for (j = 0; j <= i; j++) {
          double tmp = invKu.InProd(ddKu[i][j]);
          J[i][j] -= tmp;
          if (j != i) {
            J[j][i] -= tmp;
          }
        }
      }
    }
  }

  // complete -tr( inv(K) * dK_i ) / 2
  for (i = 0; i < LenTheta; i++) {
    F[i] /= (NumRandVec*2);
  }

  if (bNeedJacobian) {
    // complete tr( inv(K) * dK_j * inv(K) * dK_i ) / 2
    //          - tr( inv(K) * ddK_{ij} ) / 2
    for (i = 0; i < LenTheta; i++) {
      for (j = 0; j < LenTheta; j++) {
        J[i][j] /= (NumRandVec*2);
      }
    }

    // complete I[i][j]
    for (i = 0; i < LenTheta; i++) {
      for (j = 0; j <= i; j++) {
        I[i][j] = (I[i][j] + I[j][i]) / (NumRandVec*4);
        I[j][i] = I[i][j];
      }
    }

    // complete JJ[i][j]
    for (i = 0; i < LenTheta; i++) {
      double ave = 0.0;
      for (k = 0; k < NumRandVec; k++) {
        ave += uinvKdKu[i][k];
      }
      ave /= NumRandVec;
      for (k = 0; k < NumRandVec; k++) {
        uinvKdKu[i][k] -= ave;
      }
    }
    for (i = 0; i < LenTheta; i++) {
      for (j = 0; j < LenTheta; j++) {
        for (k = 0; k < NumRandVec; k++) {
          JJ[i][j] += uinvKdKu[i][k] * uinvKdKu[j][k];
        }
        if (NumRandVec > 1) {
          JJ[i][j] /= (NumRandVec-1);
        }
        else {
          JJ[i][j] = 0.0;
        }
      }
    }
  }

  // invKz = inv(K) * z
  mCovarMat->KSolve(*mGPSample, invKz);

  // dKinvKz[i] = dK_i * inv(K) * z
  for (i = 0; i < LenTheta; i++) {
    mCovarMat->dKMult(i, invKz, dKinvKz[i]);
  }

  // complete F[i]
  for (i = 0; i < LenTheta; i++) {
    F[i] += invKz.InProd(dKinvKz[i])/2;
  }

  if (bNeedJacobian) {
    // invKdKinvKz[i] = inv(K) * dK_i * inv(K) * z
    for (i = 0; i < LenTheta; i++) {
      mCovarMat->KSolve(dKinvKz[i], invKdKinvKz[i]);
    }

    // ddKinvKz[i][j] = ddK_{ij} * inv(K) * z
    for (i = 0; i < LenTheta; i++) {
      for (j = 0; j <= i; j++) {
        mCovarMat->ddKMult(i, j, invKz, ddKinvKz[i][j]);
      }
    }

    // complete J[i][j]
    for (i = 0; i < LenTheta; i++) {
      for (j = 0; j <= i; j++) {
        double tmp = - dKinvKz[i].InProd(invKdKinvKz[j])
          + invKz.InProd(ddKinvKz[i][j])/2;
        J[i][j] += tmp;
        if (j != i) {
          J[j][i] += tmp;
        }
      }
    }

    // complete invI
    InvMat(I, invI, LenTheta);

    // complete invG
    MatMat(invI, JJ, tmpJ, LenTheta);
    MatMat(tmpJ, invI, tmpJ2, LenTheta);
    for (i = 0; i < LenTheta; i++) {
      for (j = 0; j < LenTheta; j++) {
        invG[i][j] = invI[i][j] + tmpJ2[i][j] / (NumRandVec*4);
      }
    }

    // complete G
    InvMat(invG, G, LenTheta);
  }

  // Cleanup
  Delete_1D_Array<TData>(&dKinvKz);
  Delete_1D_Array<TData>(&dKu);
  Delete_1D_Array<TData>(&invKdKinvKz);
  Delete_2D_Array<TData, int>(&ddKinvKz, LenTheta);
  Delete_1D_Array<TData>(&dKinvKu);
  Delete_1D_Array<TData>(&invKdKu);
  Delete_2D_Array<TData, int>(&ddKu, LenTheta);
  Delete_2D_Array<double, int>(&uinvKdKu, LenTheta);
  Delete_2D_Array<double, int>(&tmpJ, LenTheta);
  Delete_2D_Array<double, int>(&tmpJ2, LenTheta);

  // Set bFCurrent and bJCurrent
  bFCurrent = true;
  bJCurrent = bNeedJacobian;
}

//--------------------------------------------------------------------------
template<class TData, class CovarMat>
void ApproxScoreFunc<TData, CovarMat>::
MatMat(double **A, double **B, double **AB, int n) {
  int i, j, k;
  for (i = 0; i < n; i++) {
    for (j = 0; j < n; j++) {
      AB[i][j] = 0.0;
      for (k = 0; k < n; k++) {
        AB[i][j] += A[i][k] * B[k][j];
      }
    }
  }
}

//--------------------------------------------------------------------------
template<class TData, class CovarMat>
void ApproxScoreFunc<TData, CovarMat>::
InvMat(double **A, double **invA, int n) {
  double *M = NULL;
  New_1D_Array<double, int>(&M, n*n);
  int *ipiv = NULL;
  New_1D_Array<int, int>(&ipiv, n);
  int i, j, k;
  k = 0;
  for (i = 0; i < n; i++) {
    for (j = 0; j < n; j++) {
      M[k++] = A[i][j];
    }
  }
  int info;
  dgetrf_(&n, &n, M, &n, ipiv, &info);
  int lwork = n*n;
  double *work = NULL;
  New_1D_Array<double, int>(&work, lwork);
  dgetri_(&n, M, &n, ipiv, work, &lwork, &info);
  k = 0;
  for (i = 0; i < n; i++) {
    for (j = 0; j < n; j++) {
      invA[i][j] = M[k++];
    }
  }
  Delete_1D_Array<double>(&M);
  Delete_1D_Array<int>(&ipiv);
}

#endif

// Author: J. Chen
// $Date: 2013/03/28 15:50:43$
