/*
 * toep_solver_v1.0.c
 * MPI_CG version 1
 *
 * Author: Tom Li
 * Created on: July, 19, 2012
 */
/*
 * Edited by JC:
 * 1. Append _v10 to all the data structures and functions
 * 2. Add code in toep_solver_plan_v10 to prevent computing lam_m if NULL
 */

#include "mpi.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <complex.h>
#include <math.h>
#include <time.h>
#include <fftw3-mpi.h>

#include "toep_solver_v1.0.h"


#define max(a,b) \
   ({ __typeof__ (a) _a = (a); \
       __typeof__ (b) _b = (b); \
     _a > _b ? _a : _b; })




/*
 * System interface. This are the functions users should call.
 */
/* 
 * Systen initialization 
 * Input:
 *      mpiComm     : global communicator
 *      rank        : number of levels in the system
 *      ns          : dimensionality array
 * Output:
 *      TI          : system configuration
 */
void toep_solver_init_v10(MPI_Comm mpiComm, int rank, int * ns, struct ToeplitzSolverInfo_v10 * TI)
{
    /*
     * MPI communicator management
     */
    MPI_Comm_size(mpiComm, &TI->numprocs);
    MPI_Comm_rank(mpiComm, &TI->myid);

    TI->mpiComm = mpiComm;

    TI->mpiRequests = (MPI_Request *) malloc(sizeof(MPI_Request) * 2 * TI->numprocs);
    TI->mpiStatusList = (MPI_Status *) malloc(sizeof(MPI_Status) * 2 * TI->numprocs);


    /*
     * Calculate lvlGrain, pwrLvlGrain, n, and N.
     */
    TI->lvlGrain = (long *) malloc(sizeof(long) * rank);
    TI->pwrLvlGrain = (long *) malloc(sizeof(long) * rank);
    TI->lvlGrain[rank-1] = 1;
    TI->pwrLvlGrain[rank-1] = 1;
    for (int i = rank-2; i >=0; i --)
    {
        TI->lvlGrain[i] = TI->lvlGrain[i+1] * ns[i+1];
        TI->pwrLvlGrain[i] = TI->pwrLvlGrain[i + 1] * ns[i+1] * 2;
    }
    TI->n = ns[0] * TI->lvlGrain[0];
    TI->N = ns[0] * 2 * TI->pwrLvlGrain[0];

    TI->rank = rank;

    TI->ns = (int *)malloc(sizeof(int) * rank);
    for (int i=0; i < rank; i ++) TI->ns[i] = ns[i];

    TI->ns2 = (int *) malloc ( sizeof(int) * rank);
    for (int i=0; i < rank; i ++) TI->ns2[i] = ns[i] * 2;



    /*
     * dimension sizes
     */
    TI->n0Info.size = TI->ns[0];
    TI->n1Info.size = TI->lvlGrain[0];
    TI->pwrN1Info.size = TI->pwrLvlGrain[0];


    /*
     * Calculate dimension partition and accumulation
     */
    long blksize;
    TI->n0Info.partitions = (long *) malloc(sizeof(long) * TI->numprocs);
    TI->n1Info.partitions = (long *) malloc(sizeof(long) * TI->numprocs);
    TI->pwrN1Info.partitions = (long *) malloc(sizeof(long) * TI->numprocs);

    TI->n0Info.accumulations = (long *) malloc(sizeof(long) * TI->numprocs);
    TI->n1Info.accumulations = (long *) malloc(sizeof(long) * TI->numprocs);
    TI->pwrN1Info.accumulations = (long *) malloc(sizeof(long) * TI->numprocs);

    /* partitions for n0 */
    blksize = ceil(TI->ns[0] * 1.0 / TI->numprocs);
    for (long i = 0; i < TI->numprocs; i ++)
    {
        if (i < TI->ns[0] / blksize)
            TI->n0Info.partitions[i] = blksize;
        else
            if (i == TI->ns[0] / blksize)
                TI->n0Info.partitions[i] = TI->ns[0] - blksize * (TI->ns[0] / blksize);
            else
                TI->n0Info.partitions[i] = 0;
    }

    /* partitions for n1 */
    blksize = ceil(TI->lvlGrain[0] * 1.0 / TI->numprocs);
    for (long i = 0; i < TI->numprocs; i ++)
    {
        if (i < TI->lvlGrain[0] / blksize)
            TI->n1Info.partitions[i] = blksize;
        else
            if (i == TI->lvlGrain[0] / blksize)
                TI->n1Info.partitions[i] = TI->lvlGrain[0] - blksize * (TI->lvlGrain[0] / blksize);
            else
                TI->n1Info.partitions[i] = 0;
    }

    /* partitions for pwrN1 */
    blksize = ceil(TI->pwrLvlGrain[0] * 1.0 / TI->numprocs);
    for (long i = 0; i < TI->numprocs; i ++)
    {
        if (i < TI->pwrLvlGrain[0] / blksize)
            TI->pwrN1Info.partitions[i] = blksize;
        else
            if (i == TI->pwrLvlGrain[0] / blksize)
                TI->pwrN1Info.partitions[i] = TI->pwrLvlGrain[0] - blksize * (TI->pwrLvlGrain[0] / blksize);
            else
                TI->pwrN1Info.partitions[i] = 0;
    }

    /* accumulations */
    TI->n0Info.accumulations[0] = 0;
    TI->n1Info.accumulations[0] = 0;
    TI->pwrN1Info.accumulations[0] = 0;
    for (long i =1; i < TI->numprocs; i ++)
    {
        TI->n0Info.accumulations[i] = TI->n0Info.accumulations[i-1] + TI->n0Info.partitions[i-1];
        TI->n1Info.accumulations[i] = TI->n1Info.accumulations[i-1] + TI->n1Info.partitions[i-1];
        TI->pwrN1Info.accumulations[i] = TI->pwrN1Info.accumulations[i-1] + TI->pwrN1Info.partitions[i-1];
    }


    /* variables for housekeeping and memory allocation */
    TI->n0Info.local = TI->n0Info.partitions[TI->myid];
    TI->n1Info.local = TI->n1Info.partitions[TI->myid];
    TI->pwrN1Info.local = TI->pwrN1Info.partitions[TI->myid];

    TI->n0Info.localStart = TI->n0Info.accumulations[TI->myid];

    TI->allocLocal = max(TI->n0Info.local * TI->n1Info.size, TI->n1Info.local * TI->n0Info.size);
    TI->pwrAllocLocal = max(TI->n0Info.local * TI->pwrN1Info.size, TI->pwrN1Info.local * TI->n0Info.size);


    /*
     * Transposition variables
     */
    TI->sendBuf = (complex *) malloc ( sizeof(complex) * (TI->pwrAllocLocal + TI->numprocs * 2));
    TI->sendCnts = (int * ) malloc (sizeof(int) * TI->numprocs);
    TI->sDispls = (int *) malloc (sizeof(int) * TI->numprocs);

    TI->recvBuf = (complex *) malloc ( sizeof(complex) * (TI->pwrAllocLocal + TI->numprocs * 2));
    TI->recvCnts = (int * ) malloc (sizeof(int) * TI->numprocs);
    TI->rDispls = (int *) malloc (sizeof(int) * TI->numprocs);

    /*
     * FFTW plans and buffers
     */
    TI->fftBuf = fftw_alloc_complex( max(TI->pwrN1Info.size, TI->ns2[0]));
    TI->compBuf = fftw_alloc_complex(TI->pwrAllocLocal * 2);
    TI->transBuf = fftw_alloc_complex(TI->pwrAllocLocal);

    TI->tyFFTRow = fftw_plan_dft(rank-1, TI->ns2 + 1, TI->fftBuf, TI->fftBuf, FFTW_FORWARD, FFTW_ESTIMATE);
    TI->tyIFFTRow = fftw_plan_dft(rank-1, TI->ns2 + 1, TI->fftBuf, TI->fftBuf, FFTW_BACKWARD, FFTW_ESTIMATE);

    TI->tyFFTCol = fftw_plan_dft(1, &TI->ns2[0] , TI->fftBuf, TI->fftBuf, FFTW_FORWARD, FFTW_ESTIMATE);
    TI->tyIFFTCol = fftw_plan_dft(1, &TI->ns2[0], TI->fftBuf, TI->fftBuf, FFTW_BACKWARD, FFTW_ESTIMATE);

    TI->myFFTRow = fftw_plan_dft(rank-1, TI->ns + 1, TI->fftBuf, TI->fftBuf, FFTW_FORWARD, FFTW_ESTIMATE);
    TI->myIFFTRow = fftw_plan_dft(rank-1, TI->ns + 1, TI->fftBuf, TI->fftBuf, FFTW_BACKWARD, FFTW_ESTIMATE);

    TI->myFFTCol = fftw_plan_dft(1, &TI->ns[0] , TI->fftBuf, TI->fftBuf, FFTW_FORWARD, FFTW_ESTIMATE);
    TI->myIFFTCol = fftw_plan_dft(1, &TI->ns[0], TI->fftBuf, TI->fftBuf, FFTW_BACKWARD, FFTW_ESTIMATE);

}

/*
 * Planning function. 
 * Input:
 *      t           : a Toeplitz vector. The first column of a Toeplitz matrix
 *      TI          : system configuration
 * OutPut:
 *      lam_t       : the eigenvalues of the embedded Toeplitz matrix
 *      lam_m       : the eigenvalues of the preconditioner matrix
 */
void toep_solver_plan_v10(complex * t, struct ToeplitzSolverInfo_v10 * TI, complex * lam_t, complex * lam_m)
{
    /*
     * SECTION: Compute lam_t, the eigenvalue of t
     */
    /* row-wise embed. Need to make sure TI->rank > 1 */
    for (long i=0; i < TI->n0Info.local; i ++)
    {
        toepEmbd_v10(t + i * TI->lvlGrain[0], TI->fftBuf, TI->rank - 1, TI->ns + 1);
        fftw_execute(TI->tyFFTRow);
        memcpy(TI->compBuf + i * TI->pwrLvlGrain[0], TI->fftBuf, sizeof(fftw_complex) * TI->pwrLvlGrain[0]);
    }

    /* transposition */
    transpose_v10(TI->compBuf, TI->pwrN1Info, TI->n0Info, TI, TI->transBuf);

    /* column-wise embed. */
    for (long i = 0; i < TI->pwrN1Info.local; i ++)
    {
        toepEmbd_v10(TI->transBuf + i * TI->ns[0], TI->fftBuf, 1, &TI->ns[0]);
        fftw_execute(TI->tyFFTCol);
        memcpy(lam_t + i * TI->ns2[0], TI->fftBuf, sizeof(fftw_complex) * TI->ns2[0]);
    }

    if (!lam_m) {
      return;
    }

    /*
     * SECTION: Calculate m
     */
    memcpy(lam_m, t, sizeof(complex) * TI->n0Info.local * TI->lvlGrain[0]);

    /* row-wise preconditioner compute */
    for (long i=0; i < TI->n0Info.local; i++)
        toepPreVec_v10(lam_m + i * TI->lvlGrain[0], TI->rank - 1, TI->ns + 1);


    /* transposition */
    transpose_v10(lam_m, TI->n1Info, TI->n0Info, TI, lam_m);


    /* col-wise preconditioner computation */
    for (long i = 0; i < TI->n1Info.local; i ++)
        toepPreVec_v10(lam_m + i * TI->ns[0], 1, &TI->ns[0]);


    /* transpose back */
    transpose_v10(lam_m, TI->n0Info, TI->n1Info, TI, lam_m);

    /*
     * Calculate lam_m
     */

    /* row fft for m */
    for (long i = 0; i < TI->n0Info.local; i ++)
    {
        memcpy(TI->fftBuf, lam_m + i * TI->lvlGrain[0], sizeof(complex) * TI->lvlGrain[0]);
        fftw_execute(TI->myFFTRow);
        memcpy(lam_m + i * TI->lvlGrain[0], TI->fftBuf, sizeof(complex) * TI->lvlGrain[0]);
    }

    /* transposition */
    transpose_v10(lam_m, TI->n1Info, TI->n0Info, TI, lam_m);

    /* col fft for m */
    for (long i =0; i < TI->n1Info.local; i ++)
    {
        memcpy(TI->fftBuf, lam_m + i * TI->ns[0], sizeof(complex) * TI->ns[0]);
        fftw_execute(TI->myFFTCol);
        memcpy(lam_m + i * TI->ns[0], TI->fftBuf, sizeof(complex) * TI->ns[0]);
    }
}

/*
 * Solving function.
 * Input:
 *      lam_t       : the eigenvalues of the embedded Toeplitz matrix
 *      lam_m       : the eigenvalues of the preconditioner matrix
 *      b           : the RHS of the linear system Ax = b
 *      tol         : tolerance value
 *      max_iter    : maximum iteration
 *      TI          : system configuration
 *      x           : the initial guess of x, the unknown vector
 * Output:
 *      x           : solution of the linear ysstem Ax = b
 */
void toep_solver_solve_v10(complex * lam_t, complex * lam_m, complex *b,
                       double tol, int max_iter, struct ToeplitzSolverInfo_v10 * TI, complex * x)
{
    /*
     * Algorithm control parameters
     */
    double rho;                                 /* norm of the residual r */
    double gamma;                               /* norm of b */
    complex sigma;                              /* < r, z > */
    complex prev_sigma;                         /* previous sigma */
    complex tau;                                /* < T * p, p > */
    complex alpha;                              /* sigma / tau */
    complex beta=0;                             /* beta = sigma /  prev_sigma */

    complex * r;                                /* residual vector */
    complex * z;                                /* preconditioned vector z */
    complex * p;                                /* p */
    complex * tp;                               /* T * p */

    r = (complex *) malloc (sizeof(complex) * TI->allocLocal);
    z = (complex *) malloc (sizeof(complex) * TI->allocLocal);
    p = (complex *) malloc (sizeof(complex) * TI->allocLocal);
    tp = (complex *) malloc (sizeof(complex) * TI->allocLocal);

    /*
     * SECTION: Calculate y = Tx
     */
    toep_solver_matvec_v10(lam_t, x, TI, (complex *)TI->compBuf);

    /*
     * SECTION: r0 = b - y0
     */
    for (long i=0; i < TI->n0Info.local * TI->lvlGrain[0]; i ++)
        r[i] = b[i] - TI->compBuf[i];

    /*
     * SECTION: rho0 = norm(r0), gamma =  norm(b)
     */
    double localRho;
    localRho = creal(vectorCProd_v10(r, r, TI->n0Info.local * TI->lvlGrain[0]));

    MPI_Allreduce( &localRho, &rho, 1, MPI_DOUBLE, MPI_SUM, TI->mpiComm);

    rho = sqrt(rho);

    double localGamma;
    localGamma = creal(vectorCProd_v10(b, b, TI->n0Info.local * TI->lvlGrain[0]));

    MPI_Allreduce(&localGamma, &gamma, 1, MPI_DOUBLE, MPI_SUM, TI->mpiComm);

    gamma = sqrt(gamma);

    /*
     * SECTION: z0 = M^(-1) * r0
     */
    toep_solver_precon_v10(lam_m, r, TI, z);

    /*
     * sigma0 = < r0, z0 >
     */
    complex localSigma;
    localSigma = vectorCProd_v10(r, z, TI->n0Info.local * TI->lvlGrain[0]);

    MPI_Allreduce(&localSigma, &sigma, 1, MPI_DOUBLE_COMPLEX, MPI_SUM, TI->mpiComm);

    /*
     * p0 = z0
     */
    memcpy(p, z, sizeof(complex) * TI->n0Info.local * TI->lvlGrain[0]);


    /*
     * SECTION: For iter = 0, 1, ....max_iter do
     */
    for ( int iter = 0; iter < max_iter; iter ++)
    {
        /*
         * SECTION: tp = T * p
         */
        toep_solver_matvec_v10(lam_t, p, TI, tp);

        /*
         * tau = < tp, p >
         */
        complex localTau;
        localTau = vectorCProd_v10(tp, p, TI->n0Info.local * TI->lvlGrain[0]);

        MPI_Allreduce(&localTau, &tau, 1, MPI_DOUBLE_COMPLEX, MPI_SUM, TI->mpiComm);

        /*
         * alpha = sigma / tau
         */

        alpha = sigma / tau;

        /*
         * x = x + alpha * p
         */
        for(long i = 0; i < TI->n0Info.local * TI->lvlGrain[0]; i ++) x[i] += alpha * p[i];

        /*
         * r = r - alpha * tp
         */
        for (long i = 0; i < TI->n0Info.local * TI->lvlGrain[0]; i++) r[i] -= alpha * tp[i];

        /*
         * rho = norm(r)
         */
        localRho = creal(vectorCProd_v10(r, r, TI->n0Info.local * TI->lvlGrain[0]));

        MPI_Allreduce(&localRho, &rho, 1, MPI_DOUBLE, MPI_SUM, TI->mpiComm);

        rho = sqrt(rho);

#if DEBUG_PRINT_ENABLED
        if (TI->myid == DEBUG_ID)
        {
            printf("iteration: %d, rho: %g, tau: %g, sigma: %g, alpha: %g, beta: %g\n", iter, rho, creal(tau), creal(sigma), creal(alpha), creal(beta));
            printf("***********************************************\n\n");
        }
#endif

        /*if ( rho / gamma < tol) { break; }*/
        if ( rho / gamma < tol) {
          break;
        }

        /*
         * SECTION: z = M^(-1) * r
         */
        toep_solver_precon_v10(lam_m, r, TI, z);

        /*
         * sigma = < r, z >
         */

        prev_sigma = sigma;
        localSigma = vectorCProd_v10(r, z, TI->n0Info.local * TI->lvlGrain[0]);

        MPI_Allreduce(&localSigma, &sigma, 1, MPI_DOUBLE_COMPLEX, MPI_SUM, TI->mpiComm);

        /*
         * beta = sigma_(j+1) / sigma(j)
         */
        beta = sigma / prev_sigma;

        /*
         * p = z + beta * p
         */
        for (long i = 0; i < TI->n0Info.local * TI->lvlGrain[0]; i++)
            p[i] = z[i] + beta * p[i];
    }

    /*
     * Cleaning up
     */
    free(r);
    free(z);
    free(p);
    free(tp);
}

/*
 * System finalization. It is used to free up allocated memory spaces.
 * Input:
 *      TI          : system configuration
 */
void toep_solver_finalize_v10(struct ToeplitzSolverInfo_v10 * TI)
{
    free(TI->lvlGrain);
    free(TI->pwrLvlGrain);
    free(TI->ns);
    free(TI->ns2);

    free(TI->n0Info.partitions);
    free(TI->n1Info.partitions);
    free(TI->pwrN1Info.partitions);

    free(TI->n0Info.accumulations);
    free(TI->n1Info.accumulations);
    free(TI->pwrN1Info.accumulations);

    free(TI->sendBuf);
    free(TI->sendCnts);
    free(TI->sDispls);
    free(TI->recvBuf);
    free(TI->recvCnts);
    free(TI->rDispls);

    free(TI->mpiRequests);
    free(TI->mpiStatusList);

    fftw_free(TI->fftBuf);
    fftw_free(TI->compBuf);
    fftw_free(TI->transBuf);

    fftw_destroy_plan(TI->tyFFTRow);
    fftw_destroy_plan(TI->tyIFFTRow);
    fftw_destroy_plan(TI->tyFFTCol);
    fftw_destroy_plan(TI->tyIFFTCol);

    fftw_destroy_plan(TI->myFFTRow);
    fftw_destroy_plan(TI->myIFFTRow);
    fftw_destroy_plan(TI->myFFTCol);
    fftw_destroy_plan(TI->myIFFTCol);
}

/*
 * Toeplitz matrix-vector multiplication 
 * Input:
 *      lam_t       : eigenvalues of a Toeplitz matrix
 *      x           : input vector
 *      TI          : system configuration
 * Output:
 *      y           : output vector
 */
void toep_solver_matvec_v10(complex * lam_t, complex * x, struct ToeplitzSolverInfo_v10 * TI, complex * y)
{
    /*
     * Start of multiplication
     */
    for (long i=0; i < TI->n0Info.local; i ++)
    {
        toepZeroPad_v10(x + i * TI->lvlGrain[0], TI->fftBuf, TI->rank - 1, TI->ns + 1);
        fftw_execute(TI->tyFFTRow);
        memcpy(TI->compBuf + i * TI->pwrLvlGrain[0], TI->fftBuf, sizeof(fftw_complex) * TI->pwrLvlGrain[0]);
    }

    /* transposition */
    transpose_v10(TI->compBuf, TI->pwrN1Info, TI->n0Info, TI, TI->transBuf);


    /* column-wise zeropad. */
    for (long i = 0; i < TI->pwrN1Info.local; i ++)
    {
        toepZeroPad_v10(TI->transBuf + i * TI->ns[0], TI->fftBuf, 1, &TI->ns[0]);
        fftw_execute(TI->tyFFTCol);
        memcpy(TI->compBuf + i * TI->ns2[0], TI->fftBuf, sizeof(fftw_complex) * TI->ns2[0]);
    }

    /* lam_t .* fft(x) */
    for (long i=0; i < TI->pwrN1Info.local; i ++)
        for (long j = 0; j < TI->ns2[0]; j++)
            TI->compBuf[i * TI->ns2[0] + j] *= lam_t[i * TI->ns2[0] + j];

    /*
     * Retrieve result
     */
    /* column-wise retrieve */
    for (long i =0; i < TI->pwrN1Info.local; i ++)
    {
        memcpy(TI->fftBuf, TI->compBuf + i * TI->ns2[0], sizeof(fftw_complex) * TI->ns2[0]);
        fftw_execute(TI->tyIFFTCol);
        for (long j=0; j<TI->ns[0]; j ++)
            TI->compBuf[i*TI->ns[0] + j] = TI->fftBuf[j] / TI->ns2[0];
    }

    /* transpose back */
    transpose_v10(TI->compBuf, TI->n0Info, TI->pwrN1Info, TI, TI->transBuf);

    /* row-wise retrieve */
    for (long i=0; i < TI->n0Info.local; i ++)
    {
        memcpy(TI->fftBuf, TI->transBuf + i * TI->pwrLvlGrain[0], sizeof(fftw_complex) * TI->pwrLvlGrain[0]);
        fftw_execute(TI->tyIFFTRow);
        for (long j = 0; j < TI->pwrLvlGrain[0]; j++)
            TI->fftBuf[j] = TI->fftBuf[j] / TI->pwrLvlGrain[0];
        toepRetrv_v10(TI->fftBuf, y + i * TI->lvlGrain[0], TI->rank - 1, TI->ns + 1);
    }

}

/* 
 * Toeplitz vector precondition 
 * Input:
 *      lam_m       : eigenvalues of the preconditioner matrix
 *      x           : input vector
 *      TI          : system configuration
 * Output:
 *      y           : output vector
 */
void toep_solver_precon_v10(complex * lam_m, complex * x, struct ToeplitzSolverInfo_v10 * TI, complex * y)
{
    /*
     * Start of preconditioning
     */
    /* row fft of r */
    for (long i = 0; i < TI->n0Info.local; i ++)
    {
        memcpy(TI->fftBuf, x + i * TI->lvlGrain[0], sizeof(complex) * TI->lvlGrain[0]);
        fftw_execute(TI->myFFTRow);
        memcpy(TI->compBuf + i * TI->lvlGrain[0], TI->fftBuf, sizeof(complex) * TI->lvlGrain[0]);
    }
    

    /* transpose */
    transpose_v10(TI->compBuf, TI->n1Info, TI->n0Info, TI, TI->transBuf); 

    /* col fft for r */
    for (long i =0; i < TI->n1Info.local; i ++)
    {
        memcpy(TI->fftBuf, TI->transBuf + i * TI->ns[0], sizeof(complex) * TI->ns[0]);
        fftw_execute(TI->myFFTCol);
        memcpy(TI->compBuf + i * TI->ns[0], TI->fftBuf, sizeof(complex) * TI->ns[0]);
    }

    /*
     * lam_m^-1 * r
     */
    for (long i = 0; i < TI->n1Info.local * TI->ns[0]; i ++)
        TI->compBuf[i] /= lam_m[i];

    /* col ifft for m */
    for (long i = 0; i < TI->n1Info.local; i ++)
    {
        memcpy(TI->fftBuf, TI->compBuf + i * TI->ns[0], sizeof(complex) * TI->ns[0]);
        fftw_execute(TI->myIFFTCol);
        for (long j = 0; j < TI->ns[0]; j ++)
            TI->compBuf[i*TI->ns[0] + j] = TI->fftBuf[j] / TI->ns[0];
    }

    /* transpose back */
    transpose_v10(TI->compBuf, TI->n0Info, TI->n1Info, TI, TI->transBuf);

    /* row ifft for m */
    for (long i = 0; i < TI->n0Info.local; i ++)
    {
        memcpy(TI->fftBuf, TI->transBuf + i * TI->lvlGrain[0], sizeof(complex) * TI->lvlGrain[0]);
        fftw_execute(TI->myIFFTRow);
        for (long j = 0; j < TI->lvlGrain[0]; j++)
            y[i * TI->lvlGrain[0] + j] = TI->fftBuf[j] / TI->lvlGrain[0];
    }

}


/*
 * Calculate ||b - Tx||, the norm of residual
 * Input:
 *      lam_t       : eigenvalues of a Toeplitz matrix
 *      x           : unknown variable vector in Tx = b
 *      b           : RHS of Tx = b
 *      TI          : system configuration
 * Output:
 *      rho(return) : norm of residual
 */
double toep_solver_residual_v10(complex * lam_t, complex * x, complex * b, struct ToeplitzSolverInfo_v10 * TI)
{
    complex * tx;
    complex * r;
    tx = (complex *) malloc (sizeof(complex) * TI->allocLocal);
    r = (complex *) malloc (sizeof(complex) * TI->allocLocal);

    toep_solver_matvec_v10(lam_t, x, TI, tx);

    for (long i = 0; i < TI->n0Info.local * TI->lvlGrain[0]; i ++) r[i] = b[i] - tx[i];

    complex localRes, globalRes;
    localRes = vectorCProd_v10(r, r, TI->n0Info.local * TI->lvlGrain[0]);

    MPI_Allreduce(&localRes, &globalRes, 1, MPI_DOUBLE_COMPLEX, MPI_SUM, TI->mpiComm);

    /*
     * Cleaning up
     */
    free(tx);
    free(r);

    return creal(csqrt(globalRes));
}


/*
 * SECTION: Toeplitz subroutines
 */
/* 
 * Embed a Toeplitz vector into a circulant vector 
 * Input:
 *      src         : source buffer
 *      rank        : level of the Toeplitz matrix
 *      ns          : dimensionality of the Toeplitz matrix
 * Output:
 *      dest        : destination buffer
 */
void toepEmbd_v10(complex * src, complex * dest, const int rank, const int * ns)
{
    /*
     * function's own version of lvlGrain and pwrLvlGrain. This is needed because
     *  we break the embed/zero-pad/retrieve process in parts.
     */
    long * lvlGrain;                            /* granularity of levels */
    long * pwrLvlGrain;                         /* granularity of expanded levels */
    /*int n;*/                                      /* length of the original vector t */
    int N;                                      /* length of the fully embedded vector t. N = n * 2^rank */

    /*
     * Calculate lvlGrain, pwrLvlGrain, n, and N.
     */
    lvlGrain = (long *) malloc(sizeof(long) * rank);
    pwrLvlGrain = (long *) malloc(sizeof(long) * rank);
    lvlGrain[rank-1] = 1;
    pwrLvlGrain[rank-1] = 1;
    for (int i = rank-2; i >=0; i --)
    {
        lvlGrain[i] = lvlGrain[i+1] * ns[i+1];
        pwrLvlGrain[i] = pwrLvlGrain[i + 1] * ns[i+1] * 2;
    }
    /*n = ns[0] * lvlGrain[0];*/
    N = ns[0] * 2 * pwrLvlGrain[0];

    /*
     * Calculate the corresponding source based on the destination
     */

    int * srcInds = (int *) malloc ( sizeof(int) * rank);
    int zeroFlag;
    for (long destOff = 0; destOff < N; destOff ++)
    {
        long destRem = destOff;
        zeroFlag = 0;
        for (int lvl = 0; lvl < rank; lvl++)
        {
            srcInds[lvl] = destRem / pwrLvlGrain[lvl];
            /* if we should pad zero at destOff */
            if(srcInds[lvl] == ns[lvl])
            {
                zeroFlag = 1;
                break;
            }
            /* if it is a reflection */
            if (srcInds[lvl] > ns[lvl])
            {
                srcInds[lvl] = ns[lvl] * 2 - srcInds[lvl];
            }

            destRem = destRem % pwrLvlGrain[lvl];
        }

        if ( zeroFlag != 0 )
        {
            /* pad zero */
            dest[destOff] = 0;
        } else
        {
            long srcOff = 0;
            for (int lvl=0; lvl<rank; lvl++)
            {
                srcOff += srcInds[lvl] * lvlGrain[lvl];
            }

            dest[destOff] = src[srcOff];
        }
    }

    /*
     * final clean-up
     */
    free(lvlGrain);
    free(pwrLvlGrain);
    free(srcInds);
}

/* 
 * Zero-pad a vector to prepare it for Toeplitz multiplication
 * Input:
 *      src         : source buffer
 *      rank        : level of the Toeplitz matrix
 *      ns          : dimensionality of the Toeplitz matrix
 * Output:
 *      dest        : destination buffer
 */
void toepZeroPad_v10(complex * src, complex * dest, const int rank, const int * ns)
{
    /*
     * function's own version of lvlGrain and pwrLvlGrain. This is needed because
     *  we break the embed/zero-pad/retrieve process in parts.
     */
    long * lvlGrain;                            /* granularity of levels */
    long * pwrLvlGrain;                         /* granularity of expanded levels */
    /*int n;*/                                      /* length of the original vector t */
    int N;                                      /* length of the fully embedded vector t. N = n * 2^rank */

    /*
     * Calculate lvlGrain, pwrLvlGrain, n, and N.
     */
    lvlGrain = (long *) malloc(sizeof(long) * rank);
    pwrLvlGrain = (long *) malloc(sizeof(long) * rank);
    lvlGrain[rank-1] = 1;
    pwrLvlGrain[rank-1] = 1;
    for (int i = rank-2; i >=0; i --)
    {
        lvlGrain[i] = lvlGrain[i+1] * ns[i+1];
        pwrLvlGrain[i] = pwrLvlGrain[i + 1] * ns[i+1] * 2;
    }
    /*n = ns[0] * lvlGrain[0];*/
    N = ns[0] * 2 * pwrLvlGrain[0];

    /*
     * Calculate the corresponding source based on the destination
     */

    int * srcInds = (int *) malloc ( sizeof(int) * rank);
    int zeroFlag;
    for (long destOff = 0; destOff < N; destOff ++)
    {
        long destRem = destOff;
        zeroFlag = 0;
        for (int lvl = 0; lvl < rank; lvl++)
        {
            srcInds[lvl] = destRem / pwrLvlGrain[lvl];
            /* if we should pad zero at destOff */
            if(srcInds[lvl] >= ns[lvl])
            {
                zeroFlag = 1;
                break;
            }

            destRem = destRem % pwrLvlGrain[lvl];
        }

        if ( zeroFlag != 0 )
        {
            /* pad zero */
            dest[destOff] = 0;
        } else
        {
            long srcOff = 0;
            for (int lvl=0; lvl<rank; lvl++)
            {
                srcOff += srcInds[lvl] * lvlGrain[lvl];
            }

            dest[destOff] = src[srcOff];
        }
    }

    /*
     * final clean-up
     */
    free(lvlGrain);
    free(pwrLvlGrain);
    free(srcInds);
}

/* 
 * Retrieve the original vector from its Toeplitz embedding
 * Input:
 *      src         : source buffer
 *      rank        : level of the Toeplitz matrix
 *      ns          : dimensionality of the Toeplitz matrix
 * Output:
 *      dest        : destination buffer
 */
void toepRetrv_v10(complex * src, complex * dest, const int rank, const int * ns)
{
    /*
     * function's own version of lvlGrain and pwrLvlGrain. This is needed because
     *  we break the embed/zero-pad/retrieve process in parts.
     */
    long * lvlGrain;                            /* granularity of levels */
    long * pwrLvlGrain;                         /* granularity of expanded levels */
    /*int n;*/                                      /* length of the original vector t */
    int N;                                      /* length of the fully embedded vector t. N = n * 2^rank */

    /*
     * Calculate lvlGrain, pwrLvlGrain, n, and N.
     */
    lvlGrain = (long *) malloc(sizeof(long) * rank);
    pwrLvlGrain = (long *) malloc(sizeof(long) * rank);
    lvlGrain[rank-1] = 1;
    pwrLvlGrain[rank-1] = 1;
    for (int i = rank-2; i >=0; i --)
    {
        lvlGrain[i] = lvlGrain[i+1] * ns[i+1];
        pwrLvlGrain[i] = pwrLvlGrain[i + 1] * ns[i+1] * 2;
    }
    /*n = ns[0] * lvlGrain[0];*/
    N = ns[0] * 2 * pwrLvlGrain[0];

    /*
     * Calculate the corresponding source based on the destination
     */

    int * destInds = (int *) malloc ( sizeof(int) * rank);
    int zeroFlag;
    for (long srcOff = 0; srcOff < N; srcOff ++)
    {
        long srcRem = srcOff;
        zeroFlag = 0;
        for (int lvl = 0; lvl < rank; lvl++)
        {
            destInds[lvl] = srcRem / pwrLvlGrain[lvl];
            /* a part that we should ignore */
            if(destInds[lvl] >= ns[lvl])
            {
                zeroFlag = 1;
                break;
            }
            srcRem = srcRem % pwrLvlGrain[lvl];
        }

        if ( zeroFlag != 0 )
        {
            /* do nothing */
        } else
        {
            long destOff = 0;
            for (int lvl=0; lvl<rank; lvl++)
            {
                destOff += destInds[lvl] * lvlGrain[lvl];
            }

            dest[destOff] = src[srcOff];
        }
    }

    /*
     * final clean-up
     */
    free(lvlGrain);
    free(pwrLvlGrain);
    free(destInds);
}

/* 
 * Compute the precondition vector. The calculation is in-place 
 * Input:
 *      srcdest     : source buffer
 *      rank        : level of the Toeplitz matrix
 *      ns          : dimensionality of the Toeplitz matrix
 * Output:
 *      srcdest     : destination buffer
 */
void toepPreVec_v10(complex * srcdest, const int rank, const int * ns)
{
    /*
     * function's own version of lvlGrain and pwrLvlGrain. This is needed because
     *  we break the embed/zero-pad/retrieve process in parts.
     */
    long * lvlGrain;                            /* granularity of levels */
    int n;                                      /* length of the original vector t */

    /*
     * Calculate lvlGrain, pwrLvlGrain, n, and N.
     */
    lvlGrain = (long *) malloc(sizeof(long) * rank);
    lvlGrain[rank-1] = 1;
    for (int i = rank-2; i >=0; i --)
    {
        lvlGrain[i] = lvlGrain[i+1] * ns[i+1];
    }
    n = ns[0] * lvlGrain[0];


    /* for each level, from innermost to outermost */
    for (int lvl = rank-1; lvl >=0 ; lvl --)
    {
        /* for each block in that level */
        for(long blkCount = 0; blkCount < n / lvlGrain[lvl] / ns[lvl]; blkCount ++  )
        {
            long blkOff = blkCount * lvlGrain[lvl] * ns[lvl];
            /* for each subblock inside the block */
            for(long blkNum = 1; blkNum < (ns[lvl] + 1)/2; blkNum ++)
            {
                double avg;
                /* for each element */
                for (long i = 0; i < lvlGrain[lvl]; i++)
                {
                    avg = (ns[lvl] * 1.0 - blkNum)/ns[lvl] * srcdest[blkOff + blkNum * lvlGrain[lvl] + i] +
                            blkNum * 1.0/ns[lvl] * srcdest[blkOff + (ns[lvl] - blkNum) * lvlGrain[lvl] + i];
                    srcdest[blkOff + blkNum * lvlGrain[lvl] + i] = avg;
                    srcdest[blkOff + (ns[lvl] - blkNum) * lvlGrain[lvl] + i] = avg;
                }
            }
        }
    }

    /*
     * final clean-up
     */
    free(lvlGrain);
}



/*
 * SECTION: Vector operation subroutines
 */
/* compute the inner product of two vectors */
complex vectorCProd_v10(complex *v1, complex *v2, long length)
{
    complex sum = 0;
    for (long i = 0; i < length; i ++)
    {
        sum += v1[i] * conj(v2[i]);
    }
    return sum;
}

/* randomize the vector. Each element will be in range [0, 1) */
void vectorCRand_v10(complex *v, long length)
{
    for (long i=0; i < length; i ++)
    {
        v[i] = rand() * 1.0 / RAND_MAX;
    }
}

/*
 * SECTION: Miscellaneous routines
 */
void transpose_v10(complex * srcBuf, struct DimensionInfo_v10 splitDim, struct DimensionInfo_v10 joinDim,
               struct ToeplitzSolverInfo_v10 * TI, complex * destBuf)
{
    for ( int proc = 0; proc < TI->numprocs; proc ++) {
        /* send counts and send displacements */
        TI->sendCnts[proc] = joinDim.local * splitDim.partitions[proc];
        TI->sDispls[proc] = joinDim.local * splitDim.accumulations[proc];

        /* JC: remove
        for (long i = 0; i < joinDim.local; i ++) {
            memcpy( TI->sendBuf + joinDim.local * splitDim.accumulations[proc] + i * splitDim.partitions[proc],
                    srcBuf + i * splitDim.size + splitDim.accumulations[proc],
                    sizeof(complex) * splitDim.partitions[proc]);
        }

        and replace by the following:
        */

        long sendBufDispls = joinDim.local * splitDim.accumulations[proc];
        long srcBufDispls = splitDim.accumulations[proc];
        for (long i = 0; i < joinDim.local; i ++)
        {
            memcpy(TI->sendBuf + sendBufDispls,
                   srcBuf + srcBufDispls,
                   sizeof(complex) * splitDim.partitions[proc]
                   );
            sendBufDispls += splitDim.partitions[proc];
            srcBufDispls += splitDim.size;
        }

        /* receive counts and receive displacements */
        TI->recvCnts[proc] = splitDim.local * joinDim.partitions[proc];
        TI->rDispls[proc] = splitDim.local * joinDim.accumulations[proc];
    }

#if ALT_ALLTOALLV
    /* alternative to MPI_Alltoallv */
    int mpiError = MPI_SUCCESS;
    char errorString[255];
    int errorStringLength;
    const int kAlltoallvTag = 1001;
    for (int proc_offset = 0; proc_offset < TI->numprocs; proc_offset ++) {
        int proc = (TI->myid + proc_offset) % TI->numprocs;
        mpiError = MPI_Irecv(TI->recvBuf + TI->rDispls[proc], TI->recvCnts[proc], MPI_DOUBLE_COMPLEX,
                proc, kAlltoallvTag, TI->mpiComm, &TI->mpiRequests[TI->numprocs + proc]);
        if (mpiError) {
            MPI_Error_string(mpiError, errorString, &errorStringLength);
            perror(errorString);
        }
        mpiError = MPI_Isend(TI->sendBuf + TI->sDispls[proc], TI->sendCnts[proc], MPI_DOUBLE_COMPLEX,
                proc, kAlltoallvTag, TI->mpiComm, &TI->mpiRequests[proc]);
        if (mpiError) {
            MPI_Error_string(mpiError, errorString, &errorStringLength);
            perror(errorString);
        }
    }
    mpiError = MPI_Waitall(2 * TI->numprocs, TI->mpiRequests, TI->mpiStatusList);
    if (mpiError == MPI_ERR_IN_STATUS) {
        for (int i=0; i < 2 * TI->numprocs; i ++) {
            if (TI->mpiStatusList[i].MPI_ERROR != MPI_SUCCESS) {
                MPI_Error_string(mpiError, errorString, &errorStringLength);
                perror(errorString);
            }
        }
    }
                       
#else
    MPI_Alltoallv(
            TI->sendBuf, TI->sendCnts, TI->sDispls, MPI_DOUBLE_COMPLEX,
            TI->recvBuf, TI->recvCnts, TI->rDispls, MPI_DOUBLE_COMPLEX,
            TI->mpiComm
        );
#endif

    /* pad data */
    /* JC: remove
    for (long i = 0; i < splitDim.local; i ++)
        for (long j = 0; j < joinDim.size; j ++)
            destBuf[i * joinDim.size + j] = TI->recvBuf[j * splitDim.local + i];

       and replace by the following:
    */
    complex *recvBufPtr = TI->recvBuf;
    for (long j = 0; j < joinDim.size; j ++)
        for (long i = 0; i < splitDim.local; i ++)
            destBuf[i * joinDim.size + j] = *recvBufPtr++;
}

