/*
 * toep_solver_v1.5.c
 * MPI_CG version 1.5
 *
 * Author: Tom Li
 * Created on: July, 19, 2012
 */
/*
 * Edited by JC:
 * 1. Append _v15 to all the data structures and functions
 * 2. Add code in toep_solver_plan_v15 to prevent computing lam_m if NULL
 */

#include "mpi.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <complex.h>
#include <math.h>
#include <time.h>
#include <fftw3-mpi.h>

#include "toep_solver_v1.5.h"


#define max(a,b) \
   ({ __typeof__ (a) _a = (a); \
       __typeof__ (b) _b = (b); \
     _a > _b ? _a : _b; })




/*
 * System interface. This are the functions users should call.
 */
/* 
 * Systen initialization 
 * Input:
 *      mpiComm     : global communicator
 *      rank        : number of levels in the system
 *      rowprocs    : number of rows in the partition grid
 *      ns          : dimensionality array
 * Output:
 *      TI          : system configuration
 *      error_code(return)  : whether the allocation is successful
 */
int toep_solver_init_v15(MPI_Comm mpiComm, int rowprocs, int rank, int * ns, struct ToeplitzSolverInfo_v15 * TI)
{
    /*
     * MPI communicator management
     */
    MPI_Comm_size(mpiComm, &TI->numprocs);
    MPI_Comm_rank(mpiComm, &TI->myid);

    TI->mpiComm = mpiComm;

    TI->mpiRequests = (MPI_Request *) malloc(sizeof(MPI_Request) * 2 * TI->numprocs);
    TI->mpiStatusList = (MPI_Status *) malloc(sizeof(MPI_Status) * 2 * TI->numprocs);
   
    if ( TI->numprocs % rowprocs != 0 )
    {
        return TOEP_INIT_WRONG_ROW_NUM;
    }

    TI->rowprocs = rowprocs;
    TI->colprocs = TI->numprocs / TI->rowprocs;
    TI->colid = TI->myid % TI->colprocs;
    TI->rowid = TI->myid / TI->colprocs;

    /* create MPI communicators */
    MPI_Comm_split(TI->mpiComm, TI->rowid, TI->colid, &TI->rowComm);
    MPI_Comm_split(TI->mpiComm, TI->colid + 10000, TI->rowid, &TI->colComm);


    /*
     * Calculate lvlGrain, pwrLvlGrain, n, and N.
     */
    TI->lvlGrain = (long *) malloc(sizeof(long) * rank);
    TI->pwrLvlGrain = (long *) malloc(sizeof(long) * rank);
    TI->lvlGrain[rank-1] = 1;
    TI->pwrLvlGrain[rank-1] = 1;
    for (int i = rank-2; i >=0; i --)
    {
        TI->lvlGrain[i] = TI->lvlGrain[i+1] * ns[i+1];
        TI->pwrLvlGrain[i] = TI->pwrLvlGrain[i + 1] * ns[i+1] * 2;
    }
    TI->n = ns[0] * TI->lvlGrain[0];
    TI->N = ns[0] * 2 * TI->pwrLvlGrain[0];

    TI->rank = rank;

    TI->ns = (int *)malloc(sizeof(int) * rank);
    for (int i=0; i < rank; i ++) TI->ns[i] = ns[i];

    TI->ns2 = (int *) malloc ( sizeof(int) * rank);
    for (int i=0; i < rank; i ++) TI->ns2[i] = ns[i] * 2;



    /*
     * dimension sizes
     */

    TI->n0Info.size = TI->ns[0];
    TI->n1ColInfo.size = TI->ns[1];
    TI->n1RowInfo.size = TI->ns[1];
    TI->n2Info.size = TI->lvlGrain[1];

    TI->pwrN1ColInfo.size = TI->ns2[1];
    TI->pwrN1RowInfo.size = TI->ns2[1];
    TI->pwrN2Info.size = TI->pwrLvlGrain[1];

    /*
     * Calculate dimension partition and accumulation
     */
    long blksize;
    TI->n0Info.partitions = (long *) malloc(sizeof(long) * TI->rowprocs);
    TI->n1ColInfo.partitions = (long *) malloc(sizeof(long) * TI->colprocs);
    TI->n1RowInfo.partitions = (long *) malloc(sizeof(long) * TI->rowprocs);
    TI->n2Info.partitions = (long *) malloc(sizeof(long) * TI->colprocs);
    TI->pwrN1ColInfo.partitions   = (long * ) malloc(sizeof(long) * TI->colprocs);
    TI->pwrN1RowInfo.partitions   = (long * ) malloc(sizeof(long) * TI->rowprocs);
    TI->pwrN2Info.partitions  = (long * ) malloc(sizeof(long) * TI->colprocs);

    TI->n0Info.accumulations = (long *) malloc(sizeof(long) * TI->rowprocs);
    TI->n1ColInfo.accumulations = (long *) malloc(sizeof(long) * TI->colprocs);
    TI->n1RowInfo.accumulations = (long *) malloc(sizeof(long) * TI->rowprocs);
    TI->n2Info.accumulations = (long *) malloc(sizeof(long) * TI->colprocs);
    TI->pwrN1ColInfo.accumulations   = (long * ) malloc(sizeof(long) * TI->colprocs);
    TI->pwrN1RowInfo.accumulations   = (long * ) malloc(sizeof(long) * TI->rowprocs);
    TI->pwrN2Info.accumulations  = (long * ) malloc(sizeof(long) * TI->colprocs);

    /* partition for n0 by rows */
    blksize = ceil(TI->ns[0] * 1.0 / TI->rowprocs);
    for (long i = 0; i < TI->rowprocs; i ++)
    {
        if ( i < TI->ns[0] / blksize )
            TI->n0Info.partitions[i] = blksize;
        else
            if ( i == TI->ns[0] / blksize )
                TI->n0Info.partitions[i] = TI->ns[0] - blksize * (TI->ns[0] / blksize );
            else
                TI->n0Info.partitions[i] = 0;
    }

    /* partition for n1 by columns */
    blksize = ceil(TI->ns[1] * 1.0 / TI->colprocs);
    for (long i = 0; i < TI->colprocs; i ++)
    {
        if ( i < TI->ns[1] / blksize )
            TI->n1ColInfo.partitions[i] = blksize;
        else
            if ( i == TI->ns[1] / blksize )
                TI->n1ColInfo.partitions[i] = TI->ns[1] - blksize * (TI->ns[1] / blksize );
            else
                TI->n1ColInfo.partitions[i] = 0;
    }

    /* partition for n1 by rows */
    blksize = ceil(TI->ns[1] * 1.0 / TI->rowprocs);
    for (long i = 0; i < TI->rowprocs; i ++)
    {
        if ( i < TI->ns[1] / blksize )
            TI->n1RowInfo.partitions[i] = blksize;
        else
            if ( i == TI->ns[1] / blksize )
                TI->n1RowInfo.partitions[i] = TI->ns[1] - blksize * (TI->ns[1] / blksize );
            else
                TI->n1RowInfo.partitions[i] = 0;
    }


    /* partition for n2(*n3*n4....) by columns */
    blksize = ceil(TI->lvlGrain[1] * 1.0 / TI->colprocs);
    for (long i = 0; i < TI->colprocs; i ++ )
    {
        if ( i < TI->lvlGrain[1] / blksize )
            TI->n2Info.partitions[i] = blksize;
        else
            if ( i == TI->lvlGrain[1] / blksize )
                TI->n2Info.partitions[i] = TI->lvlGrain[1] - blksize * ( TI->lvlGrain[1] / blksize );
            else
                TI->n2Info.partitions[i] = 0;
    }


    /* partition for TI->ns2[1] by column */
    blksize = ceil(TI->ns2[1] * 1.0 / TI->colprocs);
    for (long i = 0; i < TI->colprocs; i ++)
    {
        if (i < TI->ns2[1] / blksize)
            TI->pwrN1ColInfo.partitions[i] = blksize;
        else
            if (i == TI->ns2[1] / blksize)
                TI->pwrN1ColInfo.partitions[i] = TI->ns2[1] - blksize * (TI->ns2[1] / blksize);
            else
                TI->pwrN1ColInfo.partitions[i] = 0;
    }

    /* partition for TI->ns2[1] by row */
    blksize = ceil(TI->ns2[1] * 1.0 / TI->rowprocs);
    for (long i = 0; i < TI->rowprocs; i ++)
    {
        if (i < TI->ns2[1] / blksize)
            TI->pwrN1RowInfo.partitions[i] = blksize;
        else
            if (i == TI->ns2[1] / blksize)
                TI->pwrN1RowInfo.partitions[i] = TI->ns2[1] - blksize * (TI->ns2[1] / blksize);
            else
                TI->pwrN1RowInfo.partitions[i] = 0;
    }


    /* partition for n2(*n3*n4....) by columns */
    blksize = ceil(TI->pwrLvlGrain[1] * 1.0 / TI->colprocs);

    for (long i = 0; i < TI->colprocs; i ++ )
    {
        if ( i < TI->pwrLvlGrain[1] / blksize )
            TI->pwrN2Info.partitions[i] = blksize;
        else
            if ( i == TI->pwrLvlGrain[1] / blksize )
                TI->pwrN2Info.partitions[i] = TI->pwrLvlGrain[1] - blksize * ( TI->pwrLvlGrain[1] / blksize );
            else
                TI->pwrN2Info.partitions[i] = 0;
    }


    /* accumulations */
    TI->n0Info.accumulations[0] = 0;
    TI->n1ColInfo.accumulations[0] = 0;
    TI->n1RowInfo.accumulations[0] = 0;
    TI->n2Info.accumulations[0] = 0;

    TI->pwrN1ColInfo.accumulations[0] = 0;
    TI->pwrN1RowInfo.accumulations[0] = 0;
    TI->pwrN2Info.accumulations[0] = 0;
    for (long i =1; i < TI->rowprocs; i ++)
    {
        TI->n0Info.accumulations[i] = TI->n0Info.accumulations[i-1] + TI->n0Info.partitions[i-1];
        TI->n1RowInfo.accumulations[i] = TI->n1RowInfo.accumulations[i-1] + TI->n1RowInfo.partitions[i-1];
        TI->pwrN1RowInfo.accumulations[i] = TI->pwrN1RowInfo.accumulations[i-1] + TI->pwrN1RowInfo.partitions[i-1];
    }

    for (long i = 1; i < TI->colprocs; i ++)
    {
        TI->n1ColInfo.accumulations[i] = TI->n1ColInfo.accumulations[i-1] + TI->n1ColInfo.partitions[i-1];
        TI->n2Info.accumulations[i] = TI->n2Info.accumulations[i-1] + TI->n2Info.partitions[i-1];

        TI->pwrN1ColInfo.accumulations[i] = TI->pwrN1ColInfo.accumulations[i-1] + TI->pwrN1ColInfo.partitions[i-1];
        TI->pwrN2Info.accumulations[i] = TI->pwrN2Info.accumulations[i-1] + TI->pwrN2Info.partitions[i-1];
    }


    /* variables for housekeeping and memory allocation */
    TI->n0Info.local = TI->n0Info.partitions[TI->rowid];
    TI->n1ColInfo.local = TI->n1ColInfo.partitions[TI->colid];
    TI->n1RowInfo.local = TI->n1RowInfo.partitions[TI->rowid];
    TI->n2Info.local = TI->n2Info.partitions[TI->colid];

    TI->pwrN1ColInfo.local = TI->pwrN1ColInfo.partitions[TI->colid];
    TI->pwrN1RowInfo.local = TI->pwrN1RowInfo.partitions[TI->rowid];
    TI->pwrN2Info.local = TI->pwrN2Info.partitions[TI->colid];

    TI->n0Info.localStart = TI->n0Info.accumulations[TI->rowid];
    TI->n1ColInfo.localStart = TI->n1ColInfo.accumulations[TI->colid];

    TI->allocLocal = max(TI->lvlGrain[1] * TI->n1ColInfo.local * TI->n0Info.local, TI->n2Info.local * TI->ns[1] * TI->n0Info.local); 
    TI->allocLocal = max(TI->allocLocal, TI->n2Info.local * TI->n1RowInfo.local * TI->ns[0] );

    TI->pwrAllocLocal = max(TI->pwrLvlGrain[1] * TI->pwrN1ColInfo.local * TI->n0Info.local, TI->pwrN2Info.local * TI->ns2[1] * TI->n0Info.local );
    TI->pwrAllocLocal = max(TI->pwrAllocLocal, TI->pwrN2Info.local * TI->pwrN1RowInfo.local * TI->ns[0]);


    /*
     * Transposition variables
     */
    TI->sendBuf = (complex *) malloc ( sizeof(complex) * (TI->pwrAllocLocal + TI->numprocs * 2));
    TI->sendCnts = (int * ) malloc (sizeof(int) * TI->numprocs);
    TI->sDispls = (int *) malloc (sizeof(int) * TI->numprocs);

    TI->recvBuf = (complex *) malloc ( sizeof(complex) * (TI->pwrAllocLocal + TI->numprocs * 2));
    TI->recvCnts = (int * ) malloc (sizeof(int) * TI->numprocs);
    TI->rDispls = (int *) malloc (sizeof(int) * TI->numprocs);

    /*
     * FFTW plans and buffers
     */
    TI->fftBuf = fftw_alloc_complex( max (max(TI->pwrLvlGrain[1], TI->ns2[0]), TI->ns2[1]));
    TI->compBuf = fftw_alloc_complex(TI->pwrAllocLocal * 2);
    TI->transBuf = fftw_alloc_complex(TI->pwrAllocLocal);

    TI->tyFFTN2 = fftw_plan_dft(rank-2, TI->ns2 + 2, TI->fftBuf, TI->fftBuf, FFTW_FORWARD, FFTW_ESTIMATE);
    TI->tyIFFTN2 = fftw_plan_dft(rank-2, TI->ns2 + 2, TI->fftBuf, TI->fftBuf, FFTW_BACKWARD, FFTW_ESTIMATE);

    TI->tyFFTN1 = fftw_plan_dft(1, &TI->ns2[1] , TI->fftBuf, TI->fftBuf, FFTW_FORWARD, FFTW_ESTIMATE);
    TI->tyIFFTN1 = fftw_plan_dft(1, &TI->ns2[1], TI->fftBuf, TI->fftBuf, FFTW_BACKWARD, FFTW_ESTIMATE);

    TI->tyFFTN0 = fftw_plan_dft(1, &TI->ns2[0] , TI->fftBuf, TI->fftBuf, FFTW_FORWARD, FFTW_ESTIMATE);
    TI->tyIFFTN0 = fftw_plan_dft(1, &TI->ns2[0], TI->fftBuf, TI->fftBuf, FFTW_BACKWARD, FFTW_ESTIMATE);

    TI->myFFTN2 = fftw_plan_dft(rank-2, TI->ns + 2, TI->fftBuf, TI->fftBuf, FFTW_FORWARD, FFTW_ESTIMATE);
    TI->myIFFTN2 = fftw_plan_dft(rank-2, TI->ns + 2, TI->fftBuf, TI->fftBuf, FFTW_BACKWARD, FFTW_ESTIMATE);

    TI->myFFTN1 = fftw_plan_dft(1, &TI->ns[1] , TI->fftBuf, TI->fftBuf, FFTW_FORWARD, FFTW_ESTIMATE);
    TI->myIFFTN1 = fftw_plan_dft(1, &TI->ns[1], TI->fftBuf, TI->fftBuf, FFTW_BACKWARD, FFTW_ESTIMATE);

    TI->myFFTN0 = fftw_plan_dft(1, &TI->ns[0] , TI->fftBuf, TI->fftBuf, FFTW_FORWARD, FFTW_ESTIMATE);
    TI->myIFFTN0 = fftw_plan_dft(1, &TI->ns[0], TI->fftBuf, TI->fftBuf, FFTW_BACKWARD, FFTW_ESTIMATE);

    return TOEP_INIT_SUCCESS;
}

/*
 * Planning function. 
 * Input:
 *      t           : a Toeplitz vector. The first column of a Toeplitz matrix
 *      TI          : system configuration
 * OutPut:
 *      lam_t       : the eigenvalues of the embedded Toeplitz matrix
 *      lam_m       : the eigenvalues of the preconditioner matrix
 */
void toep_solver_plan_v15(complex * t, struct ToeplitzSolverInfo_v15 * TI, complex * lam_t, complex * lam_m)
{

    /*
     * SECTION: Compute lam_t, the eigenvalue of t
     */
    for (long i=0; i < TI->n0Info.local * TI->n1ColInfo.local; i ++)
    {
        toepEmbd_v15(t + i * TI->lvlGrain[1], TI->fftBuf, TI->rank - 2, TI->ns + 2);
        fftw_execute(TI->tyFFTN2);
        memcpy(TI->compBuf + i * TI->pwrLvlGrain[1], TI->fftBuf, sizeof(fftw_complex) * TI->pwrLvlGrain[1]);
    }

    /*
     * lam_t
     * row-wise transposition, pwrN2 -> n1
     */
    transpose_v15(TI->compBuf, TI->n0Info, TI->pwrN2Info, TI->n1ColInfo, TRANS_MODE_ROW, TI, TI->transBuf);

    /*
     * lam_t
     * embed and fft on n1
     */
    for (long i=0; i < TI->n0Info.local * TI->pwrN2Info.local; i ++)
    {
        toepEmbd_v15(TI->transBuf + i * TI->ns[1], TI->fftBuf, 1, &TI->ns[1]);
        fftw_execute(TI->tyFFTN1);
        memcpy(TI->compBuf + i * TI->ns2[1], TI->fftBuf, sizeof(fftw_complex) * TI->ns2[1]);
    }

    /*
     * lam_t
     * column-wise transposition, pwrN1 -> n0
     */
    transpose_v15(TI->compBuf, TI->pwrN2Info, TI->pwrN1RowInfo, TI->n0Info, TRANS_MODE_COL, TI, TI->transBuf);

    /*
     * lam_t
     * embed and fft on n0
     */
    for (long i=0; i < TI->pwrN1RowInfo.local * TI->pwrN2Info.local; i ++)
    {
        toepEmbd_v15(TI->transBuf + i * TI->ns[0], TI->fftBuf, 1, &TI->ns[0]);
        fftw_execute(TI->tyFFTN0);
        memcpy(lam_t + i * TI->ns2[0], TI->fftBuf, sizeof(fftw_complex) * TI->ns2[0]);
    }

    if (!lam_m) {
      return;
    }

    /*
     * SECTION: Calculate m
     */
    memcpy(lam_m, t, sizeof(complex) * TI->n0Info.local * TI->n1ColInfo.local * TI->lvlGrain[1]);

    /*
     * m
     * precondition on n2
     */
    for (long i=0; i < TI->n0Info.local * TI->n1ColInfo.local; i++)
        toepPreVec_v15(lam_m + i * TI->lvlGrain[1], TI->rank - 2, TI->ns + 2);

    /*
     * m
     * row-wise transposition, n2 -> n1
     */
    transpose_v15(lam_m, TI->n0Info, TI->n2Info, TI->n1ColInfo, TRANS_MODE_ROW, TI, lam_m);

    /*
     * m
     * preconditioning on n1
     */
    for (long i=0; i < TI->n0Info.local * TI->n2Info.local; i ++)
        toepPreVec_v15(lam_m + i * TI->ns[1], 1, &TI->ns[1]);


    /*
     * m
     * column-wise transposition, n1 -> n0
     */
    transpose_v15(lam_m, TI->n2Info, TI->n1RowInfo, TI->n0Info, TRANS_MODE_COL, TI, lam_m);

    /*
     * m
     * preconditioning on n0
     */
    for (long i=0; i < TI->n1RowInfo.local * TI->n2Info.local; i ++)
        toepPreVec_v15(lam_m + i * TI->ns[0], 1, &TI->ns[0]);

    /*
     * m
     * column-wise backward transposition, n0 -> n1
     */
    transpose_v15(lam_m, TI->n2Info, TI->n0Info, TI->n1RowInfo, TRANS_MODE_COL, TI, lam_m);

    /*
     * m
     * row-wise backward transposition, n1 -> n2
     */
    transpose_v15(lam_m, TI->n0Info, TI->n1ColInfo, TI->n2Info, TRANS_MODE_ROW, TI, lam_m);


    /****************************** 
     * SECTION: Calculate lam_m
     *******************************/ 

    /*
     * lam_m
     * fft on n2
     */
    for (long i=0; i < TI->n0Info.local * TI->n1ColInfo.local; i++)
    {
        memcpy(TI->fftBuf, lam_m + i * TI->lvlGrain[1], sizeof(complex) * TI->lvlGrain[1]);
        fftw_execute(TI->myFFTN2);
        memcpy(lam_m + i * TI->lvlGrain[1], TI->fftBuf, sizeof(complex) * TI->lvlGrain[1]);
    }

    /*
     * lam_m
     * row-wise transposition, n2 -> n1
     */

    transpose_v15(lam_m, TI->n0Info, TI->n2Info, TI->n1ColInfo, TRANS_MODE_ROW, TI, lam_m);

    /*
     * lam_m
     * fft on n1
     */
    for (long i=0; i < TI->n0Info.local * TI->n2Info.local; i ++)
    {
        memcpy(TI->fftBuf, lam_m + i * TI->ns[1], sizeof(complex) * TI->ns[1] );
        fftw_execute(TI->myFFTN1);
        memcpy(lam_m + i * TI->ns[1], TI->fftBuf, sizeof(complex) * TI->ns[1] );
    }

    /*
     * lam_m
     * column-wise transposition, n1 -> n0
     */
    transpose_v15(lam_m, TI->n2Info, TI->n1RowInfo, TI->n0Info, TRANS_MODE_COL, TI, lam_m);


    /*
     * lam_m
     * fft on n0
     */
    for (long i=0; i < TI->n1RowInfo.local * TI->n2Info.local; i ++)
    {
        memcpy(TI->fftBuf, lam_m + i * TI->ns[0], sizeof(complex) * TI->ns[0]);
        fftw_execute(TI->myFFTN0);
        memcpy(lam_m + i * TI->ns[0], TI->fftBuf, sizeof(complex) * TI->ns[0]);
    }

}

/*
 * Solving function.
 * Input:
 *      lam_t       : the eigenvalues of the embedded Toeplitz matrix
 *      lam_m       : the eigenvalues of the preconditioner matrix
 *      b           : the RHS of the linear system Ax = b
 *      tol         : tolerance value
 *      max_iter    : maximum iteration
 *      TI          : system configuration
 *      x           : the initial guess of x, the unknown vector
 * Output:
 *      x           : solution of the linear ysstem Ax = b
 */
void toep_solver_solve_v15(complex * lam_t, complex * lam_m, complex *b,
                       double tol, int max_iter, struct ToeplitzSolverInfo_v15 * TI, complex * x)
{
    /*
     * Algorithm control parameters
     */
    double rho;                                 /* norm of the residual r */
    double gamma;                               /* norm of b */
    complex sigma;                              /* < r, z > */
    complex prev_sigma;                         /* previous sigma */
    complex tau;                                /* < T * p, p > */
    complex alpha;                              /* sigma / tau */
    complex beta=0;                             /* beta = sigma /  prev_sigma */

    complex * r;                                /* residual vector */
    complex * z;                                /* preconditioned vector z */
    complex * p;                                /* p */
    complex * tp;                               /* T * p */

    r = (complex *) malloc (sizeof(complex) * TI->allocLocal);
    z = (complex *) malloc (sizeof(complex) * TI->allocLocal);
    p = (complex *) malloc (sizeof(complex) * TI->allocLocal);
    tp = (complex *) malloc (sizeof(complex) * TI->allocLocal);

    /*
     * SECTION: Calculate y = Tx
     */
    toep_solver_matvec_v15(lam_t, x, TI, (complex *)TI->compBuf);

    /*
     * SECTION: r0 = b - y0
     */
    for (long i=0; i < TI->n0Info.local * TI->n1ColInfo.local * TI->lvlGrain[1]; i ++)
        r[i] = b[i] - TI->compBuf[i];

    /*
     * SECTION: rho0 = norm(r0), gamma =  norm(b)
     */
    complex localRho;
    localRho = vectorCProd_v15(r, r, TI->n0Info.local * TI->n1ColInfo.local * TI->lvlGrain[1]);


    MPI_Allreduce(&localRho, &rho, 1, MPI_DOUBLE_COMPLEX, MPI_SUM, TI->mpiComm);

    rho = csqrt(rho);


    complex localGamma;
    localGamma = vectorCProd_v15(b, b, TI->n0Info.local * TI->n1ColInfo.local * TI->lvlGrain[1]);

    MPI_Allreduce(&localGamma, &gamma, 1, MPI_DOUBLE_COMPLEX, MPI_SUM, TI->mpiComm);
    gamma = csqrt(gamma);

    /*
     * SECTION: z0 = M^(-1) * r0
     */
    toep_solver_precon_v15(lam_m, r, TI, z);

    /*
     * sigma0 = < r0, z0 >
     */
    complex localSigma;
    localSigma = vectorCProd_v15(r, z, TI->n0Info.local * TI->n1ColInfo.local * TI->lvlGrain[1]);

    MPI_Allreduce(&localSigma, &sigma, 1, MPI_DOUBLE_COMPLEX, MPI_SUM, TI->mpiComm);


    /*
     * p0 = z0
     */
    memcpy(p, z, sizeof(complex) * TI->n0Info.local * TI->n1ColInfo.local * TI->lvlGrain[1]);


    /*
     * SECTION: For iter = 0, 1, ....max_iter do
     */
    for ( int iter = 0; iter < max_iter; iter ++)
    {
        /*
         * SECTION: tp = T * p
         */
        toep_solver_matvec_v15(lam_t, p, TI, tp);

        /*
         * tau = < tp, p >
         */
        complex localTau;
        localTau = vectorCProd_v15(tp, p, TI->n0Info.local * TI->n1ColInfo.local * TI->lvlGrain[1]);

        MPI_Allreduce(&localTau, &tau, 1, MPI_DOUBLE_COMPLEX, MPI_SUM, TI->mpiComm);

        /*
         * alpha = sigma / tau
         */

        alpha = sigma / tau;

        /*
         * x = x + alpha * p
         */
        for(long i = 0; i < TI->n0Info.local  * TI->n1ColInfo.local * TI->lvlGrain[1]; i ++)
            x[i] += alpha * p[i];

        /*
         * r = r - alpha * tp
         */
        for (long i = 0; i < TI->n0Info.local * TI->n1ColInfo.local * TI->lvlGrain[1]; i++)
            r[i] -= alpha * tp[i];

        /*
         * rho = norm(r)
         */
        localRho = vectorCProd_v15(r, r, TI->n0Info.local * TI->n1ColInfo.local * TI->lvlGrain[1]);

        MPI_Allreduce(&localRho, &rho, 1, MPI_DOUBLE, MPI_SUM, TI->mpiComm);

        rho = csqrt(rho);

#if DEBUG_PRINT_ENABLED
        if (TI->myid == DEBUG_ID)
        {
            printf("iteration: %d, rho: %g, tau: %g, sigma: %g, alpha: %g, beta: %g\n", iter, rho, creal(tau), creal(sigma), creal(alpha), creal(beta));
            printf("***********************************************\n\n");
        }
#endif

        /*if ( rho / gamma < tol) { break; }*/
        if ( rho / gamma < tol) {
          break;
        }

        /*
         * SECTION: z = M^(-1) * r
         */
        toep_solver_precon_v15(lam_m, r, TI, z);

        /*
         * sigma = < r, z >
         */

        prev_sigma = sigma;
        localSigma = vectorCProd_v15(r, z, TI->n0Info.local * TI->n1ColInfo.local * TI->lvlGrain[1]);

        MPI_Allreduce(&localSigma, &sigma, 1, MPI_DOUBLE_COMPLEX, MPI_SUM, TI->mpiComm);

        /*
         * beta = sigma_(j+1) / sigma(j)
         */
        beta = sigma / prev_sigma;

        /*
         * p = z + beta * p
         */
        for (long i = 0; i < TI->n0Info.local * TI->n1ColInfo.local * TI->lvlGrain[1]; i ++)
            p[i] = z[i] + beta * p[i];
    }

    /*
     * Cleaning up
     */
    free(r);
    free(z);
    free(p);
    free(tp);
}

/*
 * System finalization. It is used to free up allocated memory spaces.
 * Input:
 *      TI          : system configuration
 */
void toep_solver_finalize_v15(struct ToeplitzSolverInfo_v15 * TI)
{
    free(TI->lvlGrain);
    free(TI->pwrLvlGrain);
    free(TI->ns);
    free(TI->ns2);

    free(TI->n0Info.partitions);
    free(TI->n1ColInfo.partitions);
    free(TI->n1RowInfo.partitions);
    free(TI->n2Info.partitions);
    free(TI->pwrN1ColInfo.partitions);
    free(TI->pwrN1RowInfo.partitions);
    free(TI->pwrN2Info.partitions);


    free(TI->n0Info.accumulations);
    free(TI->n1ColInfo.accumulations);
    free(TI->n1RowInfo.accumulations);
    free(TI->n2Info.accumulations);
    free(TI->pwrN1ColInfo.accumulations);
    free(TI->pwrN1RowInfo.accumulations);
    free(TI->pwrN2Info.accumulations);


    free(TI->sendBuf);
    free(TI->sendCnts);
    free(TI->sDispls);
    free(TI->recvBuf);
    free(TI->recvCnts);
    free(TI->rDispls);

    free(TI->mpiRequests);
    free(TI->mpiStatusList);

    fftw_free(TI->fftBuf);
    fftw_free(TI->compBuf);
    fftw_free(TI->transBuf);

    fftw_destroy_plan(TI->tyFFTN0);
    fftw_destroy_plan(TI->tyIFFTN0);
    fftw_destroy_plan(TI->tyFFTN1);
    fftw_destroy_plan(TI->tyIFFTN1);
    fftw_destroy_plan(TI->tyFFTN2);
    fftw_destroy_plan(TI->tyIFFTN2);

    fftw_destroy_plan(TI->myFFTN0);
    fftw_destroy_plan(TI->myIFFTN0);
    fftw_destroy_plan(TI->myFFTN1);
    fftw_destroy_plan(TI->myIFFTN1);
    fftw_destroy_plan(TI->myFFTN2);
    fftw_destroy_plan(TI->myIFFTN2);
}

/*
 * Toeplitz matrix-vector multiplication 
 * Input:
 *      lam_t       : eigenvalues of a Toeplitz matrix
 *      x           : input vector
 *      TI          : system configuration
 * Output:
 *      y           : output vector
 */
void toep_solver_matvec_v15(complex * lam_t, complex * x, struct ToeplitzSolverInfo_v15 * TI, complex * y)
{


    /*
     * x
     * zeropadding and fft on n2
     */
    for (long i=0; i < TI->n0Info.local * TI->n1ColInfo.local; i ++)
    {
        toepZeroPad_v15(x + i * TI->lvlGrain[1], TI->fftBuf, TI->rank - 2, TI->ns + 2);
        fftw_execute(TI->tyFFTN2);
        memcpy(TI->compBuf + i * TI->pwrLvlGrain[1], TI->fftBuf, sizeof(fftw_complex) * TI->pwrLvlGrain[1]);
    }


    /*
     * x
     * row-wise transposition, pwrN2 -> n1
     */
    transpose_v15(TI->compBuf, TI->n0Info, TI->pwrN2Info, TI->n1ColInfo, TRANS_MODE_ROW, TI, TI->transBuf);


    /*
     * x
     * zeropadding and fft on n1
     */
    for (long i=0; i < TI->n0Info.local * TI->pwrN2Info.local; i ++)
    {
        toepZeroPad_v15(TI->transBuf + i * TI->ns[1], TI->fftBuf, 1, &TI->ns[1]);
        fftw_execute(TI->tyFFTN1);
        memcpy(TI->compBuf + i * TI->ns2[1], TI->fftBuf, sizeof(fftw_complex) * TI->ns2[1]);
    }

    /*
     * x
     * column-wise transposition, pwrN1 -> n0
     */
    transpose_v15(TI->compBuf, TI->pwrN2Info, TI->pwrN1RowInfo, TI->n0Info, TRANS_MODE_COL, TI, TI->transBuf);


    /*
     * x
     * zeropadding and fft on n0
     */
    for (long i=0; i < TI->pwrN1RowInfo.local * TI->pwrN2Info.local; i ++)
    {
        toepZeroPad_v15(TI->transBuf + i * TI->ns[0], TI->fftBuf, 1, &TI->ns[0]);
        fftw_execute(TI->tyFFTN0);
        memcpy(TI->compBuf + i * TI->ns2[0], TI->fftBuf, sizeof(fftw_complex) * TI->ns2[0]);
    }


/****************************** 
 * fft(x) .* lam_t and calculate dot product
 *  ******************************/ 


    /* lam_t .* fft(x) */
    for (long i=0; i < TI->pwrN2Info.local* TI->pwrN1RowInfo.local  * TI->ns2[0]; i ++)
        TI->compBuf[i] *= lam_t[i];


/****************************** 
 * Retrieve result
 *  ******************************/ 


    /*
     * tx
     * n0 ifft and retrieval
     */
    for (long i = 0; i < TI->pwrN1RowInfo.local * TI->pwrN2Info.local; i ++)
    {
        memcpy(TI->fftBuf, TI->compBuf + i * TI->ns2[0], sizeof(fftw_complex) * TI->ns2[0]);
        fftw_execute(TI->tyIFFTN0);
        for (long j = 0; j < TI->ns[0]; j ++)
            TI->transBuf[i*TI->ns[0] + j] = TI->fftBuf[j]/ TI->ns2[0];
    }


    /*
     * tx
     * column-wise backward transposition, n0 -> pwrN1
     */
    transpose_v15(TI->transBuf, TI->pwrN2Info, TI->n0Info, TI->pwrN1RowInfo, TRANS_MODE_COL, TI, TI->compBuf);


    /*
     * tx
     * n1 ifft and retrieval
     */
    for (long i=0; i < TI->n0Info.local * TI->pwrN2Info.local; i ++)
    {
        memcpy(TI->fftBuf, TI->compBuf + i * TI->ns2[1], sizeof(fftw_complex) * TI->ns2[1]);
        fftw_execute(TI->tyIFFTN1);
        for (long j = 0; j < TI->ns[1]; j ++)
            TI->transBuf[i * TI->ns[1] + j] = TI->fftBuf[j]/ TI->ns2[1];
    }


    /*
     * tx
     * row-wise backward transposition, n1 -> pwrN2
     */
    transpose_v15(TI->transBuf, TI->n0Info, TI->n1ColInfo, TI->pwrN2Info, TRANS_MODE_ROW, TI, TI->compBuf);


    /*
     * n2 ifft and retrieval
     */
    for (long i=0; i < TI->n0Info.local * TI->n1ColInfo.local; i ++)
    {
        memcpy(TI->fftBuf, TI->compBuf + i * TI->pwrLvlGrain[1], sizeof(fftw_complex) * TI->pwrLvlGrain[1]);
        fftw_execute(TI->tyIFFTN2);
        for (long j = 0; j < TI->pwrLvlGrain[1]; j ++)
            TI->fftBuf[j] /= TI->pwrLvlGrain[1];
        toepRetrv_v15(TI->fftBuf, y + i * TI->lvlGrain[1], TI->rank - 2, TI->ns + 2);
    }

}

/* 
 * Toeplitz vector precondition 
 * Input:
 *      lam_m       : eigenvalues of the preconditioner matrix
 *      x           : input vector
 *      TI          : system configuration
 * Output:
 *      y           : output vector
 */
void toep_solver_precon_v15(complex * lam_m, complex * x, struct ToeplitzSolverInfo_v15 * TI, complex * y)
{
    /*
     * Start of preconditioning
     */
    /*
     * precon, x
     * zeropadding and fft on n2
     */
    for (long i=0; i < TI->n0Info.local * TI->n1ColInfo.local; i ++)
    {
        memcpy(TI->fftBuf, x + i * TI->lvlGrain[1], sizeof(fftw_complex) * TI->lvlGrain[1]);
        fftw_execute(TI->myFFTN2);
        memcpy(TI->compBuf + i * TI->lvlGrain[1], TI->fftBuf, sizeof(fftw_complex) * TI->lvlGrain[1]);
    }



    /*
     * precon, x
     * row-wise transposition, n2 -> n1
     */
    transpose_v15(TI->compBuf, TI->n0Info, TI->n2Info, TI->n1ColInfo, TRANS_MODE_ROW, TI, TI->transBuf);


    /*
     * precon, x
     * zeropad and fft on n1
     */
    for (long i=0; i < TI->n0Info.local * TI->n2Info.local; i ++)
    {
        memcpy(TI->fftBuf, TI->transBuf + i * TI->ns[1], sizeof(fftw_complex) * TI->ns[1]);
        fftw_execute(TI->myFFTN1);
        memcpy(TI->compBuf + i * TI->ns[1], TI->fftBuf, sizeof(fftw_complex) * TI->ns[1]);
    }


    /*
     * precon, x
     * column-wise transposition, n1 -> n0
     */
    transpose_v15(TI->compBuf, TI->n2Info, TI->n1RowInfo, TI->n0Info, TRANS_MODE_COL, TI, TI->transBuf);


    /*
     * precon, x
     * zeropadding and fft on n0
     */

    for (long i=0; i < TI->n1RowInfo.local * TI->n2Info.local; i ++)
    {
        memcpy(TI->fftBuf, TI->transBuf + i * TI->ns[0], sizeof(fftw_complex) * TI->ns[0]);
        fftw_execute(TI->myFFTN0);
        memcpy(TI->compBuf + i * TI->ns[0], TI->fftBuf, sizeof(fftw_complex) * TI->ns[0]);
    }

/****************************** 
 * fft(x) ./ lam_m and calculate dot product
 *  ******************************/ 

    /* lam_m^(-1) .* fft(x) */
    for (long i=0; i < TI->n2Info.local * TI->n1RowInfo.local  * TI->ns[0]; i ++)
        TI->compBuf[i] /= lam_m[i];



/****************************** 
 * Retrieve result
 *  ******************************/ 


    /*
     * precon, x/m
     * n0 ifft and retrieval
     */
    for (long i = 0; i < TI->n1RowInfo.local * TI->n2Info.local; i ++)
    {
        memcpy(TI->fftBuf, TI->compBuf + i * TI->ns[0], sizeof(fftw_complex) * TI->ns[0]);
        fftw_execute(TI->myIFFTN0);
        for (long j = 0; j < TI->ns[0]; j ++)
            TI->transBuf[i*TI->ns[0] + j] = TI->fftBuf[j]/ TI->ns[0];
    }


    /*
     * precon, x/m
     * column-wise backward transposition, n0 -> N1
     */
    transpose_v15(TI->transBuf, TI->n2Info, TI->n0Info, TI->n1RowInfo, TRANS_MODE_COL, TI, TI->compBuf);

    
    /*
     * precon, x/m
     * n1 ifft and retrieval
     */
    for (long i=0; i < TI->n0Info.local * TI->n2Info.local; i ++)
    {
        memcpy(TI->fftBuf, TI->compBuf + i * TI->ns[1], sizeof(fftw_complex) * TI->ns[1]);
        fftw_execute(TI->myIFFTN1);
        for (long j = 0; j < TI->ns[1]; j ++)
            TI->transBuf[i * TI->ns[1] + j] = TI->fftBuf[j]/ TI->ns[1];
    }


    /*
     * precon, x/m
     * row-wise backward transposition, n1 -> n2
     */
    transpose_v15(TI->transBuf, TI->n0Info, TI->n1ColInfo, TI->n2Info, TRANS_MODE_ROW, TI, TI->compBuf);


    /*
     * precon, x/m
     * n2 ifft and retrieval
     */
    for (long i=0; i < TI->n0Info.local * TI->n1ColInfo.local; i ++)
    {
        memcpy(TI->fftBuf, TI->compBuf + i * TI->lvlGrain[1], sizeof(fftw_complex) * TI->lvlGrain[1]);
        fftw_execute(TI->myIFFTN2);
        for (long j = 0; j < TI->lvlGrain[1]; j ++)
            TI->fftBuf[j] /= TI->lvlGrain[1];
        memcpy(y + i * TI->lvlGrain[1], TI->fftBuf, sizeof(fftw_complex) * TI->lvlGrain[1]);
    }
}


/*
 * Calculate ||b - Tx||, the norm of residual
 * Input:
 *      lam_t       : eigenvalues of a Toeplitz matrix
 *      x           : unknown variable vector in Tx = b
 *      b           : RHS of Tx = b
 *      TI          : system configuration
 * Output:
 *      rho(return) : norm of residual
 */
double toep_solver_residual_v15(complex * lam_t, complex * x, complex * b, struct ToeplitzSolverInfo_v15 * TI)
{
    complex * tx;
    complex * r;
    tx = (complex *) malloc (sizeof(complex) * TI->allocLocal);
    r = (complex *) malloc (sizeof(complex) * TI->allocLocal);

    toep_solver_matvec_v15(lam_t, x, TI, tx);

    for (long i = 0; i < TI->n0Info.local * TI->n1ColInfo.local * TI->lvlGrain[1]; i ++) r[i] = b[i] - tx[i];

    complex localRes, globalRes;
    localRes = vectorCProd_v15(r, r, TI->n0Info.local * TI->n1ColInfo.local * TI->lvlGrain[1]);

    MPI_Allreduce(&localRes, &globalRes, 1, MPI_DOUBLE_COMPLEX, MPI_SUM, TI->mpiComm);

    /*
     * Cleaning up
     */
    free(tx);
    free(r);

    return creal(csqrt(globalRes));
}


/*
 * SECTION: Toeplitz subroutines
 */
/* 
 * Embed a Toeplitz vector into a circulant vector 
 * Input:
 *      src         : source buffer
 *      rank        : level of the Toeplitz matrix
 *      ns          : dimensionality of the Toeplitz matrix
 * Output:
 *      dest        : destination buffer
 */
void toepEmbd_v15(complex * src, complex * dest, const int rank, const int * ns)
{
    /*
     * function's own version of lvlGrain and pwrLvlGrain. This is needed because
     *  we break the embed/zero-pad/retrieve process in parts.
     */
    long * lvlGrain;                            /* granularity of levels */
    long * pwrLvlGrain;                         /* granularity of expanded levels */
    /*int n;*/                                      /* length of the original vector t */
    int N;                                      /* length of the fully embedded vector t. N = n * 2^rank */

    /*
     * Calculate lvlGrain, pwrLvlGrain, n, and N.
     */
    lvlGrain = (long *) malloc(sizeof(long) * rank);
    pwrLvlGrain = (long *) malloc(sizeof(long) * rank);
    lvlGrain[rank-1] = 1;
    pwrLvlGrain[rank-1] = 1;
    for (int i = rank-2; i >=0; i --)
    {
        lvlGrain[i] = lvlGrain[i+1] * ns[i+1];
        pwrLvlGrain[i] = pwrLvlGrain[i + 1] * ns[i+1] * 2;
    }
    /*n = ns[0] * lvlGrain[0];*/
    N = ns[0] * 2 * pwrLvlGrain[0];

    /*
     * Calculate the corresponding source based on the destination
     */

    int * srcInds = (int *) malloc ( sizeof(int) * rank);
    int zeroFlag;
    for (long destOff = 0; destOff < N; destOff ++)
    {
        long destRem = destOff;
        zeroFlag = 0;
        for (int lvl = 0; lvl < rank; lvl++)
        {
            srcInds[lvl] = destRem / pwrLvlGrain[lvl];
            /* if we should pad zero at destOff */
            if(srcInds[lvl] == ns[lvl])
            {
                zeroFlag = 1;
                break;
            }
            /* if it is a reflection */
            if (srcInds[lvl] > ns[lvl])
            {
                srcInds[lvl] = ns[lvl] * 2 - srcInds[lvl];
            }

            destRem = destRem % pwrLvlGrain[lvl];
        }

        if ( zeroFlag != 0 )
        {
            /* pad zero */
            dest[destOff] = 0;
        } else
        {
            long srcOff = 0;
            for (int lvl=0; lvl<rank; lvl++)
            {
                srcOff += srcInds[lvl] * lvlGrain[lvl];
            }

            dest[destOff] = src[srcOff];
        }
    }

    /*
     * final clean-up
     */
    free(lvlGrain);
    free(pwrLvlGrain);
    free(srcInds);
}

/* 
 * Zero-pad a vector to prepare it for Toeplitz multiplication
 * Input:
 *      src         : source buffer
 *      rank        : level of the Toeplitz matrix
 *      ns          : dimensionality of the Toeplitz matrix
 * Output:
 *      dest        : destination buffer
 */
void toepZeroPad_v15(complex * src, complex * dest, const int rank, const int * ns)
{
    /*
     * function's own version of lvlGrain and pwrLvlGrain. This is needed because
     *  we break the embed/zero-pad/retrieve process in parts.
     */
    long * lvlGrain;                            /* granularity of levels */
    long * pwrLvlGrain;                         /* granularity of expanded levels */
    /*int n;*/                                      /* length of the original vector t */
    int N;                                      /* length of the fully embedded vector t. N = n * 2^rank */

    /*
     * Calculate lvlGrain, pwrLvlGrain, n, and N.
     */
    lvlGrain = (long *) malloc(sizeof(long) * rank);
    pwrLvlGrain = (long *) malloc(sizeof(long) * rank);
    lvlGrain[rank-1] = 1;
    pwrLvlGrain[rank-1] = 1;
    for (int i = rank-2; i >=0; i --)
    {
        lvlGrain[i] = lvlGrain[i+1] * ns[i+1];
        pwrLvlGrain[i] = pwrLvlGrain[i + 1] * ns[i+1] * 2;
    }
    /*n = ns[0] * lvlGrain[0];*/
    N = ns[0] * 2 * pwrLvlGrain[0];

    /*
     * Calculate the corresponding source based on the destination
     */

    int * srcInds = (int *) malloc ( sizeof(int) * rank);
    int zeroFlag;
    for (long destOff = 0; destOff < N; destOff ++)
    {
        long destRem = destOff;
        zeroFlag = 0;
        for (int lvl = 0; lvl < rank; lvl++)
        {
            srcInds[lvl] = destRem / pwrLvlGrain[lvl];
            /* if we should pad zero at destOff */
            if(srcInds[lvl] >= ns[lvl])
            {
                zeroFlag = 1;
                break;
            }

            destRem = destRem % pwrLvlGrain[lvl];
        }

        if ( zeroFlag != 0 )
        {
            /* pad zero */
            dest[destOff] = 0;
        } else
        {
            long srcOff = 0;
            for (int lvl=0; lvl<rank; lvl++)
            {
                srcOff += srcInds[lvl] * lvlGrain[lvl];
            }

            dest[destOff] = src[srcOff];
        }
    }

    /*
     * final clean-up
     */
    free(lvlGrain);
    free(pwrLvlGrain);
    free(srcInds);
}

/* 
 * Retrieve the original vector from its Toeplitz embedding
 * Input:
 *      src         : source buffer
 *      rank        : level of the Toeplitz matrix
 *      ns          : dimensionality of the Toeplitz matrix
 * Output:
 *      dest        : destination buffer
 */
void toepRetrv_v15(complex * src, complex * dest, const int rank, const int * ns)
{
    /*
     * function's own version of lvlGrain and pwrLvlGrain. This is needed because
     *  we break the embed/zero-pad/retrieve process in parts.
     */
    long * lvlGrain;                            /* granularity of levels */
    long * pwrLvlGrain;                         /* granularity of expanded levels */
    /*int n;*/                                      /* length of the original vector t */
    int N;                                      /* length of the fully embedded vector t. N = n * 2^rank */

    /*
     * Calculate lvlGrain, pwrLvlGrain, n, and N.
     */
    lvlGrain = (long *) malloc(sizeof(long) * rank);
    pwrLvlGrain = (long *) malloc(sizeof(long) * rank);
    lvlGrain[rank-1] = 1;
    pwrLvlGrain[rank-1] = 1;
    for (int i = rank-2; i >=0; i --)
    {
        lvlGrain[i] = lvlGrain[i+1] * ns[i+1];
        pwrLvlGrain[i] = pwrLvlGrain[i + 1] * ns[i+1] * 2;
    }
    /*n = ns[0] * lvlGrain[0];*/
    N = ns[0] * 2 * pwrLvlGrain[0];

    /*
     * Calculate the corresponding source based on the destination
     */

    int * destInds = (int *) malloc ( sizeof(int) * rank);
    int zeroFlag;
    for (long srcOff = 0; srcOff < N; srcOff ++)
    {
        long srcRem = srcOff;
        zeroFlag = 0;
        for (int lvl = 0; lvl < rank; lvl++)
        {
            destInds[lvl] = srcRem / pwrLvlGrain[lvl];
            /* a part that we should ignore */
            if(destInds[lvl] >= ns[lvl])
            {
                zeroFlag = 1;
                break;
            }
            srcRem = srcRem % pwrLvlGrain[lvl];
        }

        if ( zeroFlag != 0 )
        {
            /* do nothing */
        } else
        {
            long destOff = 0;
            for (int lvl=0; lvl<rank; lvl++)
            {
                destOff += destInds[lvl] * lvlGrain[lvl];
            }

            dest[destOff] = src[srcOff];
        }
    }

    /*
     * final clean-up
     */
    free(lvlGrain);
    free(pwrLvlGrain);
    free(destInds);
}

/* 
 * Compute the precondition vector. The calculation is in-place 
 * Input:
 *      srcdest     : source buffer
 *      rank        : level of the Toeplitz matrix
 *      ns          : dimensionality of the Toeplitz matrix
 * Output:
 *      srcdest     : destination buffer
 */
void toepPreVec_v15(complex * srcdest, const int rank, const int * ns)
{
    /*
     * function's own version of lvlGrain and pwrLvlGrain. This is needed because
     *  we break the embed/zero-pad/retrieve process in parts.
     */
    long * lvlGrain;                            /* granularity of levels */
    int n;                                      /* length of the original vector t */

    /*
     * Calculate lvlGrain, pwrLvlGrain, n, and N.
     */
    lvlGrain = (long *) malloc(sizeof(long) * rank);
    lvlGrain[rank-1] = 1;
    for (int i = rank-2; i >=0; i --)
    {
        lvlGrain[i] = lvlGrain[i+1] * ns[i+1];
    }
    n = ns[0] * lvlGrain[0];


    /* for each level, from innermost to outermost */
    for (int lvl = rank-1; lvl >=0 ; lvl --)
    {
        /* for each block in that level */
        for(long blkCount = 0; blkCount < n / lvlGrain[lvl] / ns[lvl]; blkCount ++  )
        {
            long blkOff = blkCount * lvlGrain[lvl] * ns[lvl];
            /* for each subblock inside the block */
            for(long blkNum = 1; blkNum < (ns[lvl] + 1)/2; blkNum ++)
            {
                double avg;
                /* for each element */
                for (long i = 0; i < lvlGrain[lvl]; i++)
                {
                    avg = (ns[lvl] * 1.0 - blkNum)/ns[lvl] * srcdest[blkOff + blkNum * lvlGrain[lvl] + i] +
                            blkNum * 1.0/ns[lvl] * srcdest[blkOff + (ns[lvl] - blkNum) * lvlGrain[lvl] + i];
                    srcdest[blkOff + blkNum * lvlGrain[lvl] + i] = avg;
                    srcdest[blkOff + (ns[lvl] - blkNum) * lvlGrain[lvl] + i] = avg;
                }
            }
        }
    }

    /*
     * final clean-up
     */
    free(lvlGrain);
}



/*
 * SECTION: Vector operation subroutines
 */
/* compute the inner product of two vectors */
complex vectorCProd_v15(complex *v1, complex *v2, long length)
{
    complex sum = 0;
    for (long i = 0; i < length; i ++)
    {
        sum += v1[i] * conj(v2[i]);
    }
    return sum;
}

/* randomize the vector. Each element will be in range [0, 1) */
void vectorCRand_v15(complex *v, long length)
{
    for (long i=0; i < length; i ++)
    {
        v[i] = rand() * 1.0 / RAND_MAX;
    }
}

/*
 * SECTION: Miscellaneous routines
 */
void transpose_v15(complex * srcBuf, struct DimensionInfo_v15 fixedDim, struct DimensionInfo_v15 splitDim, 
               struct DimensionInfo_v15 joinDim, int transMode, struct ToeplitzSolverInfo_v15 * TI, 
               complex * destBuf)
{
    int splitProcs;
    /*int splitId;*/
    MPI_Comm splitComm;
    /*
     * transMode explained:
     * TRANS_MODE_ROW. For n2 <-> n1, pwrN2 <-> n1
     * TRANS_MODE_COL. For n1 <-> n0, pwrN1 <-> n0
     */
    if (transMode == TRANS_MODE_ROW) { 
        splitProcs = TI->colprocs;
        /*splitId = TI->colid;*/
        splitComm = TI->rowComm;
    } else { 
        splitProcs = TI->rowprocs;
        /*splitId = TI->rowid;*/
        splitComm = TI->colComm;
    }
    /* prepare for the transposition  */
    for (int proc = 0; proc < splitProcs; proc ++) {
        TI->sendCnts[proc] = fixedDim.local * splitDim.partitions[proc] * joinDim.local;
        TI->sDispls[proc] = fixedDim.local * splitDim.accumulations[proc] * joinDim.local;
      
        /* JC remove
        if (transMode == TRANS_MODE_ROW) {
            for (long i=0; i < fixedDim.local; i ++)
                for (long j = 0; j < joinDim.local; j ++)
                    memcpy(TI->sendBuf + fixedDim.local * joinDim.local * splitDim.accumulations[proc] +
                                (i * joinDim.local + j) * splitDim.partitions[proc],
                           srcBuf + (i * joinDim.local + j) * splitDim.size + splitDim.accumulations[proc],
                           sizeof(complex) * splitDim.partitions[proc]);
        } else {
            for (long i = 0; i < joinDim.local; i ++)
                for (long j = 0; j < fixedDim.local; j ++)
                    memcpy(TI->sendBuf + splitDim.accumulations[proc] * joinDim.local * fixedDim.local + (i*fixedDim.local + j)*splitDim.partitions[proc],
                            srcBuf + (i * fixedDim.local + j) * splitDim.size + splitDim.accumulations[proc],
                            sizeof(complex) * splitDim.partitions[proc]
                          );
        }

        and replace by the following:
        */

        long sendBufDispls = splitDim.accumulations[proc] * joinDim.local * fixedDim.local;
        long srcBufDispls = splitDim.accumulations[proc];
        for (long i = 0; i < fixedDim.local * joinDim.local; i++) {
          memcpy(TI->sendBuf + sendBufDispls,
                 srcBuf + srcBufDispls,
                 sizeof(complex) * splitDim.partitions[proc]);
          sendBufDispls += splitDim.partitions[proc];
          srcBufDispls += splitDim.size;
        }

        TI->recvCnts[proc] = fixedDim.local * joinDim.partitions[proc] * splitDim.local;
        TI->rDispls[proc] = fixedDim.local * joinDim.accumulations[proc] * splitDim.local;
    }
    

#if ALT_ALLTOALLV
    int mpiError = MPI_SUCCESS;
    char errorString[255];
    int errorStringLength;
    int splitId;
    if (transMode == TRANS_MODE_ROW) { 
        splitId = TI->colid;
    } else { 
        splitId = TI->rowid;
    }

    /* alternative to MPI_Alltoallv */
    const int kAlltoallvTag = 1001;
    for (int proc_offset = 0; proc_offset < splitProcs; proc_offset ++) {
        int proc = (splitId + proc_offset) % splitProcs;
        mpiError = MPI_Irecv(TI->recvBuf + TI->rDispls[proc], TI->recvCnts[proc], MPI_DOUBLE_COMPLEX,
                proc, kAlltoallvTag, splitComm, &TI->mpiRequests[splitProcs + proc]);
        if (mpiError) {
            MPI_Error_string(mpiError, errorString, &errorStringLength);
            perror(errorString);
        }
        mpiError = MPI_Isend(TI->sendBuf + TI->sDispls[proc], TI->sendCnts[proc], MPI_DOUBLE_COMPLEX,
                proc, kAlltoallvTag, splitComm, &TI->mpiRequests[proc]);
        if (mpiError) {
            MPI_Error_string(mpiError, errorString, &errorStringLength);
            perror(errorString);
        }
    }
    mpiError = MPI_Waitall(2 * splitProcs, TI->mpiRequests, TI->mpiStatusList);
    if (mpiError == MPI_ERR_IN_STATUS) {
        for (int i=0; i < 2 * splitProcs; i ++) {
            if (TI->mpiStatusList[i].MPI_ERROR != MPI_SUCCESS) {
                MPI_Error_string(mpiError, errorString, &errorStringLength);
                perror(errorString);
            }
        }
    }
                       
#else
    MPI_Alltoallv(
            TI->sendBuf, TI->sendCnts, TI->sDispls, MPI_DOUBLE_COMPLEX,
            TI->recvBuf, TI->recvCnts, TI->rDispls, MPI_DOUBLE_COMPLEX,
            splitComm
        );
#endif


    /* pad data */
    /* JC: remove
    if (transMode == TRANS_MODE_ROW) {
        for (int proc = 0; proc < splitProcs; proc ++) 
            for (long i = 0; i < fixedDim.local; i ++) 
                for (long j = 0; j < splitDim.local; j ++) 
                    for (long k = 0; k < joinDim.partitions[proc]; k ++)
                        destBuf[(i * splitDim.local + j) * joinDim.size + joinDim.accumulations[proc] + k] = 
                            TI->recvBuf[(fixedDim.local * joinDim.accumulations[proc] + i *  joinDim.partitions[proc] +
                                    k) * splitDim.local + j];
    } else {
        for (int proc = 0; proc < splitProcs; proc ++)
            for (long i = 0; i < joinDim.partitions[proc]; i++)
                for (long j = 0; j < splitDim.local; j ++)
                    for (long k = 0; k < fixedDim.local; k ++)
                        destBuf[(j*fixedDim.local + k) * joinDim.size + joinDim.accumulations[proc] + i] =
                            TI->recvBuf[splitDim.local * fixedDim.local * joinDim.accumulations[proc] + 
                                (i * fixedDim.local + k) * splitDim.local + j];
    }

    and replace by the following:
    */
    for (int proc = 0; proc < splitProcs; proc ++) {
      complex *recvBufPtr = TI->recvBuf + fixedDim.local * splitDim.local * joinDim.accumulations[proc];
      if (transMode == TRANS_MODE_ROW) {
        for (long i = 0; i < fixedDim.local; i ++) 
          for (long j = 0; j < joinDim.partitions[proc]; j ++)
            for (long k = 0; k < splitDim.local; k ++) {
              destBuf[(i * splitDim.local + k) * joinDim.size + joinDim.accumulations[proc] + j] = *recvBufPtr++;
            }
      }
      else {
        for (long i = 0; i < joinDim.partitions[proc]; i ++)
          for (long j = 0; j < fixedDim.local; j ++)
            for (long k = 0; k < splitDim.local; k ++)
              destBuf[(k * fixedDim.local + j) * joinDim.size + joinDim.accumulations[proc] + i] = *recvBufPtr++;
      }
    }

}

