/*
 * toep_solver_v2.5.h
 * MPI_CG version 2.5
 *
 * Author: Tom Li
 * Created on: July, 19, 2012
 */
/*
 * Edited by JC:
 * 1. Append _v25 to all the data structures and functions
 * 2. Change "complex" to "double complex"
 */

#ifndef TOEP_SOLVER_v25_H_
#define TOEP_SOLVER_v25_H_

#include "mpi.h"
#include <fftw3.h>

/*
 * Preprocessor switches
 */
//#define DEBUG_PRINT_ENABLED 1	    /* uncomment to enable DEBUG statements */
//#define DEBUG_ID 1
//#define ALT_ALLTOALLV 3           /* uncooment to use alternative to MPI_Alltoallv() calls */

#ifdef __cplusplus
    #include <complex>
    typedef std::complex<double> double_complex;
#else
    #include <complex.h>
    typedef double complex double_complex;
#endif


/*
 * Dimension information. Used mainly in transpositions.
 */
struct DimensionInfo_v25{
    long * partitions;
    long local;
    long localStart;
    long * accumulations;
    long size;
};

struct ToeplitzSolverInfo_v25{
    /*
     * SECTION: Declaration of FFTW stuffs. 
     */
    fftw_plan tyFFTN2;                          /* plan to compute (rank-22d fft for Ty = z */
    fftw_plan tyIFFTN2;                         /* plan to compute (rank-2)d ifft for Ty = z */
    fftw_plan tyFFTN1;                          /* plan to compute 1d fft on n1 for Ty = z */
    fftw_plan tyIFFTN1;                         /* plan to compute 1d ifft on n1 for Ty = z */
    fftw_plan tyFFTN0;                          /* plan to compute 1d fft on n0 for Ty = z */
    fftw_plan tyIFFTN0;                         /* plan to compute 1d ifft on n0 for Ty = z */

    fftw_plan myFFTN2;                          /* plan to compute (rank-22d fft for M^(-1)y = z */
    fftw_plan myIFFTN2;                         /* plan to compute (rank-2)d ifft for M^(-1)y = z */
    fftw_plan myFFTN1;                          /* plan to compute 1d fft on n1 for M^(-1)y = z */
    fftw_plan myIFFTN1;                         /* plan to compute 1d ifft on n1 for M^(-1)y = z */
    fftw_plan myFFTN0;                          /* plan to compute 1d fft on n0 for M^(-1)y = z */
    fftw_plan myIFFTN0;                         /* plan to compute 1d ifft on n0 for M^(-1)y = z */

    fftw_complex * fftBuf;                      /* the input and output buffer */
    fftw_complex * transBuf;                    /* the generic transposition buffer */
    fftw_complex * compBuf;                     /* the computation buffer */
    fftw_complex piggyBuf[2];                   /* piggy backing buffer */

    struct DimensionInfo_v25 n0Info, n1ColInfo, n1RowInfo, n2Info;      /* original dimension info sets */
    struct DimensionInfo_v25 pwrN1ColInfo, pwrN1RowInfo, pwrN2Info;     /* extended dimension info sets */

    /*
     * Transposition variables
     */
    double_complex * sendBuf;                          /* send buffer for MPI */
    int * sendCnts;                             /* send count vector for MPI_Alltoallv */
    int * sDispls;                              /* send displacement for MPI_Alltoallv */
    double_complex * recvBuf;                          /* receive buffer for MPI */
    int * recvCnts;                             /* receive count vector for MPI_Alltoallv */
    int * rDispls;                              /* receive displacement for MPI_Alltoallv */

    long allocLocal;                            /* memory allocation size for unexpanded FFT */
    long pwrAllocLocal;                         /* memory allocation size for expanded FFT */

    int myid;                                   /* process id of this machine */
    int colid;                                  /* column index in the processor raster */
    int rowid;                                  /* row index in the processor raster */
    int numprocs;                               /* number of process in the communicator */
    int rowprocs;                               /* number of row processors */
    int colprocs;                               /* number of column processors */

    MPI_Comm mpiComm;                           /* MPI communicator */
    MPI_Comm rowComm;                           /* MPI communicator for operation on the same row */
    MPI_Comm colComm;                           /* MPI communicator for operation on the same column */

    MPI_Request  * mpiRequests;                 /* MPI request lists for isend/irecv */
    MPI_Status * mpiStatusList;                 /* MPI status list for isend/irecv */

    /*
     * Dimension arrays 
     */ 
    int rank;                                   /* number of levels */
    int * ns;                                   /* level information */
    int n;                                      /* length of the original vector t */
    int N;                                      /* length of the fully embedded vector t. N = n * 2^rank */
    int * ns2;                                  /* ns2 = ns * 2 */
    long * lvlGrain;                            /* granularity of levels */
    long * pwrLvlGrain;                         /* granularity of expanded levels */
    
};

#ifdef __cplusplus
    extern "C" {
#endif

/*
 * System interface. This are the functions users should call.
 */
/* 
 * Systen initialization 
 * Input:
 *      mpiComm     : global communicator
 *      rank        : number of levels in the system
 *      ns          : dimensionality array
 * Output:
 *      TI          : system configuration
 */
#define TOEP_INIT_SUCCESS 0
#define TOEP_INIT_WRONG_ROW_NUM 1
int toep_solver_init_v25(MPI_Comm mpiComm, int rowprocs, int rank, int * ns, struct ToeplitzSolverInfo_v25 * TI);

/*
 * Planning function. 
 * Input:
 *      t           : a Toeplitz vector. The first column of a Toeplitz matrix
 *      TI          : system configuration 
 * OutPut:
 *      lam_t       : the eigenvalues of the embedded Toeplitz matrix
 *      lam_m       : the eigenvalues of the preconditioner matrix
 */
void toep_solver_plan_v25(double_complex * t, struct ToeplitzSolverInfo_v25 * TI, double_complex * lam_t, double_complex * lam_m);

/*
 * Solving function.
 * Input:
 *      lam_t       : the eigenvalues of the embedded Toeplitz matrix
 *      lam_m       : the eigenvalues of the preconditioner matrix
 *      x           : the initial guess of x, the unknown vector
 *      b           : the RHS of the linear system Ax = b
 *      tol         : tolerance value
 *      max_iter    : maximum iteration
 *      TI          : system configuration 
 * Output:
 *      x           : solution of the linear ysstem Ax = b
 */
void toep_solver_solve_v25(double_complex * lam_t, double_complex * lam_m, double_complex *b,
                       double tol, int max_iter, struct ToeplitzSolverInfo_v25 * TI, double_complex * x);

/*
 * System finalization. It is used to free up allocated memory spaces.
 * Input:
 *      TI          : system configuration 
 */
void toep_solver_finalize_v25(struct ToeplitzSolverInfo_v25 * TI);


/*
 * Toeplitz matrix-vector multiplication 
 * Input:
 *      lam_t       : eigenvalues of a Toeplitz matrix
 *      x           : input vector
 *      TI          : system configuration
 * Output:
 *      y           : output vector
 */
void toep_solver_matvec_v25(double_complex * lam_t, double_complex * x, struct ToeplitzSolverInfo_v25 * TI, double_complex * y);


/*
 * Toeplitz matrix-vector multiplication, piggybacking self dot product of b
 * Input:
 *      lam_t       : eigenvalues of a Toeplitz matrix
 *      x           : input vector for matvec multiplication
 *      b           : input vector for computing self dot product
 *      TI          : system configuration
 * Output:
 *      y           : output vector
 *      product     : self dot product of b
 */
void toep_solver_matvec_piggy_dotprod_v25(double_complex * lam_t, double_complex * x, double_complex * b, struct ToeplitzSolverInfo_v25 * TI, 
                                      double_complex * y, double_complex * product);

/*
 * Toeplitz matrix-vector multiplication, compute xTx as by-product
 * Input:
 *      lam_t       : eigenvalues of a Toeplitz matrix
 *      x           : input vector
 *      TI          : system configuration
 * Output:
 *      y           : output vector
 *      product     : result of x^T T x
 */
void toep_solver_matvec_dotprod_v25(double_complex * lam_t, double_complex * x, struct ToeplitzSolverInfo_v25 * TI, 
                                double_complex * y, double_complex * product);

/* 
 * Toeplitz vector precondition 
 * Input:
 *      lam_m       : eigenvalues of a preconditioner matrix
 *      x           : input vector
 *      TI          : system configuration
 * Output:
 *      y           : output vector
 */
void toep_solver_precon_v25(double_complex * lam_m, double_complex * x, struct ToeplitzSolverInfo_v25 * TI, double_complex * y);

/* 
 * Toeplitz vector precondition, calculating dot product and test for
 * convergence during the process
 * Input:
 *      lam_m       : eigenvalues of a preconditioner matrix
 *      x           : input vector
 *      gammatol    : gamma * tolerance, used in convergence test
 *      TI          : system configuration
 * Output:
 *      y           : output vector
 *      product     : x^T M^(-1) x
 *      rho         : ||x||
 *      convFlag(return)    : convergence test flag
 */
int toep_solver_precon_dotprod_convtest_v25(double_complex * lam_m, double_complex * x, const double gammatol,
                                        struct ToeplitzSolverInfo_v25 * TI, double_complex * y,
                                        double_complex * product, double_complex * rho);

/*
 * Calculate ||b - Tx||, the norm of residual
 * Input:
 *      lam_t       : eigenvalues of a Toeplitz matrix
 *      x           : unknown variable vector in Tx = b
 *      b           : RHS of Tx = b
 *      TI          : system configuration
 * Output:
 *      rho(return) : norm of residual
 */
double toep_solver_residual_v25(double_complex * lam_t, double_complex * x, double_complex * b, struct ToeplitzSolverInfo_v25 * TI);

/*
 * Toeplitz subroutines
 */
/* Embed a Toeplitz vector into a circulant vector */
void toepEmbd_v25(double_complex * src, double_complex * dest, const int rank, const int * ns);

/* Zero-pad a vector to prepare it for Toeplitz multiplication */
void toepZeroPad_v25(double_complex * src, double_complex * dest, const int rank, const int * ns);

/* Retrieve the original vector from its Toeplitz embedding */
void toepRetrv_v25(double_complex * src, double_complex * dest, const int rank, const int * ns);

/* Compute the precondition vector. The calculation is in-place */
void toepPreVec_v25(double_complex * srcdest, const int rank, const int * ns);


/*
 * Vector operation subroutines
 */
/* compute the inner product of two vectors */
double_complex vectorCProd_v25(double_complex *v1, double_complex *v2, long length);

/* randomize the vector. Each element will be in range [0, 1) */
void vectorCRand_v25(double_complex *v, long length);

/*
 * Miscellaneous routines
 */
/*
 * transMode explained:
 * TRANS_MODE_ROW. For n2 <-> n1, pwrN2 <-> n1
 * TRANS_MODE_COL. For n1 <-> n0, pwrN1 <-> n0
 */
#define TRANS_MODE_COL 1
#define TRANS_MODE_ROW 2
void transposePiggy_v25(double_complex * srcBuf, struct DimensionInfo_v25 fixedDim, struct DimensionInfo_v25 splitDim, 
                    struct DimensionInfo_v25 joinDim, int transMode, long piggySize, double_complex * piggyInBuf, 
                    struct ToeplitzSolverInfo_v25 * TI, double_complex * destBuf, double_complex * piggyOutBuf);

#ifdef __cplusplus
    }
#endif

#endif /* TOEP_SOLVER_H_ */
