#include <math.h>
#include "Utility.hpp"
#include "ToepSolverWrapper.hpp"

//-------------------------------------------------------------------------
//-------------------------------------------------------------------------
//-------------------------------------------------------------------------
//---------------------------------------------------------- ToepSolver_v10
ToepSolver_v10::
ToepSolver_v10(int ndim, long *dim, int *nproc, MPI_Comm &comm) {
  int *idim = NULL;
  New_1D_Array<int, int>(&idim, ndim);
  int i;
  for (i = 0; i < ndim; i++) {
    idim[i] = dim[i];
  }
  toep_solver_init_v10(comm, ndim, idim, &TI);
  Delete_1D_Array<int>(&idim);
}

//-------------------------------------------------------------------------
ToepSolver_v10::
~ToepSolver_v10() {
  toep_solver_finalize_v10(&TI);
}

//-------------------------------------------------------------------------
void ToepSolver_v10::
GetPartitioning(long *dim_start, long *dim_len) {
  dim_start[0] = TI.n0Info.localStart;
  dim_len[0] = TI.n0Info.local;
  int i;
  for (i = 1; i < TI.rank; i++) {
    dim_start[i] = 0;
    dim_len[i] = TI.ns[i];
  }
}

//-------------------------------------------------------------------------
void ToepSolver_v10::
GetLens(long &len_toep_embed, long &len_precond) {
  len_toep_embed = TI.pwrAllocLocal * 2;
  len_precond = TI.allocLocal;
}

//-------------------------------------------------------------------------
void ToepSolver_v10::
Planning(double *toep, double_complex *toep_embed_eig,
         double_complex *prec_eig) {
  double_complex *t = NULL;
  New_1D_Array<double_complex, long>(&t, TI.allocLocal);
  long i;
  long len = TI.n0Info.local * TI.n1Info.size;
  for (i = 0; i < len; i++) {
    t[i] = toep[i];
  }
  toep_solver_plan_v10(t, &TI, toep_embed_eig, prec_eig);
  Delete_1D_Array<double_complex>(&t);
}

//-------------------------------------------------------------------------
void ToepSolver_v10::
MatMult(double_complex *toep_embed_eig, double *x, double *y) {
  double_complex *cx = NULL, *cy = NULL;
  New_1D_Array<double_complex, long>(&cx, TI.allocLocal);
  New_1D_Array<double_complex, long>(&cy, TI.allocLocal);
  long i;
  long len = TI.n0Info.local * TI.n1Info.size;
  for (i = 0; i < len; i++) {
    cx[i] = x[i];
  }
  toep_solver_matvec_v10(toep_embed_eig, cx, &TI, cy);
  for (i = 0; i < len; i++) {
    y[i] = real(cy[i]);
  }
  Delete_1D_Array<double_complex>(&cx);
  Delete_1D_Array<double_complex>(&cy);
}

//-------------------------------------------------------------------------
void ToepSolver_v10::
MatSolve(double_complex *toep_embed_eig, double_complex *prec_eig,
         double *b, double *x, double rtol, int maxit) {
  double_complex *cb = NULL, *cx = NULL;
  New_1D_Array<double_complex, long>(&cb, TI.allocLocal);
  New_1D_Array<double_complex, long>(&cx, TI.allocLocal);
  long i;
  long len = TI.n0Info.local * TI.n1Info.size;
  for (i = 0; i < len; i++) {
    cb[i] = b[i];
    cx[i] = x[i];
  }
  toep_solver_solve_v10(toep_embed_eig, prec_eig, cb, rtol, maxit, &TI, cx);
  for (i = 0; i < len; i++) {
    x[i] = real(cx[i]);
  }
  // Should really optimize the following out -----------------------
  double res = toep_solver_residual_v10(toep_embed_eig, cx, cb, &TI);
  double normb_local = 0.0, normb;
  for (i = 0; i < len; i++) {
    normb_local += b[i] * b[i];
  }
  MPI_Allreduce(&normb_local, &normb, 1, MPI_DOUBLE, MPI_SUM, TI.mpiComm);
  normb = sqrt(normb);
  if (res/normb > rtol) {
    throw std::string("ToepSolver_v10::MatSolve: Linear solve not converge!\n");
  }
  //-----------------------------------------------------------------
  Delete_1D_Array<double_complex>(&cb);
  Delete_1D_Array<double_complex>(&cx);
}

//-------------------------------------------------------------------------
//-------------------------------------------------------------------------
//-------------------------------------------------------------------------
//---------------------------------------------------------- ToepSolver_v15
ToepSolver_v15::
ToepSolver_v15(int ndim, long *dim, int *nproc, MPI_Comm &comm) {
  int *idim = NULL;
  New_1D_Array<int, int>(&idim, ndim);
  int i;
  for (i = 0; i < ndim; i++) {
    idim[i] = dim[i];
  }
  toep_solver_init_v15(comm, nproc[0], ndim, idim, &TI);
  Delete_1D_Array<int>(&idim);
}

//-------------------------------------------------------------------------
ToepSolver_v15::
~ToepSolver_v15() {
  toep_solver_finalize_v15(&TI);
}

//-------------------------------------------------------------------------
void ToepSolver_v15::
GetPartitioning(long *dim_start, long *dim_len) {
  dim_start[0] = TI.n0Info.localStart;
  dim_len[0] = TI.n0Info.local;
  dim_start[1] = TI.n1ColInfo.localStart;
  dim_len[1] = TI.n1ColInfo.local;
  int i;
  for (i = 2; i < TI.rank; i++) {
    dim_start[i] = 0;
    dim_len[i] = TI.ns[i];
  }
}

//-------------------------------------------------------------------------
void ToepSolver_v15::
GetLens(long &len_toep_embed, long &len_precond) {
  len_toep_embed = TI.pwrAllocLocal * 2;
  len_precond = TI.allocLocal;
}

//-------------------------------------------------------------------------
void ToepSolver_v15::
Planning(double *toep, double_complex *toep_embed_eig,
         double_complex *prec_eig) {
  double_complex *t = NULL;
  New_1D_Array<double_complex, long>(&t, TI.allocLocal);
  long i;
  long len = TI.n0Info.local * TI.n1ColInfo.local * TI.n2Info.size;
  for (i = 0; i < len; i++) {
    t[i] = toep[i];
  }
  toep_solver_plan_v15(t, &TI, toep_embed_eig, prec_eig);
  Delete_1D_Array<double_complex>(&t);
}

//-------------------------------------------------------------------------
void ToepSolver_v15::
MatMult(double_complex *toep_embed_eig, double *x, double *y) {
  double_complex *cx = NULL, *cy = NULL;
  New_1D_Array<double_complex, long>(&cx, TI.allocLocal);
  New_1D_Array<double_complex, long>(&cy, TI.allocLocal);
  long i;
  long len = TI.n0Info.local * TI.n1ColInfo.local * TI.n2Info.size;
  for (i = 0; i < len; i++) {
    cx[i] = x[i];
  }
  toep_solver_matvec_v15(toep_embed_eig, cx, &TI, cy);
  for (i = 0; i < len; i++) {
    y[i] = real(cy[i]);
  }
  Delete_1D_Array<double_complex>(&cx);
  Delete_1D_Array<double_complex>(&cy);
}

//-------------------------------------------------------------------------
void ToepSolver_v15::
MatSolve(double_complex *toep_embed_eig, double_complex *prec_eig,
         double *b, double *x, double rtol, int maxit) {
  double_complex *cb = NULL, *cx = NULL;
  New_1D_Array<double_complex, long>(&cb, TI.allocLocal);
  New_1D_Array<double_complex, long>(&cx, TI.allocLocal);
  long i;
  long len = TI.n0Info.local * TI.n1ColInfo.local * TI.n2Info.size;
  for (i = 0; i < len; i++) {
    cb[i] = b[i];
    cx[i] = x[i];
  }
  toep_solver_solve_v15(toep_embed_eig, prec_eig, cb, rtol, maxit, &TI, cx);
  for (i = 0; i < len; i++) {
    x[i] = real(cx[i]);
  }
  // Should really optimize the following out -----------------------
  double res = toep_solver_residual_v15(toep_embed_eig, cx, cb, &TI);
  double normb_local = 0.0, normb;
  for (i = 0; i < len; i++) {
    normb_local += b[i] * b[i];
  }
  MPI_Allreduce(&normb_local, &normb, 1, MPI_DOUBLE, MPI_SUM, TI.mpiComm);
  normb = sqrt(normb);
  if (res/normb > rtol) {
    throw std::string("ToepSolver_v15::MatSolve: Linear solve not converge!\n");
  }
  //-----------------------------------------------------------------
  Delete_1D_Array<double_complex>(&cb);
  Delete_1D_Array<double_complex>(&cx);
}

//-------------------------------------------------------------------------
//-------------------------------------------------------------------------
//-------------------------------------------------------------------------
//---------------------------------------------------------- ToepSolver_v20
ToepSolver_v20::
ToepSolver_v20(int ndim, long *dim, int *nproc, MPI_Comm &comm) {
  int *idim = NULL;
  New_1D_Array<int, int>(&idim, ndim);
  int i;
  for (i = 0; i < ndim; i++) {
    idim[i] = dim[i];
  }
  toep_solver_init_v20(comm, ndim, idim, &TI);
  Delete_1D_Array<int>(&idim);
}

//-------------------------------------------------------------------------
ToepSolver_v20::
~ToepSolver_v20() {
  toep_solver_finalize_v20(&TI);
}

//-------------------------------------------------------------------------
void ToepSolver_v20::
GetPartitioning(long *dim_start, long *dim_len) {
  dim_start[0] = TI.n0Info.localStart;
  dim_len[0] = TI.n0Info.local;
  int i;
  for (i = 1; i < TI.rank; i++) {
    dim_start[i] = 0;
    dim_len[i] = TI.ns[i];
  }
}

//-------------------------------------------------------------------------
void ToepSolver_v20::
GetLens(long &len_toep_embed, long &len_precond) {
  len_toep_embed = TI.pwrAllocLocal * 2;
  len_precond = TI.allocLocal;
}

//-------------------------------------------------------------------------
void ToepSolver_v20::
Planning(double *toep, double_complex *toep_embed_eig,
         double_complex *prec_eig) {
  double_complex *t = NULL;
  New_1D_Array<double_complex, long>(&t, TI.allocLocal);
  long i;
  long len = TI.n0Info.local * TI.n1Info.size;
  for (i = 0; i < len; i++) {
    t[i] = toep[i];
  }
  toep_solver_plan_v20(t, &TI, toep_embed_eig, prec_eig);
  Delete_1D_Array<double_complex>(&t);
}

//-------------------------------------------------------------------------
void ToepSolver_v20::
MatMult(double_complex *toep_embed_eig, double *x, double *y) {
  double_complex *cx = NULL, *cy = NULL;
  New_1D_Array<double_complex, long>(&cx, TI.allocLocal);
  New_1D_Array<double_complex, long>(&cy, TI.allocLocal);
  long i;
  long len = TI.n0Info.local * TI.n1Info.size;
  for (i = 0; i < len; i++) {
    cx[i] = x[i];
  }
  toep_solver_matvec_v20(toep_embed_eig, cx, &TI, cy);
  for (i = 0; i < len; i++) {
    y[i] = real(cy[i]);
  }
  Delete_1D_Array<double_complex>(&cx);
  Delete_1D_Array<double_complex>(&cy);
}

//-------------------------------------------------------------------------
void ToepSolver_v20::
MatSolve(double_complex *toep_embed_eig, double_complex *prec_eig,
         double *b, double *x, double rtol, int maxit) {
  double_complex *cb = NULL, *cx = NULL;
  New_1D_Array<double_complex, long>(&cb, TI.allocLocal);
  New_1D_Array<double_complex, long>(&cx, TI.allocLocal);
  long i;
  long len = TI.n0Info.local * TI.n1Info.size;
  for (i = 0; i < len; i++) {
    cb[i] = b[i];
    cx[i] = x[i];
  }
  toep_solver_solve_v20(toep_embed_eig, prec_eig, cb, rtol, maxit, &TI, cx);
  for (i = 0; i < len; i++) {
    x[i] = real(cx[i]);
  }
  // Should really optimize the following out -----------------------
  double res = toep_solver_residual_v20(toep_embed_eig, cx, cb, &TI);
  double normb_local = 0.0, normb;
  for (i = 0; i < len; i++) {
    normb_local += b[i] * b[i];
  }
  MPI_Allreduce(&normb_local, &normb, 1, MPI_DOUBLE, MPI_SUM, TI.mpiComm);
  normb = sqrt(normb);
  if (res/normb > rtol) {
    throw std::string("ToepSolver_v20::MatSolve: Linear solve not converge!\n");
  }
  //-----------------------------------------------------------------
  Delete_1D_Array<double_complex>(&cb);
  Delete_1D_Array<double_complex>(&cx);
}

//-------------------------------------------------------------------------
//-------------------------------------------------------------------------
//-------------------------------------------------------------------------
//---------------------------------------------------------- ToepSolver_v25
ToepSolver_v25::
ToepSolver_v25(int ndim, long *dim, int *nproc, MPI_Comm &comm) {
  int *idim = NULL;
  New_1D_Array<int, int>(&idim, ndim);
  int i;
  for (i = 0; i < ndim; i++) {
    idim[i] = dim[i];
  }
  toep_solver_init_v25(comm, nproc[0], ndim, idim, &TI);
  Delete_1D_Array<int>(&idim);
}

//-------------------------------------------------------------------------
ToepSolver_v25::
~ToepSolver_v25() {
  toep_solver_finalize_v25(&TI);
}

//-------------------------------------------------------------------------
void ToepSolver_v25::
GetPartitioning(long *dim_start, long *dim_len) {
  dim_start[0] = TI.n0Info.localStart;
  dim_len[0] = TI.n0Info.local;
  dim_start[1] = TI.n1ColInfo.localStart;
  dim_len[1] = TI.n1ColInfo.local;
  int i;
  for (i = 2; i < TI.rank; i++) {
    dim_start[i] = 0;
    dim_len[i] = TI.ns[i];
  }
}

//-------------------------------------------------------------------------
void ToepSolver_v25::
GetLens(long &len_toep_embed, long &len_precond) {
  len_toep_embed = TI.pwrAllocLocal * 2;
  len_precond = TI.allocLocal;
}

//-------------------------------------------------------------------------
void ToepSolver_v25::
Planning(double *toep, double_complex *toep_embed_eig,
         double_complex *prec_eig) {
  double_complex *t = NULL;
  New_1D_Array<double_complex, long>(&t, TI.allocLocal);
  long i;
  long len = TI.n0Info.local * TI.n1ColInfo.local * TI.n2Info.size;
  for (i = 0; i < len; i++) {
    t[i] = toep[i];
  }
  toep_solver_plan_v25(t, &TI, toep_embed_eig, prec_eig);
  Delete_1D_Array<double_complex>(&t);
}

//-------------------------------------------------------------------------
void ToepSolver_v25::
MatMult(double_complex *toep_embed_eig, double *x, double *y) {
  double_complex *cx = NULL, *cy = NULL;
  New_1D_Array<double_complex, long>(&cx, TI.allocLocal);
  New_1D_Array<double_complex, long>(&cy, TI.allocLocal);
  long i;
  long len = TI.n0Info.local * TI.n1ColInfo.local * TI.n2Info.size;
  for (i = 0; i < len; i++) {
    cx[i] = x[i];
  }
  toep_solver_matvec_v25(toep_embed_eig, cx, &TI, cy);
  for (i = 0; i < len; i++) {
    y[i] = real(cy[i]);
  }
  Delete_1D_Array<double_complex>(&cx);
  Delete_1D_Array<double_complex>(&cy);
}

//-------------------------------------------------------------------------
void ToepSolver_v25::
MatSolve(double_complex *toep_embed_eig, double_complex *prec_eig,
         double *b, double *x, double rtol, int maxit) {
  double_complex *cb = NULL, *cx = NULL;
  New_1D_Array<double_complex, long>(&cb, TI.allocLocal);
  New_1D_Array<double_complex, long>(&cx, TI.allocLocal);
  long i;
  long len = TI.n0Info.local * TI.n1ColInfo.local * TI.n2Info.size;
  for (i = 0; i < len; i++) {
    cb[i] = b[i];
    cx[i] = x[i];
  }
  toep_solver_solve_v25(toep_embed_eig, prec_eig, cb, rtol, maxit, &TI, cx);
  for (i = 0; i < len; i++) {
    x[i] = real(cx[i]);
  }
  // Should really optimize the following out -----------------------
  double res = toep_solver_residual_v25(toep_embed_eig, cx, cb, &TI);
  double normb_local = 0.0, normb;
  for (i = 0; i < len; i++) {
    normb_local += b[i] * b[i];
  }
  MPI_Allreduce(&normb_local, &normb, 1, MPI_DOUBLE, MPI_SUM, TI.mpiComm);
  normb = sqrt(normb);
  if (res/normb > rtol) {
    throw std::string("ToepSolver_v25::MatSolve: Linear solve not converge!\n");
  }
  //-----------------------------------------------------------------
  Delete_1D_Array<double_complex>(&cb);
  Delete_1D_Array<double_complex>(&cx);
}

//-------------------------------------------------------------------------
//-------------------------------------------------------------------------
//-------------------------------------------------------------------------
//-------------------------------------------------------------- ToepSolver
ToepSolver::
ToepSolver() {
  NDim = 0;
  Dim = NULL;
  NProc = NULL;
  mSolverVersion = UNKNOWN_SOLVER_VERSION;
  pToepSolver_v10 = NULL;
  pToepSolver_v15 = NULL;
  pToepSolver_v20 = NULL;
  pToepSolver_v25 = NULL;
}

//-------------------------------------------------------------------------
ToepSolver::
ToepSolver(const ToepSolver& S) {
  Dim = NULL;
  NProc = NULL;
  pToepSolver_v10 = NULL;
  pToepSolver_v15 = NULL;
  pToepSolver_v20 = NULL;
  pToepSolver_v25 = NULL;
  DeepCopy(S);
}

//-------------------------------------------------------------------------
void ToepSolver::
DeepCopy(const ToepSolver& S) {
  NDim = S.NDim;
  New_1D_Array<long, int>(&Dim, NDim);
  Set_Array<long, int>(Dim, S.Dim, NDim);
  New_1D_Array<int, int>(&NProc, NDim);
  Set_Array<int, int>(NProc, S.NProc, NDim);
  Comm = S.Comm;
  mSolverVersion = S.mSolverVersion;
  switch (mSolverVersion) {
  case v10:
    pToepSolver_v10 = new ToepSolver_v10(NDim, Dim, NProc, Comm);
    break;
  case v15:
    pToepSolver_v15 = new ToepSolver_v15(NDim, Dim, NProc, Comm);
    break;
  case v20:
    pToepSolver_v20 = new ToepSolver_v20(NDim, Dim, NProc, Comm);
    break;
  case v25:
    pToepSolver_v25 = new ToepSolver_v25(NDim, Dim, NProc, Comm);
    break;
  case UNKNOWN_SOLVER_VERSION:
    break;
  }
}

//-------------------------------------------------------------------------
ToepSolver& ToepSolver::
operator= (const ToepSolver& S) {
  if (this == &S) {
    return *this;
  }
  DeAllocMem();
  DeepCopy(S);
  return *this;
}

//-------------------------------------------------------------------------
void ToepSolver::
DeAllocMem(void) {
  Delete_1D_Array<long>(&Dim);
  Delete_1D_Array<int>(&NProc);
  delete pToepSolver_v10; pToepSolver_v10 = NULL;
  delete pToepSolver_v15; pToepSolver_v15 = NULL;
  delete pToepSolver_v20; pToepSolver_v20 = NULL;
  delete pToepSolver_v25; pToepSolver_v25 = NULL;
}

//-------------------------------------------------------------------------
ToepSolver::
~ToepSolver() {
  DeAllocMem();
}

//-------------------------------------------------------------------------
void ToepSolver::
SetDimAndOpt(int ndim, long *dim, int *nproc, bool opt, MPI_Comm &comm) {
  DeAllocMem();

  NDim = ndim;
  New_1D_Array<long, int>(&Dim, NDim);
  Set_Array<long, int>(Dim, dim, NDim);
  New_1D_Array<int, int>(&NProc, NDim);
  Set_Array<int, int>(NProc, nproc, NDim);
  Comm = comm;

  // Carefully play with ndim and nproc
  int total_nproc;
  MPI_Comm_size(Comm, &total_nproc);

  int total_nproc2 = 1;
  int i;
  for (i = 0; i < NDim; i++) {
    total_nproc2 *= NProc[i];
  }
  if (total_nproc != total_nproc2) {
    throw std::string("ToepSolver::SetDim: Prod(nproc) != #processes!\n");
  }

  if (NDim == 1) {
    throw std::string("Error: This library does not support 1D data!\n");
  }

  if (NDim == 2) {
    for (i = 1; i < NDim; i++) {
      if (NProc[i] != 1) {
        throw std::string("Error: 2D data can only use 1D partitioning!\n");
      }
    }
    if (opt) {
      mSolverVersion = v20;
      pToepSolver_v20 = new ToepSolver_v20(NDim, Dim, NProc, Comm);
    }
    else {
      mSolverVersion = v10;
      pToepSolver_v10 = new ToepSolver_v10(NDim, Dim, NProc, Comm);
    }
  }

  if (NDim >= 3) {
    for (i = 2; i < NDim; i++) {
      if (NProc[i] != 1) {
        throw std::string("Error: Supports only 1D or 2D partitioning!\n");
      }
    }
    if (NProc[1] == 1) {
      if (opt) {
        mSolverVersion = v20;
        pToepSolver_v20 = new ToepSolver_v20(NDim, Dim, NProc, Comm);
      }
      else {
        mSolverVersion = v10;
        pToepSolver_v10 = new ToepSolver_v10(NDim, Dim, NProc, Comm);
      }
    }
    else {
      if (opt) {
        mSolverVersion = v25;
        pToepSolver_v25 = new ToepSolver_v25(NDim, Dim, NProc, Comm);
      }
      else {
        mSolverVersion = v15;
        pToepSolver_v15 = new ToepSolver_v15(NDim, Dim, NProc, Comm);
      }
    }
  }
}

//-------------------------------------------------------------------------
void ToepSolver::
GetPartitioning(long *dim_start, long *dim_len) {
  switch (mSolverVersion) {
  case v10:
    pToepSolver_v10->GetPartitioning(dim_start, dim_len);
    break;
  case v15:
    pToepSolver_v15->GetPartitioning(dim_start, dim_len);
    break;
  case v20:
    pToepSolver_v20->GetPartitioning(dim_start, dim_len);
    break;
  case v25:
    pToepSolver_v25->GetPartitioning(dim_start, dim_len);
    break;
  case UNKNOWN_SOLVER_VERSION:
    throw std::string("ToepSolver::GetPartitioning: Unknown solver version!\n");
    break;
  }
}

//-------------------------------------------------------------------------
void ToepSolver::
GetLens(long &len_toep_embed, long &len_precond) {
  switch (mSolverVersion) {
  case v10:
    pToepSolver_v10->GetLens(len_toep_embed, len_precond);
    break;
  case v15:
    pToepSolver_v15->GetLens(len_toep_embed, len_precond);
    break;
  case v20:
    pToepSolver_v20->GetLens(len_toep_embed, len_precond);
    break;
  case v25:
    pToepSolver_v25->GetLens(len_toep_embed, len_precond);
    break;
  case UNKNOWN_SOLVER_VERSION:
    throw std::string("ToepSolver::GetLens: Unknown solver version!\n");
    break;
  }
}

//-------------------------------------------------------------------------
void ToepSolver::
Planning(double *toep, double_complex *toep_embed_eig,
         double_complex *prec_eig) {
  switch (mSolverVersion) {
  case v10:
    pToepSolver_v10->Planning(toep, toep_embed_eig, prec_eig);
    break;
  case v15:
    pToepSolver_v15->Planning(toep, toep_embed_eig, prec_eig);
    break;
  case v20:
    pToepSolver_v20->Planning(toep, toep_embed_eig, prec_eig);
    break;
  case v25:
    pToepSolver_v25->Planning(toep, toep_embed_eig, prec_eig);
    break;
  case UNKNOWN_SOLVER_VERSION:
    throw std::string("ToepSolver::Planning: Unknown solver version!\n");
    break;
  }
}

//-------------------------------------------------------------------------
void ToepSolver::
MatMult(double_complex *toep_embed_eig, double *x, double *y) {
  switch (mSolverVersion) {
  case v10:
    pToepSolver_v10->MatMult(toep_embed_eig, x, y);
    break;
  case v15:
    pToepSolver_v15->MatMult(toep_embed_eig, x, y);
    break;
  case v20:
    pToepSolver_v20->MatMult(toep_embed_eig, x, y);
    break;
  case v25:
    pToepSolver_v25->MatMult(toep_embed_eig, x, y);
    break;
  case UNKNOWN_SOLVER_VERSION:
    throw std::string("ToepSolver::MatMult: Unknown solver version!\n");
    break;
  }
}

//-------------------------------------------------------------------------
void ToepSolver::
MatSolve(double_complex *toep_embed_eig, double_complex *prec_eig,
         double *b, double *x, double rtol, int maxit) {
  switch (mSolverVersion) {
  case v10:
    pToepSolver_v10->MatSolve(toep_embed_eig, prec_eig, b, x, rtol, maxit);
    break;
  case v15:
    pToepSolver_v15->MatSolve(toep_embed_eig, prec_eig, b, x, rtol, maxit);
    break;
  case v20:
    pToepSolver_v20->MatSolve(toep_embed_eig, prec_eig, b, x, rtol, maxit);
    break;
  case v25:
    pToepSolver_v25->MatSolve(toep_embed_eig, prec_eig, b, x, rtol, maxit);
    break;
  case UNKNOWN_SOLVER_VERSION:
    throw std::string("ToepSolver::MatSolve: Unknown solver version!\n");
    break;
  }
}

//-------------------------------------------------------------------------
//-------------------------------------------------------------------------
//-------------------------------------------------------------------------
//--------------------------------------------------------- ToepSolverParam
ToepSolverParam::ToepSolverParam() {
  RTol = 1e-6;
  MaxIt = 200;
}

//--------------------------------------------------------------------------
ToepSolverParam::ToepSolverParam(double rtol, int maxit) {
  SetToepSolverParam(rtol, maxit);
}

//--------------------------------------------------------------------------
ToepSolverParam::ToepSolverParam(const ToepSolverParam &P) {
  DeepCopy(P);
}

//--------------------------------------------------------------------------
void ToepSolverParam::DeepCopy(const ToepSolverParam &P) {
  RTol = P.RTol;
  MaxIt = P.MaxIt;
}

//--------------------------------------------------------------------------
ToepSolverParam& ToepSolverParam::operator= (const ToepSolverParam &P) {
  if (this == &P) {
    return *this;
  }
  DeAllocMem();
  DeepCopy(P);
  return *this;
}

//--------------------------------------------------------------------------
ToepSolverParam::~ToepSolverParam() {
  DeAllocMem();
}

//--------------------------------------------------------------------------
void ToepSolverParam::DeAllocMem(void) {
}

//--------------------------------------------------------------------------
void ToepSolverParam::SetToepSolverParam(double rtol, int maxit) {
  RTol = rtol;
  MaxIt = maxit;
}

// Author: J. Chen
// $Date: 2013/03/28 15:50:43$
