#ifndef _TOEP_SOLER_WRAPPER_
#define _TOEP_SOLER_WRAPPER_
// A C++ wrapper for the Toeplitz sovler implemented under ToepSolver/.
// Contains the following classes:
//  class ToepSolver_v10  // 1D partitioning, AllReduce()
//  class ToepSolver_v15  // 2D partitioning, AllReduce()
//  class ToepSolver_v20  // 1D partitioning, no AllReduce()
//  class ToepSolver_v25  // 2D partitioning, no AllReduce()
//  class ToepSolver      // The unified interface. Always use this one
//  class ToepSolverParam // Parameters

#include <mpi.h>
#include <complex>
#include "ToepSolver/toep_solver_v1.0.h"
#include "ToepSolver/toep_solver_v1.5.h"
#include "ToepSolver/toep_solver_v2.0.h"
#include "ToepSolver/toep_solver_v2.5.h"

//-------------------------------------------------------------------------
class ToepSolver_v10 {

public:

  // Constructor, destructor
  ToepSolver_v10(int ndim, long *dim, int *nproc, MPI_Comm &comm);
  ~ToepSolver_v10();

  // Get
  void GetPartitioning(long *dim_start, long *dim_len);
  void GetLens(long &len_toep_embed, long &len_precond);

  // Major computational component
  void Planning(double *toep, double_complex *toep_embed_eig,
                double_complex *prec_eig);
  void MatMult(double_complex *toep_embed_eig, double *x, double *y);
  void MatSolve(double_complex *toep_embed_eig, double_complex *prec_eig,
                double *b, double *x, double rtol, int maxit);

private:

  ToeplitzSolverInfo_v10 TI;

};

//-------------------------------------------------------------------------
class ToepSolver_v15 {

public:

  // Constructor, destructor
  ToepSolver_v15(int ndim, long *dim, int *nproc, MPI_Comm &comm);
  ~ToepSolver_v15();

  // Get
  void GetPartitioning(long *dim_start, long *dim_len);
  void GetLens(long &len_toep_embed, long &len_precond);

  // Major computational component
  void Planning(double *toep, double_complex *toep_embed_eig,
                double_complex *prec_eig);
  void MatMult(double_complex *toep_embed_eig, double *x, double *y);
  void MatSolve(double_complex *toep_embed_eig, double_complex *prec_eig,
                double *b, double *x, double rtol, int maxit);

private:

  ToeplitzSolverInfo_v15 TI;

};

//-------------------------------------------------------------------------
class ToepSolver_v20 {

public:

  // Constructor, destructor
  ToepSolver_v20(int ndim, long *dim, int *nproc, MPI_Comm &comm);
  ~ToepSolver_v20();

  // Get
  void GetPartitioning(long *dim_start, long *dim_len);
  void GetLens(long &len_toep_embed, long &len_precond);

  // Major computational component
  void Planning(double *toep, double_complex *toep_embed_eig,
                double_complex *prec_eig);
  void MatMult(double_complex *toep_embed_eig, double *x, double *y);
  void MatSolve(double_complex *toep_embed_eig, double_complex *prec_eig,
                double *b, double *x, double rtol, int maxit);

private:

  ToeplitzSolverInfo_v20 TI;

};

//-------------------------------------------------------------------------
class ToepSolver_v25 {

public:

  // Constructor, destructor
  ToepSolver_v25(int ndim, long *dim, int *nproc, MPI_Comm &comm);
  ~ToepSolver_v25();

  // Get
  void GetPartitioning(long *dim_start, long *dim_len);
  void GetLens(long &len_toep_embed, long &len_precond);

  // Major computational component
  void Planning(double *toep, double_complex *toep_embed_eig,
                double_complex *prec_eig);
  void MatMult(double_complex *toep_embed_eig, double *x, double *y);
  void MatSolve(double_complex *toep_embed_eig, double_complex *prec_eig,
                double *b, double *x, double rtol, int maxit);

private:

  ToeplitzSolverInfo_v25 TI;

};

//-------------------------------------------------------------------------
typedef enum { UNKNOWN_SOLVER_VERSION, v10, v15, v20, v25 } SOLVER_VERSION;

class ToepSolver {

public:

  // Constructor, destructor
  ToepSolver();
  ToepSolver(const ToepSolver& S);
  ToepSolver& operator= (const ToepSolver& S);
  ~ToepSolver();

  // Set
  void SetDimAndOpt(int ndim, long *dim, int *nproc, bool opt, MPI_Comm &comm);

  // Get
  void GetPartitioning(long *dim_start, long *dim_len);
  void GetLens(long &len_toep_embed, long &len_precond);

  // Major computational component
  void Planning(double *toep, double_complex *toep_embed_eig,
                double_complex *prec_eig);
  void MatMult(double_complex *toep_embed_eig, double *x, double *y);
  void MatSolve(double_complex *toep_embed_eig, double_complex *prec_eig,
                double *b, double *x, double rtol, int maxit);

private:

  // Needed by constructor and destructor
  void DeepCopy(const ToepSolver& S);
  void DeAllocMem(void);

  int NDim;
  long *Dim;
  int *NProc;
  MPI_Comm Comm;

  SOLVER_VERSION mSolverVersion;
  ToepSolver_v10 *pToepSolver_v10;
  ToepSolver_v15 *pToepSolver_v15;
  ToepSolver_v20 *pToepSolver_v20;
  ToepSolver_v25 *pToepSolver_v25;

};

//--------------------------------------------------------------------------
class ToepSolverParam {

public:

  // Constructor, destructor
  ToepSolverParam();
  ToepSolverParam(double rtol, int maxit);
  ToepSolverParam(const ToepSolverParam &P);
  ToepSolverParam& operator= (const ToepSolverParam &P);
  ~ToepSolverParam();

  // Set
  void SetToepSolverParam(double rtol, int maxit);

  // No need to conceal them
  double RTol;                  // Relative tolerance
  int MaxIt;                    // Maximum number of iterations

private:

  // Needed by constructor and destructor
  void DeepCopy(const ToepSolverParam &P);
  void DeAllocMem(void);

};

#endif

// Author: J. Chen
// $Date: 2013/03/28 15:50:43$
