#ifndef _UTILITY_
#define _UTILITY_
// Defines miscellaneous utilities.

#include <mpi.h>
#include <stdlib.h>
#include <string>
#include <string.h>
#include <sstream>
#include <iostream>

//--------------------------------------------------------------------------
// Exception handler
void ExceptionHandler(std::string msg, MPI_Comm comm);

//--------------------------------------------------------------------------
// Math functions
double square(double x);

// Modified Bessel function of the second kind
double besselk(double nu, double x);

// The following function computes y = besselk(|nu|,x).
// If y1 is not NULL, then y1 = besselk(||nu|-1|,x).
// Further, if y2 is not NULL, then y2 = besselk(|||nu|-1|-1|,x).
void besselk2(double nu, double x, double *y, double *y1 = NULL, double *y2 = NULL);

extern "C" {
// Modified Bessel function of the second kind.
// This is a Fortran code.
void zbesk_(double *zr,    // Input, real part
            double *zi,    // Input, imaginary part
            double *fnu,   // Order
            int *kode,     // Scaling option. 1 = no scaling
            int *n,        // Computes order from fnu to fnu + n -1
            double cyr[],  // Outputs, real part, for different orders
            double cyi[],  // Outputs, imaginary part, for different orders
            int *nz,       // Number of elements in cyr/cyi set to zero
            int *ierr      // Error flag
            );

// Lapack routines.
void dgetrf_(int *m, int *n, double *A, int *lda, int *ipiv, int *info);
void dgetri_(int *n, double *A, int *lda, int *ipiv, double *work, int *lwork, int *info);
}

//--------------------------------------------------------------------------
// Templated new and delete. For all the delete functions, the pointer
// is ensured to set to NULL. For all the new functions, the allocated
// memory is set to zero.
//
// Provides the following interfaces:
//
// 1a. Delete a 1D array
//   void Delete_1D_Array(T **a);
//
// 1b. New a 1D array of length d1
//   void New_1D_Array(T **a, T1 d1);
//
// 2a. Delete a 2D array (first dimension d1)
//   void Delete_2D_Array(T ***a, T1 d1);
//
// 2b. New a 2D array of size d1*d2
//   void New_2D_Array(T ***a, T1 d1, T2 d2);
//
// 2c. New a 2D array. First dimension d1, second dimension varying
//   void New_2D_Array(T ***a, T1 d1, T2 *d2);
//
// 2d. New a 2D symmetric array of size d1*d1. Only allocated memory for
//     entries [i][j] where j <= i
//   void New_2D_Array_Sym(T ***a, T1 d1);
//
// 3a. Delete a 3D array (first dimension d1, second dimension d2)
//   void Delete_3D_Array(T ****a, T1 d1, T2 d2);
//
// 3b. New a 3D array of size d1*d2*d3
//   void New_3D_Array(T ****a, T1 d1, T2 d2, T3 d3);
//
// 3c. Delete a 3D array, where the first two dimensions are symmetric.
//     The First dimension has length d1. Must be used together with
//     New_3D_Array_Sym2D()
//   void Delete_3D_Array_Sym2D(T ****a, T1 d1);
//
// 3d. New a 3D array of size d1*d2*d3, where the first two dimensions are
//     symmetric. Only allocated memory for entries [i][j][k] where j <= i
//   void New_3D_Array_Sym2D(T ****a, T1 d1, T3 d3);

template <class T>
void Delete_1D_Array(T **a) {
  delete [] (*a);
  (*a) = NULL;
}

template <class T, class T1>
void New_1D_Array(T **a, T1 d1) {
  (*a) = new T [d1];
  memset((*a), 0, d1*sizeof(T));
}

template <class T, class T1>
void Delete_2D_Array(T ***a, T1 d1) {
  T1 i;
  if (*a) {
    for (i = 0; i < d1; i++) {
      delete [] (*a)[i];
    }
    delete [] (*a);
    (*a) = NULL;
  }
}

template <class T, class T1, class T2>
void New_2D_Array(T ***a, T1 d1, T2 d2) {
  (*a) = new T * [d1];
  T1 i;
  for (i = 0; i < d1; i++) {
    (*a)[i] = new T [d2];
    memset((*a)[i], 0, d2*sizeof(T));
  }
}

template <class T, class T1, class T2>
void New_2D_Array(T ***a, T1 d1, T2 *d2) {
  (*a) = new T * [d1];
  T1 i;
  for (i = 0; i < d1; i++) {
     (*a)[i] = new T [d2[i]];
    memset((*a)[i], 0, d2[i]*sizeof(T));
  }
}

template <class T, class T1>
void New_2D_Array_Sym(T ***a, T1 d1) {
  (*a) = new T * [d1];
  T1 i;
  for (i = 0; i < d1; i++) {
    (*a)[i] = new T [i+1];
    memset((*a)[i], 0, (i+1)*sizeof(T));
  }
}

template <class T, class T1, class T2>
void Delete_3D_Array(T ****a, T1 d1, T2 d2) {
  T1 i;
  T2 j;
  if (*a) {
    for (i = 0; i < d1; i++) {
      if ((*a)[i]) {
        for (j = 0; j < d2; j++) {
          delete [] (*a)[i][j];
        }
        delete [] (*a)[i];
      }
    }
    delete [] (*a);
    (*a) = NULL;
  }
}

template <class T, class T1, class T2, class T3>
void New_3D_Array(T ****a, T1 d1, T2 d2, T3 d3) {
  (*a) = new T ** [d1];
  T1 i;
  T2 j;
  for (i = 0; i < d1; i++) {
    (*a)[i] = new T * [d2];
    for (j = 0; j < d2; j++) {
      (*a)[i][j] = new T [d3];
      memset((*a)[i][j], 0, d3*sizeof(T));
    }
  }
}

template <class T, class T1>
void Delete_3D_Array_Sym2D(T ****a, T1 d1) {
  T1 i, j;
  if (*a) {
    for (i = 0; i < d1; i++) {
      if ((*a)[i]) {
        for (j = 0; j <= i ; j++) {
          delete [] (*a)[i][j];
        }
        delete [] (*a)[i];
      }
    }
    delete [] (*a);
    (*a) = NULL;
  }
}

template <class T, class T1, class T3>
void New_3D_Array_Sym2D(T ****a, T1 d1, T3 d3) {
  (*a) = new T ** [d1];
  T1 i, j;
  for (i = 0; i < d1; i++) {
    (*a)[i] = new T * [i+1];
    for (j = 0; j <= i; j++) {
      (*a)[i][j] = new T [d3];
      memset((*a)[i][j], 0, d3*sizeof(T));
    }
  }
}

//--------------------------------------------------------------------------
// Templated functions for setting the values of an array a of length n
template <class T, class LenT>
void Set_Array(T *a, T val, LenT n) {
  LenT i;
  for (i = 0; i < n; i++) {
    a[i] = val;
  }
}

template <class T, class LenT>
void Set_Array(T *a, T *b, LenT n) {
  memcpy(a, b, n*sizeof(T));
}

//--------------------------------------------------------------------------
// Templated swap
template <class T>
void Swap(T &a, T &b) {
  T tmp = a;
  a = b;
  b = tmp;
}

#endif

// Author: J. Chen
// $Date: 2013/03/28 15:50:43$
